/**
\page alpha_tut_page Tutorial 4

\section alpha_sec Alpha Composited Windows

Some desktops allow windows to be composited or blended with the rest of the desktop (e.g. compiz for X windows based systems, Windows Vista and later).

GLK provides a number of flags which can be passed into the GLK::GlWindow constructor to control the windows appearance.
The GLK::WindowFlags enum is defined in WindowTypes.h and provides the flag GLK::AlphaComposited which will tell the window manager to composite
the window with the rest of the desktop. If supported the OpenGL framebuffer for the window will be set up with an alpha channel, allowing you to control
which areas of your window are opaque and which are blended with the desktop.

NOTE: whether compositing is supported depends on your platform, if it is not supported the flag will have no effect (the same is true for some of the other flags).
On linux systems you usually have to enable compositing by activating a 'compositing window manager'. This can often be activated from the preferences->appearance menu.

\subsection alpha_example Example

\image html alpha_window.png "Alpha Compositing with GLK" width=10cm

This example renders some text with a drop shadow over the desktop. The completed example can be found in the <em>'examples/alpha_window'</em> directory but if you want to follow along then begin with the finished code from the font tutorial \ref font_sec ( DerivedWindowModified.h in <em>'examples/hello_world'</em>).

First change the name of the window class from DerivedWindowModified to AlphaWindow. Next in the constructor set the alpha flag like this:

\code
class AlphaWindow : public GLK::GlWindow
{
public:
    AlphaWindow( GLK::String title ) : GlWindow( 10, 10, GLK::AlphaComposited )
    {
        //... leave rest of constructor unchanged

\endcode

Add a timer to the class so we can do a simple animation:

\code
private:
    Ft::Library* m_library;
    GLK::GlFont* m_font;
    
    GLK::Timer   m_timer; // <= ADD THIS LINE
\endcode

In the function AlphaWindow::InitGL() change the clear colour so that it is completely transparent (alpha component of 0).
Also find the line which sets the font size and make it bigger:

\code
        glClearColor( 0.f, 0.f, 0.f, 0.f ); // SET BLACK AND COMPLETELY TRASPARENT BACKGROUND COLOR

        // ...        
        
        if ( m_font->FontLoaded() )
        {
                m_font->SetSizePx( 64 ); // CHANGE TO 64 PIXEL FONT SIZE
                return true;
        }
\endcode

The remaining changes only affect the AlphaWindow::Render() function. (As usual basic knowledge of OpenGL is assumed.)
Find where the text is rendered and delete all the rendering code (between the call to glViewport and SwapBuffers()).
Insert the following code in its place:

\code
        // Render text with drop shadow and underline
        glTranslatef( GetWidth()/2.f,GetHeight()/2.f, 0 ); // translate text to centre of window
        glRotatef( 25.f*m_timer.GetSeconds(), 0, 0, 1 ); // time based rotation in screen plane
        float rotPointX = m_font->GetOffsetPx( "ALPHA!", 3 ); // horizontal pivot point will be immediately after 'P'
        float rotPointY = m_font->GetLineHeightPx()/2.f; // vertical pivot point is half the line height for this font
        glTranslatef( -rotPointX, -rotPointY, 0.f ); // Translate so text rotates around its centre
\endcode

This sets up the modelview transformation so that the text is drawn centred in the window, while rotating about its centre.
Below that add these lines to underline the text with a thick red line:

\code
        // Render a red under-line with a quad:
        glColor4f( 1.f, 0.f, 0.f, 0.3f );
        glBegin( GL_QUADS );
        {
            glVertex2f( -25, -5 );
            glVertex2f( 250, -5 );
            glVertex2f( 250, -20 );
            glVertex2f( -25, -20 );
        }
        glEnd();
\endcode

Note that the quad was drawn with an alpha value of 0.3 meaning it will be partly blended with the desktop background.
Next draw the drop shadow which will also be transparent (we draw it first so that it appears behind the opaque text):

\code
        // Draw shadow first so that it appears behind:
        glColor4f( 0, 0, 0, 0.6f );
        float dropOffset = -m_font->GetSizePx()/20.f;
        glPushMatrix();
        {
            glTranslatef( dropOffset, dropOffset, 0.f );
            m_font->RenderString( "ALPHA!" );
        }
        glPopMatrix();
\endcode

Note we pushed and popped matrices to loose the drop shadow offset. We can now render the opaque text by setting an alpha value of 1 in glColor4f():

\code
        // Now render opaque text:
        glColor4f( 0, 0, 0, 1.f );
        m_font->RenderString( "ALPHA!" );
\endcode

Now enable compositing in you windowing system (if necessary) and compile and run a main program which uses the AlphaWindow class (see the hello world tutorial \ref hello_sec).
You should see that the client area's background is transparent and the rotating text being opaque. Note that the window borders are still being drawn which ruins the effect a little.
Window decorations can be removed by specifying the GLK::Decorless flag as well as GLK::AlphaComposited in the constructor. Note that flags must be bitwise-OR-ed together:

\code
    AlphaWindow( GLK::String title ) : GlWindow( 10, 10, GLK::AlphaComposited | GLK::Decorless )
\endcode

On Linux/X-windows the window manager is allowed to ignore requests for borderless windows, however most modern managers allow them.
Because the window has no borders you will have to move it and close it using the desktop's taskbar.


\ref lua_tut_page
**/

