/**
\page lua_tut_page Tutorial 5

\section lua_intro_sec GLK and Lua

\image html lua.png "Window rendered from Lua script" width=5cm

Lua is a lightweight scripting laguage that is easily embedded in programs
but can also be used as a standalone language. This is not intended to be a
Lua tutorial but shows how to use GLK with Lua.

Lua is integrated with GLK in three ways:

- The namespace Lua provides a C++ wrapper for Lua's functionality. 

- The class GLK::LuaWindow allows your host C++ program to create windows
that have their rendering 'callbacks' defined in Lua scripts.

- The library <em>luaglk</em> can be used in standalone Lua programs to create
and render to GLK windows directly from Lua without a host C++ program.

<b>NOTE:</b> GLK does not provide an OpenGL to Lua binding. GLK's Lua integration
simply provides the  means the create and manage GLK windows from the Lua scripting
language. The mechanism for calling OpenGL from those scripts is up to the user. It
has been tested with LuaGL (http://luagl.sourceforge.net/) and that is the binding
used in the following examples.

The source for both the following examples can be found in <em>'src/examples/lua_window'.</em>

\section lua_example_1 Using LuaWindow

You can find a full example program in <em>'src/examples/lua_window/main.cpp'</em>. 

The corresponding lua script that contains the rendering code is found in <em>'src/examples/lua_window/lua-window.lua'</em>.

\section lua_example_2 Using the luaglk Dynamic Library

Using GLK directly from a standalone Lua program is relatively simple.

The following is a minimal set of commands that load and use luaglk. Try
it from the interactive lua interpreter (http://www.lua.org/manual/5.1/lua.html):

\code
$ lua
Lua 5.1.4  Copyright (C) 1994-2008 Lua.org, PUC-Rio
> require 'luaglk'
> lua_window = glk.NewWindow( 'LuaGLK Test', 320, 240 )
> lua_window.InitGL = function() return true end
> lua_window.Start()

\endcode

The above should create an window frame with no contents (we haven't defined any rendering functions).
If you get a seg fault see \ref sec_known_issue.

In order to render to this window we only need a few more lines. Firstly we need to import
a Lua to OpenGL binding. Try LuaGL (download and install from http://luagl.sourceforge.net/).
<b>Note:</b> the current version of LuaGL does not install itself correctly, so you will need
to do it manually.

Once it is installed try loading it in the interpreter as we did
for luaglk:
\code
> require 'luagl'
\endcode

The only thing that remains is to define the appropriate functions. We have already defined InitGL
(if we hadn't then the window would have closed immediately after seeing GLK::LuaWindow::InitGL return
false). Some suitable rendering functions are given in <em>'glk/src/examples/lua_window/lua-window.lua'</em>.
I would <b>not</b> recommend trying to type these at the command prompt. You can just import the file to
execute all the statements in it:
\code
> dofile 'lua-window.lua'
\endcode

You will need to resize the window to get it to re-draw. Note something interesting happend here - we changed the window's rendering functions at run-time. Open
the file <em>'glk/src/examples/lua_window/lua-window.lua'</em> and find the Render function. Try changing the clear color to black by swapping the comment:
\code
function lua_window:Render()
    --gl.ClearColor( 1, 0.5, 0, 0 )
    gl.ClearColor( 0, 0, 0, 0 )
\endcode

Save the change and re-execute the file:
\code
> dofile 'lua-window.lua'
\endcode

The window's background color should change immediately: we can update LuaGLK programs 'live'!
Obviously if we change the resize callback then we won't see the change until resize is next called,
which is why we had to resize the window to see the changes when we called <em>dofile</em> the first time.

\subsection subsec_lua_standalone Running Scripts Directly

From a shell prompt you can directly run a Lua script to automatically execute the above example:
\code
$ lua glk-standalone.lua
\endcode
(You can't edit the script 'live' in this case because you cannot access the lua command prompt.)

\subsection sec_known_issue Known Issue:

On Linux, if you get a segfault try the following (this is a bug in your Linux distribution):
\code
$ LD_PRELOAD=/lib/libpthread.so.0 lua glk-standalone.lua
\endcode

\section lua_how_it_works How does luaglk work?

Lua allows the loading of native dynamic libraries on some platforms. These libraries
can register an API with Lua that will allow Lua to make calls into the shared lib.
This is the mechanism <em>luaglk</em> uses to allow standalone Lua/GLK programs.

Roughly speaking, when you run a program that depends on a shared library, the dynamic
loader will fetch not only the required library, but also any other libraries upon which
that shared library depends. It places all the libraries in memory and makes sure your
program can see their symbols.

However, when Lua loads a shared library it must do so itself at runtime, without relying on the
operating system's dynamic loader. (It has to because you tell it to load the library using a 'require'
statement in your script - after the lua executable has started running). On Linux loading a library this
way is not straight-forward because the program which loads the shared library is responsible for loading
all other dependecies itself. This causes a problem for Lua - it does not know what other libraries a particular
shared library depends on, so cannot preload all other dependencies. This procludes us from, say, directly loading
libglk.so(Linux) as glk depends upon numerous other libraries (glkcore, pthread, opengl etc).

The solution used by GLK is that a stub library called luaglk is created. This library's sole responsibility is to
load all the dependencies for GLK and then to load glk itself and expose a minimal set of symbols from GLK to Lua.
You can see how this works in 'src/luaglk/linux/libluaglk.cpp' and 'src/luaglk/win32/libluaglk.cpp'. While this is
not necessary on Windows the same mechanism is used there for consistency (so that all lua scripts load 'luaglk' no
matter what platform you are on).

If you wanted to modify GLK and expose new funtionality to Lua you would follow two simple steps:

- Add a function of type LuaFuncType to GLK that exposes the functionality (see examples in LuaStandAlone.cpp).
- Dynamically load the symbol for your function in libluaglk.cpp (in RegisterLuaGlk() ).

The advantage of this mechanism is that all the complexity remains in GLK and the library loaded by Lua is just a thin
interface to the rest of GLK (you can see for yourself how short libluaglk.cpp is).
*/
