// Copyright (c) Mark Pupilli 2010, All Rights Reserved.

#include <stdio.h>

#include <glkcore.h>
#include <glk.h>

using namespace GLK;

/*
    This example shows one possible use of the LuaWindow class.
    
    By creating a LuaWindow object you can bind the OpenGL interface
    to a lua script which defines InitGL, Update, Render, etc.
    
    You can then quickly prototype OpenGL code written in Lua without
    recompiling. With some modifications you could even rebind the
    script's functions at runtime, thus modifying the render code
    'on the fly'.
    
    The OpenGL code to go with this example is in the file 'lua-window.lua'.
*/
int main( int argc, char** argv )
{
    const char script[] = "../../../src/examples/lua_window/lua-window.lua"; // Name of the script which contains the LuaGL code

    // Setup a Lua object:
    Lua::LuaState lua;
    lua.OpenLibraries();
    LUA_REGISTER_PRINT( &lua ); // So we can print to stdout/stderr.
    LUA_REGISTER_SLEEP( &lua ); // So we can call sleep( ms ) in the Lua script.
       
    // Define a variable name for accessing the window object in the script.
    String variableName( "lua_window" );

    // Attach a Lua window to the Lua state object.
    LuaWindow win( lua, variableName, 320, 240 );
    
    // Execute the script. This must be done after the window object is constructed but before
    // we start the window's event otherwise the InitGL call will be missed. However, calling
    // it here means we can't call any functions which block waiting for the event loop to start
    // or the program will deadlock. 
    int err = lua.DoFile( script );
    
    if ( err == 0 )
    {
        win.Show(); // Make the window visible
        win.StartEventLoopAsync(); // Enter the window's event loop
        win.SetWindowName( "Examples: Lua Window" );
        return win.WaitForClose();
    }
    else
    {
        // Get error message from Lua and print it:
        if ( lua.GetStackSize() > 0 && lua.IsString() )
        {
            size_t l = 0;
            const char* s = lua.ToString( -1, &l );
            fwrite( s, l, 1, stderr );
            fwrite( "\n", 1,1, stderr );
            lua.Pop(1);
        }
        
        return err;
    }
}
