#include "ClickWindow.h"

#define NUM_PTS 1

unsigned int ClickWindow::ms_count = 0;
GLK::Mutex ClickWindow::ms_initLock;

/**
	Each of instance of a ClickWindow window
	is initialised with a message queue and
	a unique construction id.
**/
ClickWindow::ClickWindow( unsigned int w, unsigned int h )
:
	GlThreadWindow    ( w, h, GLK::Normal ), // window is constructed in base class
	m_receive   (0),				  // message queue for receiving clicks sent to this instance.
	m_send      (0),				  // pointer to another ClickWindow instance that will receive clicks
	m_readTime  (0.f)
{
	GLK::MutexLock locker( ms_initLock );

	m_receive = new GLK::MessageQueue<Point2i>( NUM_PTS );
	m_id = ms_count++;
    PushMouseHandler( &m_mouse );
}

ClickWindow::~ClickWindow()
{
	delete m_receive;
}

bool ClickWindow::InitGL()
{

	// background color depends on construction id
    if ( m_id == 1 )
    {
        glClearColor(1,1,1,1);
        glColor3f(0,0,0);
    }
    else
    {
        glClearColor(0,0,0,0);
        glColor3f(1,1,1);
    }

    glPointSize(5.f);
    
    return true;
}

void ClickWindow::DestroyGL()
{
}

void ClickWindow::Resize(const unsigned int w, const unsigned int h)
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glViewport(0,0,w,h);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho( 0,w,h,0,-1,1);

    glMatrixMode(GL_MODELVIEW);

    Render();
}

/**
	Send and receive click messages.

	Return value is only true if there is a mouse click to receive
	hence, this window is only redrawn when a new click arrives.
**/
bool ClickWindow::Update( unsigned int time )
{
	bool rval = false;

    {
		Point2i pt;
        m_readTime = static_cast<float>( time );

		// Check the message queue for incoming mouse clicks:
        if ( m_receive->TryRead( &pt, 1 ) )
        {
			m_pts = pt;
            rval = true;
        }

		// If a click receiver was registered
		// and the left button is down then send
		// the mouse coords to the receiver.
		bool ok;
		{
		    GLK::MutexLock lock( ms_initLock );
		    ok = ( m_send != 0 );
		}
		
        if ( ok && m_mouse.GetButton( GLK::MouseHandler::Left ) )
        {
			pt.x = m_mouse.GetX();
            pt.y = m_mouse.GetY();

            if ( m_send->PostClick( pt ) )
            {
                // Wake window so it can call Update() and Render()
                m_send->Notify(); 
            }

		}
    }

    return rval;
}

void ClickWindow::Render()
{
	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glBegin( GL_POINTS );
    {
		for ( unsigned int i=0; i<NUM_PTS; ++i )
        {
			glVertex2f( static_cast<GLfloat>(m_pts.x), static_cast<GLfloat>(m_pts.y) );
        }
    }
    glEnd();

    SwapBuffers();
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
}

/**
	Posts a click in this window's message queue.
	(Can be called from another thread - message
	queue ensures thread safety).

	If the message queue is full (or TryWrite()
	fails for some other reason) then the click
	message will simply be lost.
**/
bool ClickWindow::PostClick( Point2i pt )
{
    bool rval = false;

    if ( m_receive->TryWrite( &pt, 1) )
    {
        rval = true;
    }
    
    return rval;
}

