#ifndef ClickWindow_H
#define ClickWindow_H

#include <glkcore.h>
#include <glk.h>

struct Point2i
{
    int x;
    int y;
};

/**
    Simple OpenGL window which passes mouse click events to a registered receiver.
**/
class ClickWindow : public GLK::GlThreadWindow
{
    friend class ThreadWindow;

public:
    ClickWindow( unsigned int w, unsigned int h );
    virtual ~ClickWindow();

    virtual bool InitGL();
    virtual void DestroyGL();
    virtual void Resize( unsigned int w, unsigned int h );
    virtual bool Update( unsigned int time_ms );
    virtual void Render();
    virtual void TimerExpired( int ) {};

    void SetClickReceiver( ClickWindow* other ) { GLK::MutexLock lock( ms_initLock ); m_send = other; };
    bool PostClick( Point2i p );

protected:
    GLK::MessageQueue<Point2i>*  m_receive; //< message queue for receiving mouse clicks

private:
    static unsigned int ms_count;
    static GLK::Mutex	ms_initLock;

    ClickWindow*	m_send;		//< Window that will receive mouse clicks
    Point2i     m_pts;		//< buffer for storing received clicks
    float       m_readTime;	//< time of last read
    unsigned int m_id;

    GLK::BasicMouse m_mouse;
};

// global list of windows
#define MAX_WINDOWS 12
// global list of windows
static ClickWindow* g_win[MAX_WINDOWS] = { 0 };
static unsigned int g_winIndex = 1;

/*
    Lua-function to create new windows.

    Expects 2 or 3 arguments:
    1:  window width    (NUMBER)
    2:  window height   (NUMBER)
    3:  window title    (STRING) (OPTIONAL)

    Returns 0 results.

    NOTE: All Lua C functions should be documented using the above format for clarity.
*/
static int createWindow( lua_State* L )
{
    const Lua::LuaState lua( L ); // creates a LuaState object which is just a reference to L
                                 // doesn't have to be const, but let's test const correctness...

    // if there are 2 arguments and both are numbers
    if( lua.GetStackSize()>=2 && lua.IsNumber(1) && lua.IsNumber(2) )
    {
        int w = lua.ToInteger(1); // read width parameter from call stack
        int h = lua.ToInteger(2); // read height parameter from call stack

        const char* title = 0;
        if ( lua.GetStackSize() == 3 && lua.IsString(3) )
        {
            title = lua.ToString(3);
        }

        g_win[g_winIndex] = new ClickWindow( w, h ); // create a window of dimension (w x h)

        // The first window will always send its clicks to the newly created window
        g_win[0]->SetClickReceiver( g_win[g_winIndex] );

        // The newly created window sends clicks to the previously created one
        unsigned int prevIndex = g_winIndex-1;
        if ( prevIndex >= 0 )
        {
            g_win[g_winIndex]->SetClickReceiver( g_win[prevIndex] );
        }

        g_win[g_winIndex]->StartEventLoopAsync();
        g_win[g_winIndex]->Move( 20*g_winIndex, 20*g_winIndex );
        g_win[g_winIndex]->SetWindowName( title );

        g_winIndex++;
    }

    return 0; // indicate that no return values are on the stack (Lua functions can return multiple results)
}

#endif // end ClickWindow_H

