#include "ConsoleBuffer.h"

#include <memory.h>
#include <assert.h>
#include <stdio.h>

namespace GLK
{

ConsoleBuffer::ConsoleBuffer( size_t cols, size_t rows, size_t lines )
:
    m_size      ( cols * lines ),
    m_cols      ( cols ),
    m_rows      ( rows < lines ? rows : lines ), // can't have more visible rows than there are lines
    m_lines     ( lines ),
    m_pStart    ( new char[m_size] ),
    m_pEnd      ( m_pStart + m_size )
{
    Clear();
}

ConsoleBuffer::~ConsoleBuffer()
{
    delete [] m_pStart;
}

/**
    Move the cursor n characters in the buffer.
    Negative values for 'n' move cursor left, positive right.

    @param n Number of chars to move cursor.
**/
void ConsoleBuffer::MoveCursor( int n )
{
    m_cursor = m_cursor + n;

    if ( m_cursor < m_pStart )
    {
        m_cursor = m_pEnd - ( m_pStart - m_cursor );
    }
    else if ( m_cursor >= m_pEnd )
    {
        m_cursor = m_pStart + ( m_cursor - m_pEnd );
    }
}

/**
    Erase the character under the cursor by replacing
    with 'c' then moving the cursor left one character.

    @param c Character to overwrite with.
**/
void ConsoleBuffer::Erase( char c )
{
    MoveCursor( -1 );
    *m_cursor = c;
}

/**
    Put a character at the current cursor position and
    advance the cursor by one character.

    @param c Character to put in buffer.
**/
void ConsoleBuffer::Put( char c )
{
    *m_cursor++ = c;

    char* end = GetLineEnd();
    if ( end < m_pEnd )
    {
        if ( m_cursor == end ) 
        {
            // cursor wrapped around so we need to register a newline
            NewLine();
        }
        else if ( c == '\r' )
        {
            while ( m_cursor != end )
            {
                *m_cursor++ = ' ';
            }

            NewLine();
        }

    }

    if ( m_cursor == m_pEnd )
    {
        m_cursor = m_pStart;
        NewLine();
    }

    AdjustView();
}

/**
    Copy some text into the buffer at the current cursor position.
    
    @param text Buffer to copy from.
    @param numBytes The number of bytes to copy from text.
**/
void ConsoleBuffer::WriteText( const char* text, size_t numBytes )
{
    assert( numBytes < m_size ); // If the buffer is this small we should give up and go home!

    assert( m_cursor <= m_pEnd );
    size_t space = m_pEnd - m_cursor;
    if ( numBytes > space )
    {
        numBytes -= space;
        Write( text, numBytes );
        text += numBytes;
    }

    Write( text, numBytes );
   
    // If we reach end of m_buffer m_cursor will be at m_pEnd because
    // we filled the line with spaces:
    assert( m_cursor <= m_pEnd );
    if ( m_cursor == m_pEnd )
    {
        m_cursor = m_pStart;
        NewLine();
    }

    AdjustView();
}

/**
    Write formatted text into the text buffer at the current cursor position.
    
    @param formatString printf style format string (see C language stdio documentation).
    @return The number of chars written (not including null terminator) or negative value if there was an error.
**/
int ConsoleBuffer::Printf( const char* formatString, va_list args )
{
    va_list args2;
#ifdef va_copy
    va_copy(args2, args );
#else
    // No va_copy provided so assume this is safe:
    args2 = args;
#endif

    // Count how much space we need
    
    int numChars = 1 + vsnprintf( 0, 0, formatString, args ); //_vscprintf( formatString, args ); 

    assert( numChars < (int)m_size ); // If the buffer is this small we should give up and go home!

    char* tmp = new char[numChars];
    numChars = vsnprintf( tmp, numChars, formatString, args2 );
    
    assert( numChars != -1 );
    WriteText( tmp, numChars );
        
    delete [] tmp;

    return numChars;
}

/**
    Scroll the buffers visible region by 'n' lines.
    (Negative n is scrolls backwards).
**/
void ConsoleBuffer::Scroll( int n )
{
    if ( n != 0 )
    {
        m_viewStart += n;

        int max = m_line - m_rows + 1;
      
        if ( m_viewStart > max )
        {
            m_viewStart = max;    
        }
        
        if ( m_viewStart < 1 )
        {
            m_viewStart = 1;
        }
    }
}

/**
    Clear and reset the console buffer.
**/
void ConsoleBuffer::Clear()
{
    m_cursor = m_pStart;
    m_line = 1;
    m_viewStart = m_line;

    // fill visible buffer with spaces:
    FillChar( ' ', GetVisibleSize() );

    // fill visible line starts with newlines (speeds rendering of blank page)
    for ( unsigned int i=0; i<GetRows(); ++i )
    {
        *(m_cursor + (i*m_cols)) = '\n';
    }  
}

/**
    Fill buffer with 'repeat' chars from cursor but do not advance the cursor.
**/
void ConsoleBuffer::FillChar( char c, int repeat )
{
    char* p = m_cursor;
    assert(m_cursor<=m_pEnd);
    while ( repeat-- &&  p != m_pEnd )
    {
        *p++ = c;
    }

    if ( repeat > 0 )
    {
        p = m_pStart;
        while ( repeat-- )
        {
            *p++ = c;
        }
    }
}

/**
    
**/
void ConsoleBuffer::Write( const char* text, size_t numBytes )
{
    char c;
    char* end = GetLineEnd();

    if ( end == m_pStart )
    {
        end = m_pEnd;
    }
    
    if ( m_cursor == m_pEnd )
    {
        m_cursor = m_pStart;
    }

    while ( numBytes-- )
    {
        c = *text++;
        *m_cursor++ = c;
        if ( c == '\n' )
        {
            while ( m_cursor != end )
            {
                *m_cursor++ = ' ';
            }
        }

        if ( m_cursor == end )
        {
            NewLine();
            end = GetLineEnd();
            if ( end == m_pStart )
            {
                end = m_pEnd;
            }

            if ( m_cursor == m_pEnd )
            {
                m_cursor = m_pStart;
            }
        }

    }    
}

/**
    Progress the console buffer onto a new line.

    Increments line counter and fills new line with spaces.
**/
void ConsoleBuffer::NewLine()
{
    m_line++;
    FillChar( ' ', m_cols );
}

/**
    Adjust the first visible line to keep up with any changes
    to current line.
**/
void ConsoleBuffer::AdjustView()
{
    if ( m_line > (int)m_rows )
    {
        m_viewStart = m_line - m_rows + 1;
    }
}

} // end of namespace GLK
