#include "ConsoleWindow.h"

#include "../glkcore/free_type/FreeTypeSupport.h"
#include "GlFont.h"

namespace GLK
{

/**
    Replacement for lua print function which prints to a ConsoleWindow.
**/
int ConsoleWindow::consolePrint( lua_State* L )
{
    const Lua::LuaState lua( L );
    int numParams = lua.GetStackSize();

    assert( lua.IsLightUserData( lua_upvalueindex( 1 ) ) );
    ConsoleWindow* console = reinterpret_cast<ConsoleWindow*>( lua.GetUserData( lua_upvalueindex( 1 ) ) );

    int nl = -99;
    for ( int i=1; i<=numParams; ++i )
    {
        switch( lua.GetType(i) )
        {
        case LUA_TNIL:
            console->Write( "nil", 3 );
            break;
        case LUA_TNUMBER:
            nl = console->Printf( "%g", lua.ToDouble(i) );
            break;

        case LUA_TBOOLEAN:
            if ( lua.ToBool(i) )
            {
                console->Write( "true", 4 );
            }
            else
            {
                console->Write( "false", 5 );
            }
            break;

        case LUA_TSTRING:
            {
                size_t n;
                const char* s = lua.ToString( i, &n );
                console->Write( s, n );
            }
            break;

        case LUA_TTABLE:
            console->Printf( "table: " );
            break;

        case LUA_TFUNCTION:
            console->Printf( "function: " );
            break;
        case LUA_TUSERDATA:
            console->Printf( "user data: " );
            break;
        case LUA_TTHREAD:
            console->Printf( "thread: " );
            break;
        case LUA_TLIGHTUSERDATA:
            console->Printf( "light user data: " );
            break;

        default:
            break;
        }

        if ( i == numParams )
        {
            console->Write( "\n", 1 );
        }
        else
        {
            console->Write( "    ", 4 );
        }
    }

    return 0;
}

/**
    Changes the font of the console.

    Windows: The font should be located in Window's standard fonts folder.
    Linux: The full path to the font must be given.

    Expects 1 argument:
    1:  fontName    (STRING)

    Returns 0 results.

    @param fontName Should be the file name without the path or extension.
**/
int ConsoleWindow::font( const Lua::LuaState& lua, ConsoleWindow* console )
{
    if( lua.GetStackSize() > 0 && lua.IsString(1) )
    {
        const char* s = lua.ToString();

#ifdef __linux
	    String fileName( "" );
#else
        String fileName( "C:\\WINDOWS\\Fonts\\" );
#endif

        fileName += String( s );

#ifndef __linux
        fileName += String( ".ttf" );
#endif

        if ( !console->ChangeFont( fileName.cStr() ) )
        {
            console->Printf( "\nFont \'%s\' not found\n", fileName.cStr() );
        }
    }

    return 0;
}

/**
    Setup as a lua object and print Lua acknowledgement/copyright message to console.

    If the commands returned from the console are to be Lua commands then initialising
    the console by calling this function is recommended.

    The object 'console' is registered.

    'console' contains the following values:

    console.rows
    console.cols

    'console' contains the following functions:

    console.exit( number )   -- closes console window with error code set to 'number'
    console.clear()     -- clears the console's text buffer
    console.font( "font_name" ) will change the console's font to 'font_name.ttf'
    console.print( ... ) prints all arguments to the console
    console.windowed() Switch the console to a window
    console.fullscreen() Switch the console to fullscreen mode
    console.decorOff() Turn off window decorations
    console.decorOn() Turn on window decorations

    The following global functions are also added to the lua environment:

    print()     -- same as console.print() - replaces Luas built-in print()

    @param lua Reference to the LuaState object which this ConsoleWindow is to
    be associated with (i.e. commands returned by the console should be executed by this LuaState).
**/
void ConsoleWindow::SetupLua( Lua::LuaState& lua )
{   
    Printf( "Powered by %s %s\n\n", LUA_RELEASE, LUA_COPYRIGHT );

    RegisterObject( &lua, "console" );

    // 'members':
    lua.PushNumber( m_buffer->GetRows() );
    SetLuaValue( "rows" );
    lua.PushNumber( m_buffer->GetCols() );
    SetLuaValue( "cols" );

    // 'methods':
    RegisterMethod( "exit",  0 );
    RegisterMethod( "clear", 1 );
    RegisterMethod( "font",  2 );
    RegisterMethod( "fullscreen", 3 );
    RegisterMethod( "windowed", 4 );
    RegisterMethod( "decorOff", 5 );
    RegisterMethod( "decorOn", 6 );
    
    // globals:
    lua.RegisterFunction( "print", consolePrint, this );
    lua.DoString( "exit = console.exit" );
}

/**
    Lua method call dispatcher.
**/
int ConsoleWindow::Dispatch( lua_State* L, lua_Number id )
{
    int returnCount = 0;
    const Lua::LuaState lua( L );

    int i = static_cast<int>(id);
    switch ( i )
    {
    case 0:
        {
            int exitCode = 0;
            if( lua.GetStackSize()>0 && lua.IsNumber(1) )
            {
                exitCode = lua.ToInteger( 1 );
            }

            GlWindow::Quit( exitCode );
        }
        break;
    case 1:
        Clear();
        break;
    case 2:
        returnCount = font( lua, this );
        break;
    case 3:
        FullscreenOn();
        break;
    case 4:
        FullscreenOff();
        break;
    case 5:
        DecorationsOff();
        break;
    case 6:
        DecorationsOn();
        break;
    default:
        break;
    }

    return returnCount;
}

/**
    Initialises font and buffer for console then
    launches the console window (which runs asynchronously).
    
    The event loop of the window is guaranteed to have started by the time the constructor returns.
    
    @todo  ConsoleBuffer size is fixed but should be altered every time a new font is loaded!
    Will add resizing support to ConsoleBuffer at some point which will solve this.
**/
ConsoleWindow::ConsoleWindow( const char* title, int width, int height, const String& prompt )
:
    GlThreadWindow  ( width, height, Normal ),
    m_buffer        (0),
    m_library       (0),
    m_font          (0),
    m_mutex         ( Mutex::Recursive ),
    m_prompt        ( prompt ),
    m_command       ( "", 128 ),
    m_cmdPos        ( 0 ),
    m_histItr       ( m_history ),
    m_msg           ( 8 ),
    m_update        ( false ),
    m_blink         ( true ),
    m_quitting      ( false )
{   
    m_mutex.Lock();

    m_library = new Ft::Library();
    m_buffer  = new ConsoleBuffer( 80, 16, m_kNumLines ); // ConsoleBuffer needs a resize method

    GlThreadWindow::PushKeyboardHandler( this );
    GlThreadWindow::PushMouseHandler( &m_mouse );

    // Finally start event loop
    m_mutex.Unlock();

    GlThreadWindow::StartEventLoopAsync();
    GlThreadWindow::SetWindowName( title );

    StartTimer( 330 ); // need to set blink rate for cursor
    
    WindowBase::WaitForEventLoop();
}

/**
    The destructor blocks until the console window is closed.
**/
ConsoleWindow::~ConsoleWindow()
{
    GlThreadWindow::WaitForClose();

    MutexLock lock( m_mutex );
    delete m_buffer;
    delete m_library;
}

/**
    This can only be called after an opengl context has been created and
    must be called from the same thread in which it was created!
    
    @todo Need to set rows and cols in a more sensible way than below:
**/
bool ConsoleWindow::GlFontSetup()
{
    MutexLock lock( m_mutex );

    m_font = new GlFont( *m_library, DEFAULT_FONT );

#ifdef WIN32
    // On windows we attempt to fall back to courier:
    if ( !m_font->FontLoaded() )
    {
        delete m_font;
        m_font = new GlFont( *m_library, "C:\\WINDOWS\\Fonts\\cour.ttf" );
    }
#endif

    assert( m_font->FontLoaded() );
    if ( m_font->FontLoaded() )
    {
        m_font->SetSizePx( 16 );

        // setup console buffer for given font:
        size_t rows;
        size_t cols;

        m_monoSpace = m_font->GetFace().GetMaxAdvance();
    
        if ( m_font->GetFace().IsFixedWidth() )
        {
            float lineHeight = m_font->GetLineHeightPx();

            rows = (GlThreadWindow::GetHeight() / (int)lineHeight) - 2;
            cols = (GlThreadWindow::GetWidth() / (int)m_monoSpace) - 2;
        }
        else
        {
            // @todo  need to set rows and cols in a more sensible way here:
            rows = 16;
            cols = 80;    
        }

        delete m_buffer;
        m_buffer  = new ConsoleBuffer( cols, rows, m_kNumLines );
    }
    
    return m_font->FontLoaded();
}

/**
    Signal the console that something changed.
    The console can then decide if it needs to redraw
    its window as a result.
**/
void ConsoleWindow::Notify()
{
    m_update = true;
    GlThreadWindow::Notify();
}

/**
    Change the font used to render text in the console.

    If the specified font could not be loaded then the
    old font continues to be used.

    Can only be called when opengl context is active.

    @param fontFile This must be the full path and filename of the font to be used.
**/
bool ConsoleWindow::ChangeFont( const char* fontFile )
{
    bool ok = false;

    MutexLock lock( m_mutex );
    ConsoleWindow::Notify();

    if ( m_font->ChangeFont( fontFile ) )
    {
        ok = true;

        m_monoSpace = m_font->GetFace().GetMaxAdvance();
    }

    return ok;
}

/**
    Clear and reset the console buffer.
**/
void ConsoleWindow::Clear()
{
    m_buffer->Clear();
}

/**
    Copy some text into the console's text buffer for display.

    Thread safe.

    @param text Buffer to copy from.
    @param numBytes The number of bytes to copy from text.
**/
size_t ConsoleWindow::Write( const char* text, size_t numBytes )
{
    MutexLock lock( m_mutex );
    ConsoleWindow::Notify();
    m_buffer->WriteText( text, numBytes );

    return numBytes;
}

/**
    Write formatted text into the console's text buffer for display.

    Thread safe.

    @param formatString printf style format string (see C language stdio documentation).
    @return The number of chars written (not including null terminator) or negative value if there was an error.
**/
int ConsoleWindow::Printf( const char* formatString, ... )
{
    va_list args;
    va_start( args, formatString );

    MutexLock lock( m_mutex );
    ConsoleWindow::Notify();

    int n = m_buffer->Printf( formatString, args );
    va_end(args);

    return n;
}

/**
    Block until a command string is available then return it in the reference parameter.
**/
void ConsoleWindow::GetCommand( String& s )
{
    m_msg.Read( &s );
}

/**
    If command is available remove it from the queue,
    retrive it into the reference parameter, and return true.
    Otherwise return false.
**/
bool ConsoleWindow::TryCommand( String& s )
{
    return m_msg.TryRead( &s );
}

/**
    Caller will sleep until there is a new command available for reading.

    When the window exits its message loop, it passes an empty message to its own
    queue. This then causes any threads that are blocking on WaitCommand() to wake
    up. After this further calls to WaitCommand() will have no effect.
**/
void ConsoleWindow::WaitCommand()
{
    if ( !m_quitting )
    {
        m_msg.WaitMessage();
    }
}

/**
    Paste a string into the command buffer.
**/
void ConsoleWindow::PasteCommand( const String& s )
{
    m_command.Insert( s, m_cmdPos );
    m_cmdPos += s.Size();
}

void ConsoleWindow::Key( char c )
{
    MutexLock lock( m_mutex );
    ConsoleWindow::Notify();

    switch (c)
    {
    case '\v': // Vertical tab remapped as left arrow
        if ( m_cmdPos ) { m_cmdPos--; }
        m_blink = true;
        return;

    case '\f': // Form feed remapped as right arrow
        m_cmdPos++;
        if ( m_cmdPos > m_command.Size() )
        {
            m_cmdPos = m_command.Size();
        }
        m_blink = true;
        return;

    default:
        break;
    }

    if ( c == '\r' )
    {
        m_buffer->WriteText( m_prompt.cStr(), m_prompt.Size() );

        unsigned int size = m_command.Size();
        if ( size )
        {
            // permanently print command
            m_buffer->WriteText( m_command.cStr(), size );
            m_buffer->WriteText( "\n", 1 );

            // 'Execute' command
            m_history.AddFront( m_command );
            m_command += '\0';

            // Need to unlock mutex before doing a blocking write to avoid deadlocks
            m_mutex.Unlock();
            m_msg.Write( &m_command );
            m_mutex.Lock();

            m_command.Clear();
            m_cmdPos = 0;
        }

        m_buffer->WriteText( "\n", 1 );

        m_histItr.Begin();
    }
    else
    {
        switch (c)
        {
        case '\b': // BACKSPACE
            if ( m_cmdPos )
            {
                m_command.Delete(--m_cmdPos);
            }
            break;
        case 16: // DEL
            if ( m_command.Size() > 0 && m_cmdPos < m_command.Size() )
            {
                m_command.Delete( m_cmdPos );
            }
            break;
        case 28: // history lookup
            {
                if ( m_histItr.HasNext() )
                {
                    String tmp = m_histItr.Next();
                    if ( tmp.Size() )
                    {
                        m_command = tmp;
                        m_cmdPos = m_command.Size();
                    }
                }
                break;
            }
        case 29: // history lookdown
            {
                if ( m_histItr.HasPrev() )
                {
                    String tmp = m_histItr.Prev();
                    if ( tmp.Size() )
                    {
                        m_command = tmp;
                        m_cmdPos = m_command.Size();
                    }
                }
                break;
            }

        default:
            // Append char to command
            m_command.Insert( c, m_cmdPos );
            m_cmdPos++;
            break;
        }

    }

}

bool ConsoleWindow::InitGL()
{
    glClearColor( 1,1,1, 1 );
 
    return GlFontSetup();
}

void ConsoleWindow::DestroyGL()
{
    // send empty command to wake up anyone waiting for messages
    m_quitting = true;
    String final("");
    m_msg.Write( &final );

    delete m_font; // needs to be done here while gl context is still active
}

void ConsoleWindow::Resize(const unsigned int w, const unsigned int h )
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glViewport( 0, 0, w, h );

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho( 0, w, 0, h, 0.f, 5000.f );

    Render();
}

/*
    Called in response to Notify().
    If keyboard was updated ( WasUpdated() ) then we will redraw.
*/
bool ConsoleWindow::Update( unsigned int )
{
    bool rval = false;

    bool m =  m_mouse.WasUpdated();
    bool kb = WasUpdated();

    if ( kb ) { rval |= true; }

    if ( m )
    {
        int ds = m_mouse.GetDscroll();
        if (ds)
        {
            m_buffer->Scroll( -ds );
            rval |= true;
        }

        if ( m_mouse.GetDrop() )
        {
            String s( m_mouse.GetDrop() );
            for ( unsigned int i=0; i<=s.Size(); ++i )
            {
                if ( s[i] == '\\' )
                {
                    s.Insert( '\\', i++ );
                }
            }
            PasteCommand( s );
            m_mouse.ClearDrop();
            rval |= true;
        }
    }

    m_mouse.FinishUpdate();
    return rval;
}

/*
    Render a view of the console's current text buffer.
*/
void ConsoleWindow::Render()
{
    float lineHeightPx = m_font->GetLineHeightPx();

    const char* start = m_buffer->GetFirstVisibleLine();

	glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    glPushMatrix();
    glTranslatef( m_monoSpace, GetHeight() - lineHeightPx, 0.f );
    glColor3f( 0.f, 0.f, 0.f );

    m_mutex.Lock();
    m_font->RenderConsole( start, m_buffer->GetCols(), m_buffer->GetRows(), m_buffer->GetStart(), m_buffer->GetEnd() );

    String cmd = m_prompt;
    cmd += m_command;
    glPopMatrix();
    glTranslatef( m_monoSpace, lineHeightPx, 0.f );

    // render a cursor
    if ( m_blink )
    {
        glColor3f(0,0,0);
        glLineWidth( 1.f );
        float cursPosX;
        if ( m_font->GetFace().IsFixedWidth() )
        {
            cursPosX = m_monoSpace * (m_cmdPos+m_prompt.Size());
        }
        else
        {
            cursPosX = m_font->GetOffsetPx( cmd.cStr(), m_cmdPos+m_prompt.Size() );
        }
        glBegin( GL_LINES );
        {
            glVertex2f( cursPosX, 0 );
            glVertex2f( cursPosX, lineHeightPx );
        }
        glEnd();
    }

    glColor3f( 0.f, 0.4f , 0.f );
    m_font->RenderText( cmd.cStr(), cmd.Size() );

    m_mutex.Unlock();

    SwapBuffers();
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
}

/**
    The timer has been set to go off at the blink rate of the cursor, so switch blink state.
**/
void ConsoleWindow::TimerExpired( int id )
{
    assert( id == 1 );
    m_blink = !m_blink;
    Notify();
}

} // end namespace GLK
