#include "GlFont.h"

#include <assert.h>

#define DEFAULT_FONT_SIZE 16

namespace GLK
{

/**
    Constructs an OpenGl font which will load a face using the
    specified Freetype2 library reference.

    Construction sets up opengl texture and display lists objects so must
    be called while a valid opengl context is active (and from the
    same thread in which the opengl context was created).

    @params lib Reference to a valid free type library instance.
    @params font The name of the font to load (must include full path).
**/
GlFont::GlFont( Ft::Library& lib, const char* font, FT_Long index )
:
    m_freeType  ( lib ),
    m_fontFace  ( 0 ),
    m_sizePx    ( DEFAULT_FONT_SIZE )
{
    InitCache();

    m_fontFace = new Ft::Face( m_freeType, font, index );
    m_fontFace->SetPixelSizes( m_sizePx, m_sizePx );
}

/**
    The destructor should be called while the OpenGL context (in which the font object was constructed) still active.
**/
GlFont::~GlFont()
{
    glDeleteLists( m_listIndex, 256 ); // frees the glyph display lists
    glDeleteTextures( 1, &m_tabTex ); // free the tableau textureg

    // Must clear cache first because Glyphs reference objects inside
    // the loaded font face.
    ClearCache();

    delete [] m_cache;
    delete m_fontFace;
}

/**
    Change the font face loaded for this GlFont object.

    This will be expensive as the new face will have to be
    loaded and will not yet be cached. It is better practice
    to create a different GlFont object for each font that
    is regularly used and switch between them instead.

    If the new font cannot be loaded then the existing one is
    left unchanged.

    @param font New font name (full path).
    @return True if the font was loaded successfully - false otherwise.
**/
bool GlFont::ChangeFont( const char* font )
{
    Ft::Face* newFont = new Ft::Face( m_freeType, font );

    if ( newFont->LoadSuccessful() )
    {
        // Must clear cache first because Glyphs reference objects inside
        // the loaded font face.
        ClearCache();
        delete m_fontFace;
        m_fontFace = newFont;
        m_fontFace->SetPixelSizes( m_sizePx, m_sizePx );
        return true;
    }
    else
    {
        delete newFont;
        return false;
    }
}

/**
    Sets the font size in pixels.

    This is the em-square size, so actual glyph sizes vary depending on the
    exact face being used: this is simply how fonts work, no way around it.

    This is a non-trivial operation and invalidates the whole glyph cache.
    If you want to render the same font at different sizes then create two
    GlFont objects and set different pixels sizes for each, or pick a size
    in between and scale up and down using OpenGL.

    If your graphics hardware does not support non-power of two textures
    then (size x 16) must be a power of two (as this will be the size
    of the texture tableau).
**/
void GlFont::SetSizePx( int size )
{
    ClearCache();
    m_sizePx = size;
    m_fontFace->SetPixelSizes( m_sizePx, m_sizePx );

    CreateGlTableau();
};

/**
    Retreive a glyph from the cache.

    If it is not in the cache it is added and then retrieved;
**/
Ft::Glyph& GlFont::GetGlyph( unsigned char c )
{
    unsigned int index = c;
    if ( m_cache[index] == 0 )
    {
        CacheGlyph(c);
    }

    Ft::Glyph& glyph = *m_cache[index];
    return glyph;
}

/**
    Render a string at the origin using OpenGL.
    You will need to modify the model view matrix to position the text.
**/
void GlFont::RenderString( const char* pStr, bool useKerning )
{
    assert( m_fontFace );

    StartGlTexture();

    // OpenGL setup
    glPushMatrix();

    // set the face size and get required metrics
    useKerning = m_fontFace->HasKerning() && useKerning;
    float lineHeight = GetLineHeightPx();

    bool in = false;
    // loop over the character string rendering as we go
    char nextc = *pStr;
    char c;
    FT_Vector advance;
    FT_Vector kerning;

    while ( nextc != '\0' )
    {
        c = nextc;
        nextc = *(++pStr);

        if ( c == '"' && !in )
        {
            glPushAttrib( GL_CURRENT_BIT );
            glColor3f(1,0,0);
        }

        if ( c == '\r' || c=='\n' )
        {
            glPopMatrix();
            glTranslatef( 0.f, -lineHeight, 0.f );
            glPushMatrix();
        }
        else
        {
            Ft::Glyph& glyph = GetGlyph( c );
            glCallList( m_listIndex + c );
            advance = glyph.GetAdvance();
            glTranslatef( advance.x/65536.f, advance.y/65536.f, 0 );

            // Kerning:
            if ( useKerning )
            {
                kerning = m_fontFace->GetKerning( m_fontFace->GetCharIndex(c), m_fontFace->GetCharIndex(nextc) );
                glTranslatef( kerning.x/64.f, kerning.y/64.f, 0 );
            }
        }

        if ( c == '"' )
        {
            if (in)
            {
                glPopAttrib();
                in = false;
            }
            else
            {
                in = true;
            }
        }
    }

    glPopMatrix();
    StopGlTexture();
}

/**
    Render a text buffer of specified size. Does not stop for any character ( e.g. terminator '\0').

    @param pStr Pointer to text.
    @param numChars Number of characters to render.
    @param useKerning If true then adjust spacing between letters for improved aesthetics.
**/
void GlFont::RenderText( const char* pStr, size_t numChars, bool useKerning )
{
    assert( m_fontFace );

    StartGlTexture();

    // OpenGL setup
    glPushMatrix();

    // set the face size and get required metrics
    useKerning = m_fontFace->HasKerning() && useKerning;
    float lineHeight = GetLineHeightPx();

    bool in = false;
    // loop over the character string rendering as we go
    const char* const pEnd = pStr + numChars;
    char nextc = *pStr;
    char c;
    FT_Vector advance;
    FT_Vector kerning;
    while ( pStr != pEnd )
    {
        c = nextc;
        nextc = *(++pStr);

        if ( c == '"' && !in )
        {
            glPushAttrib( GL_CURRENT_BIT );
            glColor3f(1,0,0);
        }

        if ( c == '\r' || c=='\n' )
        {
            glPopMatrix();
            glTranslatef( 0.f, -lineHeight, 0.f );
            glPushMatrix();
        }
        else
        {
            Ft::Glyph& glyph = GetGlyph( c );
            glCallList( m_listIndex + c );
            advance = glyph.GetAdvance();
            glTranslatef( advance.x/65536.f, advance.y/65536.f, 0 );

            // Kerning:
            if ( useKerning )
            {
                kerning = m_fontFace->GetKerning( m_fontFace->GetCharIndex(c), m_fontFace->GetCharIndex(nextc) );
                glTranslatef( kerning.x/64.f, kerning.y/64.f, 0 );
            }

        }

        if ( c == '"' )
        {
            if ( in )
            {
                glPopAttrib();
                in = false;
            }
            else
            {
                in = true;
            }
        }
    }

    glPopMatrix();
    StopGlTexture();
}

/**
    Renders tex but don't apply new-lines and carriage returns -
    instead wrap after every 'cols' characters. Also colours strings.
**/
void GlFont::RenderConsole( const char* pStr, size_t cols, size_t rows, const char* const start, const char* const split, bool useKerning )
{
    assert( m_fontFace );

    StartGlTexture();

    // OpenGL setup
    glPushMatrix();

    // Set the face size and get required metrics
    useKerning = m_fontFace->HasKerning() && useKerning;
    float lineHeight = GetLineHeightPx();

    unsigned int count = cols;
    bool in = false;
    // loop over the character string rendering as we go
    const char* pEnd = pStr + cols*rows;

    if ( pEnd >= split )
    {
        pEnd -= split-start;
    }

    char nextc = *pStr;
    char c;
    FT_Vector advance;
    FT_Vector kerning;
    while ( pStr != pEnd )
    {
        if ( ++pStr >= split )
        {
            pStr = start;
        }

        c = nextc;
        nextc = *pStr;

        if ( count-- == 0 )
        {
            glPopMatrix();
            glTranslatef( 0.f, -lineHeight, 0.f );
            glPushMatrix();
            count = cols - 1;
        }

        if ( c == '\n' )
        {
            pStr += count;
            nextc = *pStr;
            count = cols;
            glPopMatrix();
            glTranslatef( 0.f, -lineHeight, 0.f );
            glPushMatrix();
        }
        else
        {

            if ( c == '"' && !in )
            {
                glPushAttrib( GL_CURRENT_BIT );
                glColor3f( 1, 0, 0 );
            }

            Ft::Glyph& glyph = GetGlyph( c );
            glCallList( m_listIndex + c );
            advance = glyph.GetAdvance();
            glTranslatef( advance.x/65536.f, advance.y/65536.f, 0 );

            // Kerning:
            if ( useKerning )
            {
                kerning = m_fontFace->GetKerning( m_fontFace->GetCharIndex(c), m_fontFace->GetCharIndex(nextc) );
                glTranslatef( kerning.x/64.f, kerning.y/64.f, 0 );
            }

            if ( c == '"' )
            {
                if ( in )
                {
                    glPopAttrib();
                    in = false;
                }
                else
                {
                    in = true;
                }
            }
        }

    }

    glPopMatrix();
    StopGlTexture();
}

/**
    Return the offset that the character at index 'i' would be located if the string were rendered using RenderText().

    Index must be within the bounds of the string.
**/
float GlFont::GetOffsetPx(  const char* pStr, unsigned int i, bool useKerning )
{
    assert( m_fontFace );
    float x = 0.f;
    float y = 0.f;

    // Set the face size and get required metrics
    useKerning = m_fontFace->HasKerning() && useKerning;
    float lineHeight = GetLineHeightPx();

    const char* const pEnd = pStr + i;
    char nextc = *pStr;
    char c;
    FT_Vector advance;
    FT_Vector kerning;
    while ( pStr != pEnd )
    {
        c = nextc;
        nextc = *(++pStr);

        if ( c == '\r' || c=='\n' )
        {
            x = 0.f;
            y -= lineHeight;
        }
        else
        {
            Ft::Glyph& glyph = GetGlyph( c );
            advance = glyph.GetAdvance();
            x += advance.x/65536.f;
            y += advance.y/65536.f;

            // Kerning:
            if ( useKerning )
            {
                kerning = m_fontFace->GetKerning( m_fontFace->GetCharIndex(c), m_fontFace->GetCharIndex(nextc) );
                x += kerning.x/64.f;
                y += kerning.y/64.f;
            }

        }

    }

    return x;
}

/**
    Render a glyph as a textured quad.
    Translates the model view so that the character is rendered on the baseline.
    Restores model-view transform on exit.

    This is used in a display list so no calls which are disallowed in display lists
    should be added here.

    @param glyph Reference to the glyph object that is to be drawn.
    @param sizePX The pixel size which the glyph has been setup to use.
**/
void GlFont::RenderGlyph( const Ft::Glyph& glyph, int sizePx, unsigned int index )
{
    glPushMatrix();

    int top = glyph.GetTop();
    int left = glyph.GetLeft();
    FT_BBox bbox = glyph.GetControlBox();

    // Translate so that character is shifted by left bearing and aligned to baseline
    glTranslatef( static_cast<GLfloat>(left), static_cast<GLfloat>(top-sizePx), 0 );

    int w = bbox.xMax - bbox.xMin;
    int h = bbox.yMax - bbox.yMin;
    float tx = static_cast<float>(w);
    float ty = static_cast<float>(h);
    tx *= 1.f / (16.f*sizePx);
    ty *= 1.f / (16.f*sizePx);

    // Compute the top corner of the glyph's square in tableau:
    int sq_x = (index % 16);
    int sq_y = (index / 16);
    
    // Compute in floating point tex coords:
    float x = sq_x / 16.f;
    float y = sq_y / 16.f;

    glBegin( GL_QUADS );
    {
        glTexCoord2f( x, y );
        glVertex2i( 0,sizePx );

        glTexCoord2f( x+tx, y );
        glVertex2i( w,sizePx );

        glTexCoord2f( x+tx, y+ty );
        glVertex2i( w,sizePx-h );

        glTexCoord2f( x, y+ty );
        glVertex2i( 0,sizePx-h );
    }
    glEnd();

    glPopMatrix();
}

/**
    Setup OpenGL texture modes ready for text rendering.
**/
void GlFont::StartGlTexture()
{
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glEnable( GL_TEXTURE_2D );
}

/**
    Clean up texture modes that OpenGL used for text rendering.
**/
void GlFont::StopGlTexture()
{
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
    glDisable( GL_BLEND );
    glDisable(GL_TEXTURE_2D);
}

/**
    Creates the tableau texture in opengl and fills it black.
**/
void GlFont::CreateGlTableau()
{
    glBindTexture( GL_TEXTURE_2D, m_tabTex );

    // set critical texture parameters
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

    unsigned char* blank = new unsigned char[16*16*m_sizePx*m_sizePx];
    memset( blank, 0, 16*16*m_sizePx*m_sizePx );
    glTexImage2D( GL_TEXTURE_2D, 0, GL_ALPHA, 16*m_sizePx, 16*m_sizePx, 0, GL_ALPHA, GL_UNSIGNED_BYTE, blank );
    delete [] blank;
}

/**
    Initialise all glyph and bitmap cache pointers to null.
**/
void GlFont::InitCache()
{
    m_cache = new Ft::Glyph*[256];

    for ( unsigned int i=0; i<256; ++i )
    {
        m_cache[i] = 0;
    }

    m_listIndex = glGenLists( 256 ); // each glyph will have its own cached display list.
    assert( m_listIndex != 0 );

    // Generate the tableau texture id:
    glGenTextures( 1, &m_tabTex );

    CreateGlTableau();
}

/**
    Erase all the cached bitmaps for the current font face.
**/
void GlFont::ClearCache()
{
    for (unsigned int i=0;i<256;++i)
    {
        delete m_cache[i];
        m_cache[i] = 0;
    }
}

/**
    Make sure all glyphs are pre-loaded and rendered into the cache.
**/
void GlFont::FillCache()
{
    for (unsigned int c=0; c<256; ++c )
    {
        if ( m_cache[c] == 0 )
        {
            CacheGlyph(c);
        }
    }
}

/**
    Cache the glyph to speed rendering.

    Caches the glyph by loading and rendering it.
**/
void GlFont::CacheGlyph( unsigned int i )
{
    m_fontFace->LoadChar( (unsigned char)i );
    m_cache[i] = new Ft::Glyph( *m_fontFace );
    m_cache[i]->Render();

    Ft::Glyph& glyph = *m_cache[i];
    int width  = glyph.GetWidth();
    int height = glyph.GetRows();
    assert( glyph.GetWidth() == glyph.GetPitch() );

    // There is no reason why the glyph size cannot be larger than the specified pixel size.
    // Fortunately for most fonts it is only a single pixel larger which isn't usually noticable,
    // even at small pixel sizes. Ignoring the extra bitmap row/column greatly simplifies
    // rendering so live with it for now.
    width  = width  > m_sizePx ? m_sizePx : width;
    height = height > m_sizePx ? m_sizePx : height;

    // Extract the bitmap so we can cache it:
    unsigned char* pImg = new unsigned char[ width*height ];
    glyph.ExtractBitmap( pImg, width, height, width );

    // Upload glyph's bitmap to the tableau:
    glBindTexture( GL_TEXTURE_2D, m_tabTex );
    glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
    int x = (i % 16) * m_sizePx; // x coord in tableau
    int y = (i / 16) * m_sizePx; // y coord in tableau
    glTexSubImage2D( GL_TEXTURE_2D, 0, x, y, width, height, GL_ALPHA, GL_UNSIGNED_BYTE, pImg );
    glPixelStorei( GL_UNPACK_ALIGNMENT, 4 ); // put alignment back to default

    // Now create the display list for this glyph:
    glNewList( m_listIndex+i, GL_COMPILE );
    glBindTexture( GL_TEXTURE_2D, m_tabTex );
    GlFont::RenderGlyph( glyph, m_sizePx, i );
    glEndList();

    delete [] pImg;
}

} // end namespace GLK
