#include "WindowBase.h"

#include "KeyboardHandler.h"
#include "MouseHandler.h"

#include "../glkcore/thread/MutexLock.h"
#include "../glkcore/containers/ListIterator.h"

namespace GLK
{

WindowBase::WindowBase( int w, int h, int flags, const GlConfig* const glConfig )
:
    m_width  ( w ),
    m_height ( h ),
    m_flags  ( flags ),
    m_eventLoopMutex    ( Mutex::Recursive ),
    m_processingEvents  ( false )
{
    if ( glConfig )
    {
        m_glConfig = *glConfig; // take a copy of the requested OpenGL framebuffer config
    }
}

WindowBase::~WindowBase()
{
}

/**
    Return the width in pixels of the window's client (drawable area excluding window boarders and decorations).
**/
int WindowBase::GetWidth() const
{
    return m_width;
}

/**
    Return the height in pixels of the window's client (drawable area excluding window boarders and decorations).
**/
int WindowBase::GetHeight() const
{
    return m_height;
}

/**
    Return the GlConfig describing this window's framebuffer.
    
    For asynchronous windows this call will block until the Window's event
    loop is running to ensure it returns the config that was used for the
    framebuffer.
**/
const GlConfig& WindowBase::GetGlConfig() const
{
    WaitForEventLoop();
    return m_glConfig;
}

/**
    Returns true if the window is setup for asynchronous operation.
**/
bool WindowBase::IsAsynchronous() const
{
    return ( m_flags & Async ) != 0;
}

/**
    Returns true if the window is setup to be composited on the desktop.
**/
bool WindowBase::IsComposited() const
{
    return ( m_flags & AlphaComposited ) != 0;
}

/**
    Returns true if the window is in fullscreen mode.
**/
bool WindowBase::IsFullscreen() const
{
    return ( m_flags & FullScreen ) != 0;
}

/**
    Returns true if the window has no decorations (no border or title bar).
**/
bool WindowBase::IsDecorless() const
{
    return ( m_flags & Decorless ) != 0;
}

/**
    @return true if the core event loop is running, false otherwise.
**/
bool WindowBase::EventLoopIsRunning() const
{
    return m_processingEvents;
}

/**
    Blocks until this Window's event loop is started.
    Returns immediately if the event loop is running or if the window is not asynchronous.
**/
void WindowBase::WaitForEventLoop() const
{
    if ( IsAsynchronous() )
    {
        MutexLock lock( m_eventLoopMutex );
        while ( !m_processingEvents )
        {
            m_eventLoopCond.Wait( m_eventLoopMutex );
        }
    }
}

/**
    Signal any threads that are waiting for the event lop to start.
**/
void WindowBase::SignalEventLoopStart()
{
    m_eventLoopMutex.Lock();
    m_processingEvents = true;
    m_eventLoopCond.WakeAll();
    m_eventLoopMutex.Unlock();
}

/**
    Add a new handler to the stack.
    
    This new handler becomes the active handler (it will receive
    all keyboard character events).

    It probably an error to push a handler to multiple windows but this
    depends on the implementation of the specified handler.

    @return Pointer to the previously active handler. Null if there were no previous handlers (stack was empty).
**/
KeyboardHandler* WindowBase::PushKeyboardHandler( KeyboardHandler* const newHandler )
{
    KeyboardHandler* oldHandler = 0;

    MutexLock lock( m_kbhMutex );

    if ( m_kbh.Length() )
    {
        oldHandler = m_kbh.GetLast();
    }

    m_kbh.AddBack( newHandler );

    newHandler->Activated();

    return oldHandler;
}

/**
    Remove the current keyboard handler from the stack.
    The previous handler on the stack now becomes the
    active handler.

    Popping when there are no handlers in the stack causes an assertion in debug builds.

    @return Null if the stack was empty, otherwise return a pointer to the removed handler.
**/
KeyboardHandler* WindowBase::PopKeyboardHandler()
{
    KeyboardHandler* oldHandler = 0;

    MutexLock lock( m_kbhMutex );

    if ( m_kbh.Length() )
    {
        oldHandler = m_kbh.RemoveLast();
        oldHandler->Deactivated();
    }
    else
    {
        assert(0); // popping empty stack
    }

    return oldHandler;
}

/**
    Find a specific handler in the stack and remove it - shifting all other
    elements above it in the stack down accordingly.

    It is an error to remove the active handler. (PopKeyboardHandler() should
    be used instead).

    Need to use this if we push an object onto the stack and then delete it.
    Even if a deleted handler is not currently active it could become so if
    it remains in the stack.

    @return The number of times the handler was found in the stack.
**/
unsigned int WindowBase::RemoveKeyboardHandler( KeyboardHandler* const handler )
{
    unsigned int found = 0;

    MutexLock lock( m_kbhMutex );

    KeyboardHandler* active = m_kbh.GetLast();

    if ( handler == active )
    {
        assert(0); // should have used pop instead!
    }
    else
    {
        ListIterator<KeyboardHandler*> itr( m_kbh );
        while ( itr.HasNext() )
        {
            KeyboardHandler* current = itr.Next();
            if ( current == handler )
            {
                itr.Remove();
            }
        }
    }

    return found;
}

/**
    Add a new handler to the stack.
    
    This new handler becomes the active handler (it will receive
    all keyboard character events).

    It probably an error to push a handler to multiple windows but this
    depends on the implementation of the specified handler.

    @return Pointer to the previously active handler. Null if there were no previous handlers (stack was empty).
**/
MouseHandler* WindowBase::PushMouseHandler( MouseHandler* const newHandler )
{
    MouseHandler* oldHandler = 0;

    MutexLock lock( m_mhMutex );

    if ( m_mh.Length() )
    {
        oldHandler = m_mh.GetLast();
    }

    m_mh.AddBack( newHandler );

    newHandler->Activated();

    return oldHandler;
}

/**
    Remove the current mouse handler from the stack.
    The previous handler on the stack now becomes the
    active handler.

    Popping when there are no handlers in the stack causes an assertion in debug builds.

    @return Null if the stack was empty, otherwise return a pointer to the removed handler.
**/
MouseHandler* WindowBase::PopMouseHandler()
{
    MouseHandler* oldHandler = 0;

    MutexLock lock( m_mhMutex );

    if ( m_mh.Length() )
    {
        oldHandler = m_mh.RemoveLast();
        oldHandler->Deactivated();
    }
    else
    {
        assert(0); // popping empty stack
    }

    return oldHandler;
}

/**
    Find a specific handler in the stack and remove it - shifting all other
    elements above it in the stack down accordingly.


    It is an error to remove the active handler. (PopKeyboardHandler() should
    be used instead).

    Need to use this if we push an object onto the stack and then delete it.
    Even if a deleted handler is not currently active it could become so if
    it remains in the stack.

    @return The number of times the handler was found in the stack.
**/
unsigned int WindowBase::RemoveMouseHandler( MouseHandler* const handler )
{
    unsigned int found = 0;

    MutexLock lock( m_mhMutex );

    MouseHandler* active = m_mh.GetLast();

    if ( handler == active )
    {
        assert(0); // should have used pop instead!
    }
    else
    {
        ListIterator<MouseHandler*> itr( m_mh );
        while ( itr.HasNext() )
        {
            MouseHandler* current = itr.Next();

            if ( current == handler )
            {
                itr.Remove();
            }
        }
    }

    return found;
}

} // end namespace GLK

