#ifndef WINDOW_BASE_HEADER_H
#define WINDOW_BASE_HEADER_H

#include "WindowTypes.h"
#include "../glkcore/thread/Mutex.h"
#include "../glkcore/thread/ConditionVariable.h"
#include "../glkcore/containers/List.h"
#include "GlConfig.h"

namespace GLK
{

// fwd decl
class KeyboardHandler;
class MouseHandler;

/**
    This is a base class for Window objects on all platforms.
    
    It defines a consistent interface and also some high-level
    functionality which ensures consistent behaviour regarding
    bring up of windows on different platforms (particularly with
    regards to multi-threading).
**/
class FW_API WindowBase
{

public:
    WindowBase ( int width, int height, int windowType = Normal, const GlConfig* const glConfig = 0 );
    virtual ~WindowBase();

    int GetWidth()  const;
    int GetHeight() const;
    const GlConfig& GetGlConfig() const;
    bool IsAsynchronous() const;
    bool IsComposited()   const;
    bool IsFullscreen()   const;
    bool IsDecorless()    const;
        
    // Mouse and keyboard handler management:
    KeyboardHandler* PushKeyboardHandler( KeyboardHandler* const newHandler );
    KeyboardHandler* PopKeyboardHandler();
    unsigned int RemoveKeyboardHandler( KeyboardHandler* const handler );
    MouseHandler* PushMouseHandler( MouseHandler* const newHandler );
    MouseHandler* PopMouseHandler();
    unsigned int RemoveMouseHandler( MouseHandler* const handler );

    void WaitForEventLoop() const;

    // Window interface
    virtual int  StartEventLoop() = 0;
    virtual void SetWindowName( const char* name ) = 0;
    virtual void SwapBuffers() = 0;
    virtual void Show() = 0;
    virtual void Hide() = 0;
    virtual void Minimise() = 0;
    virtual void Maximise() = 0;
    virtual void Move( int x, int y, int w = 0, int h = 0 ) = 0;
    virtual void Quit( int exitCode ) = 0;
    virtual void Notify() = 0;
    virtual void StartTimer( unsigned int millisecs ) = 0;
    virtual void StopTimer() = 0;

    // OpenGL interface
    virtual bool InitGL()    = 0;
    virtual void DestroyGL() = 0;
    virtual void Resize( unsigned int, unsigned int ) = 0;
    virtual bool Update( unsigned int ) = 0;
    virtual void Render()               = 0;
    virtual void TimerExpired( int id ) = 0;

    virtual void DecorationsOff() = 0;
    virtual void DecorationsOn()  = 0;
    virtual void FullscreenOn()   = 0;
    virtual void FullscreenOff()  = 0;

protected:
    void SignalEventLoopStart();
    bool EventLoopIsRunning() const;
        
    int m_width;
    int m_height;
    int m_flags;
    GlConfig m_glConfig;
        
    // keyboard handling
    Mutex m_kbhMutex;
    List<KeyboardHandler*> m_kbh;

    // mouse handling
    Mutex m_mhMutex;  
    List<MouseHandler*> m_mh;
    
    // signalling - the following mutable members allow the class can remain logically const correct.
    mutable Mutex m_eventLoopMutex;
    mutable ConditionVariable m_eventLoopCond;
    bool m_processingEvents;

private:
};

} // end namespace GLK

#endif // WINDOW_BASE_HEADER_H

