#include "GlWindow.h"

#include "../KeyboardHandler.h"
#include "../MouseHandler.h"

#include "../../glkcore/thread/MutexLock.h"
#include "../../glkcore/containers/ListIterator.h"

#include <assert.h>
#include <windowsx.h>
#include <shellapi.h>
#include <Dwmapi.h>

namespace GLK
{

#define WM_MY_NOTIFY        (WM_USER)
#define WM_CUSTOM_QUIT      (WM_USER+1)
#define WM_FULLSCREEN_ON    (WM_USER+2)
#define WM_FULLSCREEN_OFF   (WM_USER+3)
#define WM_DECORATIONS_ON   (WM_USER+4)
#define WM_DECORATIONS_OFF  (WM_USER+5)

/**
        The typical 'window procedure' for processing the window's messages.
        (The appropriate instance gets called by WndProcRouter).
**/
LONG GlWindow::WndProc( UINT msg, WPARAM wParam, LPARAM lParam )
{
    LONG rval = 0;

    bool state = false; // used to speed processing of mouse button messages

	switch (msg)
	{
	case WM_SYSCOMMAND:
		// Disable system screensaver if in full screen mode
		switch (wParam & 0xFFF0)
		{
		case SC_SCREENSAVE:
			if ( IsFullscreen() )
            {
                return 0;
            }
			else
            {
                PostMessage( m_windowHandle, WM_CLOSE, 0, 0 );
            }
		default:
			return DefWindowProc( m_windowHandle, msg, wParam, lParam );
		}

	case WM_DISPLAYCHANGE:
		if ( IsFullscreen() )
        {
            PostMessage( m_windowHandle, WM_CLOSE, 0, 0 );
        }
		break;

    case WM_CUSTOM_QUIT:
		PostQuitMessage( wParam );
        break;
	case WM_CLOSE:
		PostQuitMessage(0);
		break;

	case WM_ERASEBKGND:
        rval = TRUE; // returning TRUE indicates that we did erase the background
        break;

	case WM_PAINT:
		Render();
		ValidateRect( m_windowHandle, 0 );
		break;

	case WM_SIZE:
        WindowBase::m_width  = LOWORD(lParam);
		WindowBase::m_height = HIWORD(lParam);

        if ( !IsFullscreen() )
        {
            // need this to record width after fullscreen switch
            GlWindow::m_winWidth  = LOWORD(lParam);
            GlWindow::m_winHeight = HIWORD(lParam);
        }

		Resize( LOWORD(lParam), HIWORD(lParam) );
		break;

    case WM_ACTIVATE:
	case WM_ACTIVATEAPP:
		break;

    case WM_TIMER:
        TimerExpired( wParam );
        break;

	case WM_CREATE:
	    m_deviceContext = GetDC( m_windowHandle );
	    m_renderingContext = SetupGl( m_glConfig );
	    if (
				m_deviceContext == 0 ||
				m_renderingContext == 0 ||
				wglMakeCurrent( m_deviceContext, m_renderingContext ) == FALSE
			)
		{

			MessageBox( m_windowHandle ,"Device or Rendering Context setup failed", "Error", MB_OK|MB_ICONERROR );
			if ( m_renderingContext ) wglDeleteContext( m_renderingContext );
			if ( m_deviceContext ) ReleaseDC( m_windowHandle, m_deviceContext );
			PostMessage( m_windowHandle, WM_CLOSE, 0, 0 );
		} else {
			InitGL(); // When the window is first created the base class InitGL is called (which does nothing)
                      // subsequent calls invoke any derived class' InitGL
			m_timer.Reset();
		}
		break;

	case WM_DESTROY:
		DestroyGL(); // When the window is finally destroyed the base class DestroyGL is called (which does nothing)
                     // all previous calls invoke any derived class' DestroyGL

		wglMakeCurrent( 0, 0 );
		wglDeleteContext(m_renderingContext);
		ReleaseDC( m_windowHandle, m_deviceContext );
		break;

    case WM_CHAR:
        {
            MutexLock lock( m_kbhMutex );
            if ( m_kbh.Length() )
            {
                char c = static_cast<char>( wParam );
                m_kbh.GetLast()->Key( c );
            }
        }
        break;

	case WM_KEYDOWN:
        {
            if ( wParam == VK_ESCAPE )
            {
                PostMessage( m_windowHandle, WM_CLOSE, 0, 0 );
            }

            switch ( wParam )
            {
            case VK_LEFT:
                SendNotifyMessage( m_windowHandle, WM_CHAR, '\v', 0 );
                break;
            case VK_RIGHT:
                SendNotifyMessage( m_windowHandle, WM_CHAR, '\f', 0 );
                break;
            case VK_UP:
                SendNotifyMessage( m_windowHandle, WM_CHAR, 28, 0 );
                break;
            case VK_DOWN:
                SendNotifyMessage( m_windowHandle, WM_CHAR, 29, 0 );
                break;
            case VK_DELETE:
                SendNotifyMessage( m_windowHandle, WM_CHAR, 16, 0 );
                break;
            }
        }
        break;

    case WM_LBUTTONDBLCLK:
    case WM_RBUTTONDBLCLK:
    case WM_MBUTTONDBLCLK:
        {
            MutexLock lock( m_mhMutex );
            if ( m_mh.Length() != 0 )
            {
                MouseHandler::ButtonId id = static_cast<MouseHandler::ButtonId>( (msg >> 2) & 3 );
                m_mh.GetLast()->DoubleClick( id, GET_X_LPARAM( lParam ), GET_Y_LPARAM( lParam ) );
            }
        }
        break;

    case WM_LBUTTONDOWN:
    case WM_RBUTTONDOWN:
    case WM_MBUTTONDOWN:
        state = true;
    case WM_LBUTTONUP:
    case WM_RBUTTONUP:
    case WM_MBUTTONUP:
        {
            MutexLock lock( m_mhMutex );
            if ( m_mh.Length() != 0 )
            {
                MouseHandler::ButtonId id = MouseHandler::Left;
                if ( msg == WM_RBUTTONDOWN || msg == WM_RBUTTONUP )
                {
                    id = MouseHandler::Right;
                }
                else if ( msg == WM_MBUTTONDOWN || msg == WM_MBUTTONUP )
                {
                    id = MouseHandler::Middle;
                }

                m_mh.GetLast()->ButtonEvent( id, state );
            }
        }
        break;


    case WM_MOUSEWHEEL:
        {
            MutexLock lock( m_mhMutex );
            if ( m_mh.Length() != 0 )
            {
                m_mh.GetLast()->Scroll( GET_WHEEL_DELTA_WPARAM(wParam) / WHEEL_DELTA );
            }
        }
        break;

	case WM_MOUSEMOVE:
        {
            MutexLock lock( m_mhMutex );
            if ( m_mh.Length() != 0 )
            {
                m_mh.GetLast()->NewPosition( GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) );
            }
        }
		break;

    case WM_MOVE:
        // window moved
		m_posX = GET_X_LPARAM(lParam);
		m_posY = GET_Y_LPARAM(lParam);
		break;

    case WM_DROPFILES:
        {
            UINT size = 1 + DragQueryFile( (HDROP)wParam, 0, 0, 0 );
            char* tmp = new char[size];
            DragQueryFile( (HDROP)wParam, 0, tmp, size );

            MutexLock lock( m_mhMutex );
            if ( m_mh.Length() != 0 )
            {
                m_mh.GetLast()->FileDrop( tmp  );
            }

            delete [] tmp;
            DragFinish( (HDROP)wParam );
        }
        break;

    case WM_FULLSCREEN_ON:
        m_flags ^= FullScreen;
        ReInit();
        break;
    case WM_FULLSCREEN_OFF:
        m_flags &= ~FullScreen;
        m_desiredWidth  = m_winWidth;
        m_desiredHeight = m_winHeight;
        ReInit();
        break;
    case WM_DECORATIONS_ON:
        m_flags &= ~Decorless;
        ReInit();
        break;
    case WM_DECORATIONS_OFF:
        m_flags ^= Decorless;
        ReInit();
        break;

	default:
		rval = DefWindowProc( m_windowHandle, msg, wParam, lParam );
        break;
    }

    return rval;
}

/**
    Construct an OpenGL window.

    A small amount of control over window style is provided by the windowType parameter.
    The effect of the bitmask's flags are platform dependent but in genral:
    Normal      - The default. Would typically be used for an application's main window.
    Slim        - Minimal window with slimmer borders. May not be resizeable. May not appear in task-bar/dock etc. Useful for toolbars etc.
    Fullscreen  - Window's client area (OpenGL rendered area) occupies the whole screen. The window has no borders/title bar. A typical use would be for games.

    @param w The width of the windows client area (area available for OpenGL drawing).
    @param h The height of the windows client area (area available for OpenGL drawing).
    @param windowType Bitmask, created from WindowType enums, that controls window appearance.
    @param delayInit Delays initialisation of the window until StartEventLoop is called - designed exclusively for use by GlThreadedWindow.
**/
GlWindow::GlWindow( unsigned int w, unsigned int h, int flags, const GlConfig* const glConfig )
:
    WindowBase          ( w, h, flags, glConfig ),
    m_windowHandle	    (0),
    m_deviceContext     (0),
    m_renderingContext  (0),

    m_windowStyle       (0),
    m_extendedStyle     (0),

    m_posX              (0),
    m_posY              (0),

    // Initialise client and window dimension to same since we dont know any different at this point.
    m_desiredWidth      (w),
    m_desiredHeight     (h),
    m_winWidth          (w),
    m_winHeight         (h)
{
    if ( !IsAsynchronous() )
    {
        Init();
    }
}

GlWindow::~GlWindow()
{
    if ( !IsAsynchronous() )
    {
        Destroy();
    }
}

/**
    Setup a new window.

    Init is called in the constructor of GlWindow.
    The window creation system causes virtual function
    calls in the window's message handler - effectively
    virtual functions (InitGL and DestroyGL) are called
    in the constructor. The code is designed so that
    these have no effect - do not modify the base InitGL
    or DestroyGL functions (they should do nothing).

    The first and final calls to InitGL and DestroyGL
    respectively are actually performed in StartEventLoop(),
    with the dud calls to the base versions having no effect.
    Hence, at the high level it appears that the derived
    versions get called at all appropriate times.
**/
bool GlWindow::Init()
{
    char appName[] = "GlWindow";
    HINSTANCE hInstance = GetModuleHandle(0);

    if ( m_flags & FullScreen )
    {
        RECT area;
        SystemParametersInfo( SPI_GETWORKAREA, 0, &area, 0 );
        m_width = area.right;
        m_height = area.bottom;
        m_desiredWidth  = m_width;
        m_desiredHeight = m_height;
        m_posX	 = 0;
        m_posY   = 0;
        m_windowStyle = WS_POPUP;
    }
    else
    {
        if ( m_flags & Decorless)
        {
            // borderless, titleless window (which can still be moved)
            m_windowStyle   = WS_POPUP | WS_SYSMENU;
            m_extendedStyle = 0;
        }
        else if ( m_flags & Slim )
        {
            // Window with slim 'toolbox-style' borders
            m_windowStyle   = WS_SYSMENU | WS_CAPTION;
            m_extendedStyle = WS_EX_TOOLWINDOW;
        }
        else 
        {
            // A regular window
            m_windowStyle   = WS_OVERLAPPEDWINDOW;
            m_extendedStyle = 0;
        }
    }

    // These are required styles for OpenGL windows:
    m_windowStyle |= WS_CLIPCHILDREN;
    m_windowStyle |= WS_CLIPSIBLINGS;

    // Describe the window and register it
    m_wc.cbSize		   = sizeof( WNDCLASSEX );
    m_wc.style         = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
    m_wc.lpfnWndProc   = (WNDPROC)GlWindow::WndProcRouter;
    m_wc.cbClsExtra    = 0;
    m_wc.cbWndExtra    = 0;
    m_wc.hInstance     = hInstance;
    m_wc.hIcon         = 0; // LoadIcon(hInstance, MAKEINTRESOURCE(APP_ICON));
    m_wc.hIconSm       = 0;
    m_wc.hCursor       = LoadCursor(NULL, IDC_CROSS);
    m_wc.hbrBackground = 0;
    m_wc.lpszMenuName  = 0;
    m_wc.lpszClassName = appName;

    // In multi-window scenarios the following call 'fails' (returns 0)
    // because the class will have already been registered.
    // This is ok - the class is already registered so we can just use it
    // (so we ignore the return value).
    RegisterClassEx( &m_wc );

    HWND hWnd = CreateWindowEx(
                                m_extendedStyle,            // extended style
                                m_wc.lpszClassName,         // class name
                                0,                          // we give title bar no text by default
                                m_windowStyle, // window style
				m_posX, m_posY, m_desiredWidth, m_desiredHeight,
				0,                          // parent window
				0,                          // window class menu.
				m_wc.hInstance,             // owner of this window
				(LPVOID)this                // lpParams (extra arguments)
		      );

    assert( hWnd == m_windowHandle ); // NCCREATE message should have caused m_windowHandle to be set correctly in WndProcRouter

    // If window could not be created give error message
    if ( m_windowHandle == 0 )
    {
        MessageBox( 0, "A window could not be created", "Error", MB_OK|MB_ICONERROR );
    }
    else
    {
        AdjustSizeForWin32();
    }

    DragAcceptFiles( hWnd, TRUE );

// Composting only supported since Vista.
#ifdef NTDDI_VISTA
	if ( m_flags & AlphaComposited )
	{
		// If it is supported this is how we enable it:
		DWM_BLURBEHIND bb = {0};
		bb.dwFlags = DWM_BB_ENABLE;
		bb.fEnable = true;
		bb.hRgnBlur = NULL;
		HRESULT hr = DwmEnableBlurBehindWindow( hWnd, &bb );
	}
#endif

    return true;
}

/**
    Destroy the window and all resources.

    Unregisters the window from the Win32 subsystem and
    sets displays settings back to their default.
**/
void GlWindow::Destroy()
{
    if ( m_windowHandle )
    {
        char appName[] = "GlWindow";

        DestroyWindow( m_windowHandle );

        // :FIXME: dodgy? all windows use same window class so if it is deleted in one thread
        // a window in another thread may stop working (not sure)
        // should really instance count and unregister for last window
        UnregisterClass( appName, GetModuleHandle(0) );

        m_windowHandle = 0;
    }
}

/**
    Tell the windowing system to make the window visible.
**/
void GlWindow::Show()
{
    if ( m_windowHandle )
    {
        ShowWindowAsync( m_windowHandle, SW_SHOWNORMAL );
    }
}

/**
	Make the window not visible.
**/
void GlWindow::Hide()
{
    if ( m_windowHandle )
    {
        ShowWindowAsync( m_windowHandle, SW_HIDE );
    }
}

/**
    Tell the windowing system to minimise the window.
**/
void GlWindow::Minimise()
{
    if ( m_windowHandle )
    {
        ShowWindowAsync( m_windowHandle, SW_SHOWMINIMIZED );
    }
}

/**
	Tell the windowing system to maximise the window.
**/
void GlWindow::Maximise()
{
    if ( m_windowHandle )
    {
        ShowWindowAsync( m_windowHandle, SW_SHOWMAXIMIZED );
    }
}

/**
    Move and (optionally) resize window.

    Move requests are ignored if the window is in fullscreen mode.
**/
void GlWindow::Move( int x, int y, int w, int h )
{
    if ( w==0 || h==0 )
    {
        w = m_desiredWidth;
        h = m_desiredHeight;
    }
    else
    {
        m_desiredWidth = w;
        m_desiredHeight = h;
    }

    WaitForEventLoop();
    if ( m_windowHandle && !IsFullscreen() )
    {
        MoveWindow( m_windowHandle, x, y, w, h, TRUE );
		AdjustSizeForWin32();
    }
}

/**
    Asks the window to return from its message loop.

    @param exitCode to return from message loop.
**/
void GlWindow::Quit( int exitCode )
{
    // loop to ensure message gets posted (if queue is full it could be ignored otherwise).
    while ( PostMessage( m_windowHandle, WM_CUSTOM_QUIT, exitCode, 0 ) == 0 )
    {
        Thread::Sleep( 5 );
    }
}

/**
    Notifies the window that it should call Update().

    Effectively this can be used to wake the Window when
    it waiting for activity. (The window will wait in this
    way if Update() returns false).
**/
void GlWindow::Notify()
{
    // Nothing is (or should be) done in response to the WM_USER/WM_MY_NOTIFY message.
    // It is just intended to cause the window to wake up.
    PostMessage( m_windowHandle, WM_MY_NOTIFY, 0, 0 );
}

/**
    Start a timer which causes the window to update at a specified interval.

    @param millisecs The desired time interval (approx.) between calls to Update().
**/
void GlWindow::StartTimer( unsigned int millisecs )
{
    SetTimer( m_windowHandle, 1, millisecs, 0 );
}

/**
    Stop the timer.
**/
void GlWindow::StopTimer()
{
    KillTimer( m_windowHandle, 1 );
}

/**
    Starts the window's message loop.

    @return Exit code.
**/
int GlWindow::StartEventLoop()
{
    int exitCode;

    if ( IsAsynchronous() )
    {
        Init();
        Show();
    }

	if ( m_windowHandle )
	{
        bool initOk = InitGL();
        if ( initOk )
        {
            SignalEventLoopStart();
	
		    // Enter message loop
		    MSG msg;
		    do {
                if ( Update( m_timer.GetMilliSeconds() ) )
			    {
			        m_timer.Reset();

				    // Cause a repaint
				    InvalidateRect( m_windowHandle, 0, FALSE );
				    UpdateWindow( m_windowHandle );
                }
                else
                {
                    // Suspend this thread until there is new window activity:
                    WaitMessage();
                }

			    while ( PeekMessage( &msg, 0, 0, 0, PM_REMOVE ) != 0 && msg.message != WM_QUIT )
			    {
				    TranslateMessage( &msg );
				    DispatchMessage( &msg );
			    }

		    } while( msg.message != WM_QUIT );

            exitCode = msg.wParam;
        }
        else
        {
            exitCode = -2;
        }

        m_eventLoopMutex.Lock();
        m_processingEvents = false;
        m_eventLoopMutex.Unlock();

        DestroyGL();
    }
    else
    {
        exitCode = -1;
    }
    return exitCode;
}

/**
	The width and height used by windows includes the
	window decorations (title bar and border) whereas
	the API requires width and height to define the
	client area. The following compensates for this:
**/
void GlWindow::AdjustSizeForWin32()
{
    RECT windowRect;
    RECT clientRect;
    GetWindowRect( m_windowHandle, &windowRect ); // including border
    GetClientRect( m_windowHandle, &clientRect ); // excluding border

    LONG dw = m_desiredWidth  - (clientRect.right - clientRect.left); // desired minus actual
    LONG dh = m_desiredHeight - (clientRect.bottom - clientRect.top); // desired minus actual

    LONG w = windowRect.right - windowRect.left;
    LONG h = windowRect.bottom - windowRect.top;

    MoveWindow( m_windowHandle, windowRect.left, windowRect.top, w + dw, h + dh, TRUE );
}

/**
    Set/change the text show in the window's title bar.

    In asynchronous windows (GlThreadWindow) waits for the event
    loop mutex to become free, so event loop must have started
    or be able to start while this calling thread is blocking.
    The following will deadlock:

    window.SetWindowName(...); // <- deadlock here.
    window.StartEventLoopAsynchronous();
**/
void GlWindow::SetWindowName( const char* name )
{
    WaitForEventLoop();
    if ( m_windowHandle && name )
    {
        m_windowName = String( name );
        SetWindowText( m_windowHandle, m_windowName.cStr() );
    }
}

void GlWindow::DecorationsOff()
{
    if ( !IsDecorless() )
    {
        PostMessage( m_windowHandle, WM_DECORATIONS_OFF, 0, 0 );
    }
}

/**
    Request that decorations are removed from the window.

    For asynchronous windows this call may block until the
    async message queue has space.
**/
void GlWindow::DecorationsOn()
{
    if ( IsDecorless() )
    {
        PostMessage( m_windowHandle, WM_DECORATIONS_ON, 0, 0 );
    }
}

/**
    Switch to fullscreen mode.
    
    This works for single threaded windows only, GlThreadedWindow has its own derived implementation.
**/
void GlWindow::FullscreenOn()
{
    if ( !IsFullscreen() )
    {
        PostMessage( m_windowHandle, WM_FULLSCREEN_ON, 0, 0 );
    }
}

/**
    Switch back to windowed mode if in fullscreen mode.
    
    This works for single threaded windows only, GlThreadedWindow has its own derived implementation.
**/
void GlWindow::FullscreenOff()
{
    if ( IsFullscreen() )
    {
        PostMessage( m_windowHandle, WM_FULLSCREEN_OFF, 0, 0 );
    }
}

/**
	Static function to set the windows desktop display mode.
**/
LONG GlWindow::SetDisplay( DWORD width, DWORD height, DWORD depth )
{
    DEVMODE settings;
    settings.dmSize	= sizeof( DEVMODE );

    int i = 0;
    while ( EnumDisplaySettings( NULL, i++, &settings ) )
    {
        if (
            settings.dmPelsWidth  == width &&
            settings.dmPelsHeight == height &&
            settings.dmBitsPerPel == depth
           )
        {
            return ChangeDisplaySettings( &settings, CDS_RESET | CDS_FULLSCREEN );
        }
    }

    return DISP_CHANGE_BADMODE;
}

/**
	Static function to create an OpenGL rendering context.

    @param config Specifies both the desired configuration and returns
    the actaul configuration of the window's pixel format.
**/
HGLRC GlWindow::SetupGl( GlConfig& config )
{
    DWORD drawingFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
    if ( IsComposited() )
    {
        drawingFlags |= PFD_SUPPORT_COMPOSITION; // Not needed anymore it seems (on Windows 7 at least) - leave in for now
        if ( config.GetAlphaSize() == 0 ) 
        {		
            config.SetAlphaSize(8); // Must have alpha channel to for window compositing
        }
    }

    if ( config.IsDoubleBuffered() )
    {
        drawingFlags |= PFD_DOUBLEBUFFER;
    }

    if ( config.GetDepthSize() == 0 )
    {
        drawingFlags |= PFD_DEPTH_DONTCARE;
    }

	PIXELFORMATDESCRIPTOR pfd =
    {
		sizeof (PIXELFORMATDESCRIPTOR), // size of structure
		1,                              // structure version
        drawingFlags,
		PFD_TYPE_RGBA,		// pixel type
		32,                 // pixel depth
		static_cast<BYTE>(config.GetRedSize()),   0,  // red bits and shifts
        static_cast<BYTE>(config.GetGreenSize()), 0,  // green bits and shifts
        static_cast<BYTE>(config.GetBlueSize()),  0,  // blue bits and shifts
		static_cast<BYTE>(config.GetAlphaSize()), 0,  // alpha bits and shift
		0,					// no accumulation buffer
		0, 0, 0, 0,			// accum bits (ignored)
		static_cast<BYTE>(config.GetDepthSize()),   // depth buffer
		static_cast<BYTE>(config.GetStencilSize()), // stencil buffer
		0,					// auxiliary buffers
		PFD_MAIN_PLANE,		// main layer
		0,					// reserved
		0, 0, 0,			// layer, visible, damage masks
	};

    int pixelformatID = ChoosePixelFormat( m_deviceContext, &pfd );

    /* If no pixel suitable pixel format is available error message and return NULL */
    if ( !pixelformatID )
    {
        MessageBox( WindowFromDC(m_deviceContext), "OpenGL Pixel Format could not be created", "Error", MB_OK | MB_ICONERROR );
        return 0;
    }

    SetPixelFormat( m_deviceContext, pixelformatID, &pfd );

    // read back the exact pixel format that was set:
    DescribePixelFormat( m_deviceContext, pixelformatID, sizeof(PIXELFORMATDESCRIPTOR), &pfd );
    config.SetRedSize( pfd.cRedBits );
    config.SetGreenSize( pfd.cGreenBits );
    config.SetBlueSize( pfd.cBlueBits );
    config.SetAlphaSize( pfd.cAlphaBits );
    config.SetDepthSize( pfd.cDepthBits );
    config.SetStencilSize( pfd.cStencilBits );
    config.SetDoubleBuffered( pfd.dwFlags & PFD_DOUBLEBUFFER );

    return wglCreateContext(m_deviceContext);
}

/**
    Destroys the window and then reinitialises it
    by calling GlWindow::Init().

    @todo  will this cause InitGL to get called again?
**/
void GlWindow::ReInit()
{
    DestroyWindow( m_windowHandle );
    bool ok = Init();
    assert( ok );
    SetWindowText( m_windowHandle, m_windowName.cStr() );
    ShowWindow( m_windowHandle, SW_SHOW );
}

/**
	Routes window messages to the correct GlWindow instance.

	Every window in Win32 must have a static WndProc for receiving window messages.
	In order to abstract this awful mechanism away, we register the same static
	WndProc function (GlWindow::WndProcRouter) for every class. This function
	then looks up the actual GlWindow to which the message should be forwarded.
**/
LONG WINAPI GlWindow::WndProcRouter( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
	GlWindow* window;

	if ( msg == WM_NCCREATE )
	{
		// store the GlWindow's address (passed to CreateWindowEx) as its unique id
		window = reinterpret_cast<GlWindow*>( ((LPCREATESTRUCT)lParam)->lpCreateParams );
		::SetWindowLongPtr( hWnd, GWL_USERDATA, reinterpret_cast<LONG>(window) );

		window->m_windowHandle = hWnd; // store the window handle for the newly created window
	}
	else
	{
		window = reinterpret_cast<GlWindow*>( ::GetWindowLong( hWnd, GWL_USERDATA ) ); // get pointer to the approriate GlWindow
	}

	// forward the message
	if ( window != 0 )
	{
		return window->WndProc( msg, wParam, lParam );
	}
	else
	{
		return DefWindowProc( hWnd, msg, wParam, lParam );
	}
}

} // End of namespace GLK

