#ifndef GL_WIN32_WINDOW_H
#define GL_WIN32_WINDOW_H

#include "../../glkcore/fwconfig.h"
#include "../../glkcore/time/Timer.h"
#include "../../glkcore/thread/Mutex.h"
#include "../../glkcore/utility/GlobalDefs.h"
#include "KeyEnums.h"
#include "../../glkcore/containers/List.h"
#include "../Windowbase.h"
#include "../../glkcore/utility/String.h"

#include <stdio.h>

#define WIN32_MEAN_AND_LEAN
#include <windows.h>
#include <mmsystem.h>
#include <GL/gl.h>

namespace GLK
{

// fwd decl
class KeyboardHandler;
class MouseHandler;

/**
   Facilitates OpenGL applications on Win32.

   Notes:

   The value returned from Update is important.
   If true then you are indicating that the window needs to be redrawn (i.e. it has been 'updated')
   If false, not only is the window not redrawn, but will suspend until the window receives a new message
   ( i.e. the thread waits for window activity ). You can force it to wake by calling Notify() which
   will send the event loop an empty message.
   
   If update always returns true then the rendering loop will execute as fast as possible but will
   consume 100% of cpu (or 100% of one core of CPU). Whether this is desirable depends on the application.
**/
class FW_API GlWindow : public WindowBase
{
public:
    GlWindow(unsigned int w, unsigned int h, int flags = Normal, const GlConfig* const glConfig = 0 );
    virtual ~GlWindow();

    void Show();
    void Hide();
    void Minimise();
    void Maximise();

    void Move( int x, int y, int w=0, int h=0 );
    void Quit( int exitCode );
    void Notify();
    void StartTimer( unsigned int millisecs );
    void StopTimer();

    int StartEventLoop();
    void SetWindowName(const char* name);
    void SwapBuffers() { ::SwapBuffers( m_deviceContext ); };

    virtual bool InitGL() { return true; };
    virtual void DestroyGL() {};
    virtual void Resize( unsigned int, unsigned int ) {};
    virtual bool Update( unsigned int ) { return true; };
    virtual void Render() {};
    virtual void TimerExpired( int ) {};

    virtual void DecorationsOff();
    virtual void DecorationsOn();
    virtual void FullscreenOn();
    virtual void FullscreenOff();

protected:
    bool EventLoopIsRunning() const;

private:
    SET_UNCOPYABLE(GlWindow)

    Timer   m_timer;
    HWND    m_windowHandle; ///< Window handle
    HDC	    m_deviceContext; ///< Device context for GDI
    HGLRC   m_renderingContext; ///< rendering context for opengl
    DWORD   m_windowStyle; ///< For use with CreateWindowEx
    DWORD   m_extendedStyle; ///< For use with CreateWindowEx
    WNDCLASSEX m_wc; ///< stores record of window attributes

    int m_desiredWidth;
    int m_desiredHeight;

    unsigned int m_posX;
    unsigned int m_posY;

    // Use these to remember windowed dimensions when switching to fullscreen
    unsigned int m_winWidth;
    unsigned int m_winHeight;

    String m_windowName;

    // Private member functions
    LONG WndProc( UINT, WPARAM, LPARAM );
    bool Init();
    void Destroy();
    void AdjustSizeForWin32();
    HGLRC SetupGl( GlConfig& config );
    void ReInit();

    // Private and static members
    static LONG         SetDisplay( DWORD width, DWORD height, DWORD depth );
    static LONG WINAPI  WndProcRouter( HWND, UINT, WPARAM, LPARAM );
};

}
#endif

