#include "GlWindow.h"
#include "XKeyboard.h"

#include <assert.h>
#include <string.h>
#include <X11/extensions/Xrender.h>
#include <time.h>
#include <signal.h>

#include "../../glkcore/containers/ListIterator.h"
#include "../../glkcore/thread/MutexLock.h"
#include "../GlWindow.h"
#include "../KeyboardHandler.h"
#include "../MouseHandler.h"

namespace GLK
{

//! Used to serialise initialisation of windows when in multi-threaded use.
Mutex GlWindow::ms_xlibLock( Mutex::Recursive );

void eventTimerCallback( sigval_t val )
{
    GlWindow* w = static_cast<GlWindow*>( val.sival_ptr );
    w->TimerExpired( 1 );
}

GlWindow::GlWindow( int w, int h, int flags, const GlConfig* const glConfig, const char* const displayName )
:
    WindowBase( w, h, flags, glConfig ),
    m_display (0),
    m_displayName ( displayName ),
    m_eventTimer  ( 0 )
{
    // Last time I checked it was ok to call XInitThreads
    // multiple times because it checks _Xglobal_lock first.
    // (And last time I checked it wasn't thread safe, despite what some
    // people might claim about Xlib!)
    ms_xlibLock.Lock();
    Status status = XInitThreads();
    assert( status != 0 );
    ms_xlibLock.Unlock();

    if ( !IsAsynchronous() )
    {
        Init();
    }
}

GlWindow::~GlWindow()
{
    if ( !IsAsynchronous() )
    {
        Destroy();
    }
}

/**
	Starts the application's event loop

   	If initialisation has been delayed then we do that first.

   	Next we unlock a mutex which signals that the event loop has started.
   	Calls are then made to initialise GL and resize before starting the event loop.

   	When the app exits the event loop then a request is made to destroy GL.
**/
int GlWindow::StartEventLoop()
{
	int exitCode;

    if ( IsAsynchronous() )
    {
        Init();
        Show();
    }

    if ( m_display )
    {
        bool initOk = InitGL();
        
        if ( initOk )
        {                    
            SignalEventLoopStart();
            Resize( m_width, m_height );

            // Start X message loop
            int go;
            do
            {
                // Application update and render
                if ( Update( m_timer.GetMilliSeconds() ) )
                {
                    m_timer.Reset();
                    Render();
                }
                else
                {   
                    // Xlibs's clumsy way of sleeping until there is a new event. Fucking idiots.
                    // Just have a (threadsafe) 'WaitForEvent()' function, even Microsoft worked that one out.
                    // And just how am i supposed to lock xlib and then sleep without blocking all xlib threads huh? Does XLockDisplay() release the lock while blocking? I don't think so.
                    XEvent event;
                    XNextEvent( m_display, &event );
                    XPutBackEvent( m_display, &event ); // Put back the event so it is dealt with in ProcessXEvents()
                }

                go = ProcessXEvents( exitCode );

            } while ( go );

            m_eventLoopMutex.Lock();
            m_processingEvents = false;
            m_eventLoopMutex.Unlock();
        }
        else
        {
            fprintf( stderr,"InitGL() returned false: (application specific initialisation unsuccessful)\n" );
        }

        DestroyGL();

        if ( IsAsynchronous() )
        {
            Destroy();
        }
    }
    else
    {
    	exitCode = -1;
    }

    return exitCode;
}

/**
   Get window manager to display the applications window.
   
   For asynchronous windows can only be called in the window's
   thread (e.g. in its Run() method) because it calls glXMakeCurrent.
**/
void GlWindow::Show()
{
    
    if ( m_display )
    {
        XMapWindow(m_display, m_window);
        glXMakeCurrent(m_display, m_window, m_context);
    }
    
}

/**
   Get the application to hide the application window.
   
   For asynchronous windows can only be called in the window's
   thread (e.g. in its Run() method) because it calls glXMakeCurrent.
**/
void GlWindow::Hide()
{
    if ( m_display )
    {
        glXMakeCurrent(m_display, None, 0);
        XUnmapWindow(m_display, m_window);
    }
}

/**
    Minimise window.    
**/
void GlWindow::Minimise()
{
    WaitForEventLoop();

    if ( m_display )
    {
        XMoveResizeWindow( m_display, m_window, 0, 0, 0, 0 );
    }
}

/**
    Maximise the window by moving it to origin and
    setting its width and height to match the default
    screen size.    
**/
void GlWindow::Maximise()
{
    WaitForEventLoop();

    if ( m_display )
    {
        XWindowAttributes xwa;
        XGetWindowAttributes( m_display, DefaultRootWindow(m_display), &xwa);
        int w = XDisplayWidth(m_display, DefaultScreen(m_display));
        int h = XDisplayHeight(m_display, DefaultScreen(m_display));
        XMoveResizeWindow( m_display, m_window, 0, 0, w, h );
    }
}

/**
    Change the position and size of the window.

    This function blocks until the event loop has started.
    Hence, if the event loop is only started at a later point
    in the same thread then this function will block forever!

    If called before the event loop is started (in a different thread) then
    the new width and height will be applied but the window will not be moved.
    (This limitation is due to Xlib).

    If w and h are zero (default) then width and height of window are not changed.
    @param x New x coord of window.
    @param y New y coord of window.
    @param w New width of window.
    @param h new height of window.
**/
void GlWindow::Move( int x, int y, int w, int h )
{
    if ( m_display && !IsFullscreen() )
    {
        WaitForEventLoop();

        if ( w == 0 || h == 0 )
        {
            w = static_cast<int>( m_width );
            h = static_cast<int>( m_height );
        }

        XMoveResizeWindow( m_display, m_window, x, y, w, h );
    }
}

/**
    Asks the window to return from its message loop.

    @param exitCode to return from message loop.
**/
void GlWindow::Quit( int exitCode )
{
    

    if ( m_window )
    {
        XClientMessageEvent xevent;
        Display* dpy = XOpenDisplay( 0 );
        xevent.type = ClientMessage;
        xevent.message_type = XInternAtom( dpy, "MY_QUIT", False );
        xevent.format = 32;
        xevent.data.l[0] = exitCode;
        XSendEvent( dpy, m_window, 0, 0, (XEvent*)&xevent );
        XCloseDisplay(dpy);
    }
    
    
}

/**
    Notifies the window that it should call Update().

    Effectively this can be used to wake the Window when
    it waiting for activity. (The window will wait in this
    way if Update() returns false).

    If the window's event loop is not running the call blocks until it starts,
    then sends the notification.
**/
void GlWindow::Notify()
{
    if ( m_window )
    {
        WaitForEventLoop(); // Having this after testing m_window stops some deadlocks,
        // but m_window could be deleted at any time - needs mutex
        
        XClientMessageEvent xevent;
        Display* dpy = XOpenDisplay(0);
        if (dpy)
        {
            XSetCloseDownMode( dpy, DestroyAll );
            xevent.type = ClientMessage;
            xevent.message_type = XInternAtom( dpy, "MY_NOTIFY", False );
            xevent.format = 32;
            xevent.data.l[0] = 0;
            XSendEvent(dpy, m_window, 0, 0, (XEvent *)&xevent);
            XCloseDisplay(dpy);
        }
    }
}

/**
    Start a timer which causes the window to update at a specified interval.

    @param millisecs The desired time interval (approx.) between calls to Update().
**/
void GlWindow::StartTimer( unsigned int millisecs )
{
    if ( m_eventTimer )
    {
        timer_delete( m_eventTimer );
    }

    // setup the signal description - want multi-threaded notifier
    struct sigevent sig;
    sig.sigev_notify = SIGEV_THREAD;
    sig.sigev_notify_function = eventTimerCallback;
    sig.sigev_value.sival_ptr = this;
    sig.sigev_notify_attributes = 0;

    // Create an event timer for regular callbacks
    int err = timer_create( CLOCK_REALTIME, &sig, &m_eventTimer );
    assert( err == 0 );

    itimerspec spec;
    spec.it_interval.tv_sec = millisecs / 1000;
    millisecs -= spec.it_interval.tv_sec * 1000;
    spec.it_interval.tv_nsec = millisecs * 1000000;

    spec.it_value.tv_sec = spec.it_interval.tv_sec;
    spec.it_value.tv_nsec = spec.it_interval.tv_nsec;

    err = timer_settime( m_eventTimer, 0, &spec, 0 );
    assert( err == 0 );
}


/**
    Stop the timer.
**/
void GlWindow::StopTimer()
{
    if ( m_eventTimer )
    {
        timer_delete( m_eventTimer );
        m_eventTimer = 0;
    }
}

/**
   Setup an X window for OpenGL rendering.
**/
void GlWindow::Init()
{
    assert( Button1 == 1 && Button3 == 3 );

    // Connect to X server.
    ms_xlibLock.Lock();
    m_display = XOpenDisplay( m_displayName );
    ms_xlibLock.Unlock();

    if ( m_display )
    {
        fprintf(
                stderr,
                "Opened %s's X Display\n",
                ServerVendor( m_display )
                );

        /*
        int defaultScreen = DefaultScreen( m_display );
            
        fprintf(
                stderr,
                "Default Screen: %dx%d %dbpp\n",
                DisplayWidth( m_display, defaultScreen ),
                DisplayHeight( m_display, defaultScreen ),
                DefaultDepth( m_display, defaultScreen )
               );

        fprintf( stderr, "(Claimed) Physical Size: %d x %d [mm]\n", DisplayWidthMM( m_display, DefaultScreen(m_display) ), DisplayHeightMM( m_display, DefaultScreen(m_display) )  );*/

        // Check glx is available.
        if ( glXQueryExtension( m_display, 0, 0 ) )
        {
            int glxMajor, glxMinor;
            glXQueryVersion( m_display, &glxMajor, &glxMinor );
            fprintf( stderr, "GLX version %d.%d available (GLK not tested with version < 1.2)\n", glxMajor, glxMinor );

            if ( IsComposited() && ( glxMajor==1 && glxMinor <3 )  )
            {
                m_flags ^= AlphaComposited; // turn off alpha because its not supported by this glx version
                fprintf( stderr, "Alpha compositing is only available if GLX version is 1.3 or greater." );
            }

            // Get a frame buffer configuration.
            XVisualInfo* xVisual = ChooseVisual( m_glConfig );
            if ( xVisual )
            {
                // Create a direct rendered GLX context.
                m_context = glXCreateContext( m_display, xVisual, 0, True );
                if ( !glXIsDirect( m_display, m_context ) )
                {
                    fprintf( stderr,"Warning: Couldn't get a direct rendered context\n" );
                }
               
                // Create an X window
                Window root = RootWindow( m_display, xVisual->screen );
                int x = 0, y = 0;
                unsigned int border = 0;
                XSetWindowAttributes att;
                //att.border_pixel = 0;
                att.colormap = XCreateColormap( m_display, root, xVisual->visual, AllocNone );

                att.event_mask =
                    StructureNotifyMask | KeyPressMask | KeyReleaseMask |
                    ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
                    FocusChangeMask | VisibilityChangeMask | ExposureMask;

                m_window = XCreateWindow( m_display, root, x, y, m_width, m_height, border, xVisual->depth,
                        InputOutput, xVisual->visual, CWBorderPixel | CWEventMask | CWColormap, &att );

                // Window should receive close events.
                Atom del = XInternAtom( m_display, "WM_DELETE_WINDOW", True );
                XSetWMProtocols( m_display, m_window, &del, 1 );

                if ( m_flags & FullScreen )
                {
                    SetDecorations( 0 );
                    SetFullscreen();
                }
                else if ( m_flags & Decorless )
                {
                    // If Decorless is specified then hint to the window manager that we want no window decorations
                    // Some window managers may not support this and some may ignore it deliberately.
                    SetDecorations( 0 );
                }

                XFree( xVisual );
            }
            else
            {
                fprintf( stderr, "Error: no suitable X visual was found\n" );
                XCloseDisplay( m_display );
                m_display = 0;
            }
        }
        else
        {
            fprintf( stderr, "Error: GLX extension not available.\n" );
            
            XCloseDisplay( m_display );
            m_display = 0;
        }
    }
    else
    {
        fprintf( stderr,"Error: Could not get X display (%s)\n", m_displayName );
    }

}

/**
    Chooses a GLX framebuffer configuration for use with OpenGL.
    The config parameter contains the desired config on input and
    stores the config that was seldected on output. These may not
    be the same if an identical config was not available on the system.
    
    @param config Specifies the desired config and returns the resulting config.
**/
XVisualInfo* GlWindow::ChooseVisual( GlConfig& config )
{
    XVisualInfo* xVisual = 0;

    if ( IsComposited() )
    {
        if ( config.GetAlphaSize() == 0 )
        {
            config.SetAlphaSize(1); // Request only one bit alpha channel for maximum compatibility
        }        
    }
    
    int doubleBuffer;
    if ( config.IsDoubleBuffered() )
    {
        doubleBuffer = True;
    }
    else
    {
        doubleBuffer = False;
    }

    // Get a frame buffer configuration.
    int xAttribs[] = {
        GLX_X_VISUAL_TYPE, GLX_TRUE_COLOR,
        GLX_DRAWABLE_TYPE, GLX_WINDOW_BIT,
        GLX_RENDER_TYPE,   GLX_RGBA_BIT,
        GLX_DOUBLEBUFFER, doubleBuffer,
        GLX_RED_SIZE,     config.GetRedSize(),
        GLX_GREEN_SIZE,   config.GetGreenSize(),
        GLX_BLUE_SIZE,    config.GetBlueSize(),
        GLX_ALPHA_SIZE,   config.GetAlphaSize(),
        GLX_DEPTH_SIZE,   config.GetDepthSize(),
        GLX_STENCIL_SIZE, config.GetStencilSize(),
        None
    };

    int nConfigs = 0;
    const int& xScreen = DefaultScreen( m_display );
    GLXFBConfig* xFB = glXChooseFBConfig( m_display, xScreen, xAttribs, &nConfigs );

    if ( xFB )
    {
        int fb;
        if ( IsComposited() )
        {
            // If a composited window was required we have more work to do.
            // (We can't just choose a visual with alpha - it needs to be suitable for Xrender.)
            fb = FindFormatWithAlpha( xFB, nConfigs );
            if ( fb == -1 )
            {
                fb = 0;
            }
        }
        else
        {
            fb = 0;
        }

        xVisual = glXGetVisualFromFBConfig( m_display, xFB[fb] );
        
        int value;
        glXGetFBConfigAttrib( m_display, xFB[fb], GLX_RED_SIZE, &value );
        config.SetRedSize( value );
        glXGetFBConfigAttrib( m_display, xFB[fb], GLX_GREEN_SIZE, &value );
        config.SetGreenSize( value );
        glXGetFBConfigAttrib( m_display, xFB[fb], GLX_BLUE_SIZE, &value );
        config.SetBlueSize( value );
        glXGetFBConfigAttrib( m_display, xFB[fb], GLX_ALPHA_SIZE, &value );
        config.SetAlphaSize( value );
        glXGetFBConfigAttrib( m_display, xFB[fb], GLX_DEPTH_SIZE, &value );
        config.SetDepthSize( value );
        glXGetFBConfigAttrib( m_display, xFB[fb], GLX_STENCIL_SIZE, &value );
        config.SetStencilSize( value );
        glXGetFBConfigAttrib( m_display, xFB[fb], GLX_DOUBLEBUFFER, &value );
        config.SetDoubleBuffered( value == True );
        XFree(xFB);
    }
    else
    {
        config.Show( stderr );
        fprintf( stderr, "Error: No suitable GLX frame buffer config available.\n" );
    }
    
    return xVisual;
}

/**
   Search for framebuffer with an Xrender compatible alpha channel (we can't just
   choose any old visual with alpha). This is useful for desktop compositing amongst
   other things.
**/
int GlWindow::FindFormatWithAlpha( GLXFBConfig* xFB, int nConfigs )
{
    int fb = -1;
    int xra, xrb;
    if ( XRenderQueryExtension( m_display, &xra, &xrb ) )
    {
        fprintf( stderr, "XRender extension available: " );
        int i;
        for ( i=0; i<nConfigs && fb == -1; ++i )
        {
            XVisualInfo* vi = glXGetVisualFromFBConfig(m_display, xFB[i]);
            if ( vi )
            {
                XRenderPictFormat* pFormat = XRenderFindVisualFormat(m_display, vi->visual);

                if ( pFormat )
                {
                    if ( pFormat->direct.alphaMask > 0 )
                    {
                        fprintf( stderr, "Frame buffer %d has alpha (%d)\n\n", i, pFormat->direct.alphaMask );
                        fb = i;
                    }
                }
                XFree(vi);
             }
         }
    }

    return fb;
}

/**
   Hide then destroy the window.
**/
void GlWindow::Destroy()
{
    Hide();
    
    if  ( m_display )
    {
        
        XDestroyWindow( m_display, m_window );
        glXDestroyContext( m_display, m_context );
        XCloseDisplay(m_display);
        m_window = 0;
        m_display = 0;
    }

    StopTimer();
}

/**
   Clear the keyboard status array.
**/
void GlWindow::ClearKeys(bool* p)
{
    bool* pEnd = p + GLK::NumKeys-1;

    while ( p != pEnd )
    {
        assert( p < pEnd );
        *p++ = false;
    }
}

/**
   Set window title bar string.
**/
void GlWindow::SetWindowName( const char* name )
{
    WaitForEventLoop();
    
    if ( m_display )
    {
        XTextProperty title;
        XmbTextListToTextProperty( m_display, const_cast<char**>(&name), 1, XStringStyle, &title );
        XSetWMName( m_display, m_window, &title );
        XFree( title.value );
        
        Notify();
    }    
}

/**
    Private utility function for processing events in the window's event loop.
    Called by StartEventLoop();

    @param Reference into which exit code is written (iff return value is 0).
    @return Returns 1 if event loop should continue, 0 if it should exit.
**/
int GlWindow::ProcessXEvents( int& exitCode )
{
    XEvent event;
    int go = 1;

     // if you release the lock temporarily in any case below you must relock it before the end of case.
    
    while ( XPending( m_display ) > 0 )
    {
        XNextEvent( m_display, &event );

        bool mouseDown = false;
        switch ( event.type )
        {
        case ConfigureNotify:
        {
            if ( event.xconfigure.width != m_width || event.xconfigure.height != m_height )
            {
                glXMakeCurrent( m_display, m_window, m_context );               
                m_width  = event.xconfigure.width;
                m_height = event.xconfigure.height;                
                Resize( m_width, m_height );
            }
            break;
        }

        case MotionNotify:
        {
            
            MutexLock lock( m_mhMutex );
            if ( m_mh.Length() != 0 )
            {
                m_mh.GetLast()->NewPosition( event.xmotion.x, event.xmotion.y );
            }
            
            break;
        }

        case KeyPress:
        {
            int index = 0 + (event.xkey.state & ShiftMask);
            KeySym ks = XLookupKeysym( &event.xkey, index );

            if ( IsModifierKey(ks) == False && IsFunctionKey(ks) == False )
            {
                char k = static_cast<char>(ks);

                // Special conversion required for some keys
                switch (ks)
                {
                case XK_Left:
                    k = '\v';
                    break;
                case XK_Up:
                    k = 28;
                    break;
                case XK_Right:
                    k = '\f';
                    break;
                case XK_Down:
                    k = 29;
                    break;
                case XK_Delete:
                    k = 16;
                    break;
                default:
                    k = static_cast<char>(ks);
                    break;
                }

                if ( ks == XK_Escape )
                {
                    go = 0;
                }
                else
                {
                    
                    MutexLock lock( m_kbhMutex );
                    if ( m_kbh.Length() != 0 )
                    {
                        m_kbh.GetLast()->Key( k );
                    }
                    
                }

            }

            break;
        }

        case KeyRelease:
        {
            break;
        }

        case ButtonPress:
            mouseDown = true;
        case ButtonRelease:
        {
            
            MutexLock lock( m_mhMutex );
            if ( m_mh.Length() != 0 )
            {
                if ( event.xbutton.button < Button4 )
                {
                    m_mh.GetLast()->ButtonEvent( (MouseHandler::ButtonId)(event.xbutton.button - Button1), mouseDown );
                }
                else
                {
                    int delta = 0;
                    if ( event.xbutton.button == Button4 ) { delta = 1; }
                    if ( event.xbutton.button == Button5 ) { delta = -1; }
                    m_mh.GetLast()->Scroll( delta );
                }
            }
            
            break;
        }

        case ClientMessage:
        {
            if ( event.xclient.message_type == XInternAtom( m_display, "WM_PROTOCOLS", True ) )
            {
                go = 0;
                exitCode = 0;
            }
            if ( event.xclient.message_type == XInternAtom( m_display, "MY_QUIT", True ) )
            {
                exitCode = event.xclient.data.l[0];
                go = 0;
            }
            break;
        }

        case Expose:
            
            Render();
            
            break;

        default:
            break;
        }

    }

    
    return go;
}

/**
    Tell the window manager that it should stop drawing decorations (borders, title bar) for this window.
    The window manager is free to ignore this request (it is a hint only).
**/
void  GlWindow::DecorationsOff()
{
    WaitForEventLoop();
    SetDecorations( 0 );
    Notify();
}

/**
    Tell the window manager that it should show decorations (borders, title bar) for this window.
**/
void  GlWindow::DecorationsOn()
{
    WaitForEventLoop();
    SetDecorations( 1 );
    Notify();
}

/**
    Set window decorations to be on or off.
    Some windowing systems and window managers do not support this.
    In such cases the function has no effect.
    
    @param state Zero turns decorations off, non-zero turns them on.
**/
void  GlWindow::SetDecorations( int state )
{
    if  ( m_display )
    {
        // The following is just a hint to the window manager - it might be ignored.
        struct {
            unsigned long flags;
            unsigned long functions;
            unsigned long decorations;
            long input_mode;
            unsigned long status;
        } hints = { 2, 0, 0, 0, 0, };
    
        hints.decorations = state;
    
        Atom property = XInternAtom ( m_display, "_MOTIF_WM_HINTS", True );
        assert( property != None );
    
        XChangeProperty (
            m_display, m_window,
            property,
            property,
            32, PropModeReplace, (const unsigned char *) &hints,
            sizeof(hints) / sizeof (long)
            );
 
        if ( state == 0 )
        {
            m_flags |= Decorless; // decor turned off so set the decorless flag
        }
        else
        {
            m_flags &= ~Decorless; // decor turned on so clear the decorless flag
        }
        
    }
}

/**
    Sets the window to take up the entire screen. The window has its borders and title bar removed and is
    then set to the dimensions of the default display. It also captures all mouse and keyboard input.
    
    Calls to this function block until the window's event loop is started.
    
    Only works with NETWM compliant window managers.
**/
void GlWindow::FullscreenOn()
{   
    WaitForEventLoop();
    
    XEvent xev;
    Atom wm_state = XInternAtom(m_display, "_NET_WM_STATE", False );
    Atom fullscreen = XInternAtom(m_display, "_NET_WM_STATE_FULLSCREEN", False );

    memset(&xev, 0, sizeof(xev));
    xev.type = ClientMessage;
    xev.xclient.window = m_window;
    xev.xclient.message_type = wm_state;
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = 1;
    xev.xclient.data.l[1] = fullscreen;
    xev.xclient.data.l[2] = 0;

    XSendEvent( m_display, DefaultRootWindow(m_display), False, SubstructureNotifyMask, &xev );
    
    m_flags |= FullScreen; // set fullscreen flag
    assert( WindowBase::IsFullscreen() == true );
    
    Notify();
}

/**
    Switch fullscreen mode off.
    
    The window reverts to its last (non-fullscreen) dimensions.
    
    Calls to this function block until the window's event loop is started.
    
    Only works with NETWM compliant window managers.
**/
void GlWindow::FullscreenOff()
{
    WaitForEventLoop();
    
    XEvent xev;
    Atom wm_state = XInternAtom(m_display, "_NET_WM_STATE", False );
    Atom fullscreen = XInternAtom(m_display, "_NET_WM_STATE_FULLSCREEN", False );

    memset(&xev, 0, sizeof(xev));
    xev.type = ClientMessage;
    xev.xclient.window = m_window;
    xev.xclient.message_type = wm_state;
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = 0;
    xev.xclient.data.l[1] = fullscreen;
    xev.xclient.data.l[2] = 0;

    XSendEvent( m_display, DefaultRootWindow(m_display), False, SubstructureNotifyMask, &xev );
    
    m_flags &= ~FullScreen; // clear fullscreen flag
    
    assert( WindowBase::IsFullscreen() == false );
    
    Notify();
}

/**
    Utility function used internally to activate fullscreen mode.    
**/
void GlWindow::SetFullscreen()
{
    int defaultScreen = DefaultScreen( m_display );
    int w = DisplayWidth( m_display, defaultScreen );
    int h = DisplayHeight( m_display, defaultScreen );
    XMoveResizeWindow( m_display, m_window, 0, 0, w, h );
    XMapRaised( m_display, m_window );
    XGrabPointer( m_display, m_window, True, 0, GrabModeAsync, GrabModeAsync, m_window, 0L, CurrentTime );
    XGrabKeyboard( m_display, m_window, False, GrabModeAsync, GrabModeAsync, CurrentTime );
}

} // end namespace GLK

