#ifndef GLX_WINDOW_H
#define GLX_WINDOW_H

#include <stdio.h>

#include <X11/Xlib.h>
#include <GL/glx.h>

#include "KeyEnums.h"

#include "../WindowBase.h"
#include "../../glkcore/time/Timer.h"

namespace GLK
{

// fwd decl
class KeyboardHandler;
class MouseHandler;

/**
    Facilitates OpenGL applications on X Windows systems.
    Supports framebuffer configurations with alpha component.

    The following property changing methods,@n
        - Show@n
        - Hide@n
        - Minimise@n
        - Maximise@n
        - Move@n
    can be called asynchronously (from a thread
    other than the one in which the window was
    created) but changes will not be visible to the
    calling thread until the window's thread has had
    time to update itself. There is no mechanism for
    determining when the update has completed. So for
    example calling GetWidth() immediately after calling
    Maximise may return the old width and height.
*/
class GlWindow : public WindowBase
{
public:
    GlWindow( int w, int h, int flags = Normal, const GlConfig* const glConfig = 0, const char* const displayName = 0 );
    virtual ~GlWindow();

    int StartEventLoop();
    void SetWindowName(const char* name);
    void SwapBuffers() { glXSwapBuffers( m_display, m_window ); };
    void Show();
    void Hide();
    void Minimise();
    void Maximise();
    void Move( int x, int y, int w = 0, int h = 0 );
    void Quit( int exitCode );
    void Notify();
    void StartTimer( unsigned int millisecs );
    void StopTimer();

    virtual bool InitGL() = 0;
    virtual void DestroyGL() = 0;
    virtual void Resize(const unsigned int, const unsigned int) = 0;
    virtual bool Update( unsigned int ) = 0;
    virtual void Render() = 0;
    virtual void TimerExpired( int id ) = 0;

    void DecorationsOff();
    void DecorationsOn();
    void FullscreenOn();
    void FullscreenOff();

private:
    int ProcessXEvents( int& exitCode );
    void SetDecorations( int boolean );
    void SetFullscreen();

    XVisualInfo* ChooseVisual( GlConfig& config );

    static Mutex ms_xlibLock;

    Timer m_timer;
    Display*   m_display;
    Window     m_window;
    GLXContext m_context;

    const char* const m_displayName;

    void Init();
    void Destroy();
    int  FindFormatWithAlpha( GLXFBConfig* xFB, int nConfigs );
    void ClearKeys( bool* p );

    timer_t m_eventTimer;
};

} // end of namespace GLK

#endif // GLX_WINDOW_H
