#include "LuaWindow.h"

int DynamicLoadTest()
{
    return 99;
}

namespace GLK
{

const int SWAP_BUFFERS       = 0; 
const int SET_WINDOW_NAME    = 1;
const int SET_FULLSCREEN_ON  = 2;
const int SET_FULLSCREEN_OFF = 3;
const int START_EVENTS       = 4;
const int WAIT_FOR_CLOSE     = 5;

/**
    Construct a new LuaWIndow, register it with the specified LuaState object and bind it to a variable.
    
    @param lua The LuaState object to which this object belongs.
    @param objectName The variable name which can be used to reference this window in Lua.
    @param w Window width in pixels.
    @param h Window height in pixels.
    @param windowType Creation flags. Defaults used if this is not specified.
    @param glConfig The desired OpenGL framebuffer configuration. Defaults used if this is null or not specified.
**/
LuaWindow::LuaWindow( Lua::LuaState& lua, const String& objectName, unsigned int w, unsigned int h, int windowType, const GlConfig* const glConfig )
:
    GlThreadWindow( w, h, windowType, glConfig ),
    m_name ( objectName )
{ 
    // setup the lua object
    LuaObject::RegisterObject( &lua, m_name.cStr() );
    RegisterMethods();
}

/**
    Nothing to do here in current implementation.
**/
LuaWindow::~LuaWindow()
{
}

/**
    Register methods that can be called on a window object in Lua.
**/
void LuaWindow::RegisterMethods()
{
    RegisterMethod( "SwapBuffers",   SWAP_BUFFERS );
    RegisterMethod( "SetWindowName", SET_WINDOW_NAME );
    RegisterMethod( "FullscreenOn",  SET_FULLSCREEN_ON );
    RegisterMethod( "FullscreenOff", SET_FULLSCREEN_OFF );
    RegisterMethod( "Start", START_EVENTS );
    RegisterMethod( "WaitForClose", WAIT_FOR_CLOSE );
}

/**
    Handles any method calls that come from Lua.
**/
int LuaWindow::Dispatch( lua_State* L, lua_Number id )
{
    Lua::LuaState lua( L );    
    int callId = static_cast<int>( id );

    switch ( callId )
    {
    case SWAP_BUFFERS:
        SwapBuffers();
        break;
    case SET_WINDOW_NAME:
        if ( lua.GetStackSize() == 1 && lua.IsString() )
        {
            SetWindowName( lua.ToString( -1 ) );
            lua.Pop( 1 );
        }
        break;

    case SET_FULLSCREEN_ON:
        FullscreenOn();
        break;

    case SET_FULLSCREEN_OFF:
        FullscreenOff();
        break;
        
    case START_EVENTS:
        Show();
        StartEventLoopAsync();
        break;

    case WAIT_FOR_CLOSE:
        WaitForClose();
        break;

    default:
        break;
    }
    return 0;
}

/**
    If the InitGL function is bound then call it. If no function is bound then returns false.
        
    \note If this function returns false the window will be destroyed immediately.
    
    @return True if the initialisation function exists and was successful, false otherwise.
**/
bool LuaWindow::InitGL()
{
    bool retVal = false;
    
    Lua::LuaStackRestore restore( *GetLua() );
        
    GetLuaValue( "InitGL" );
    
    if ( GetLua()->IsFunction() )
    {
        GetLuaObject(); // pass self as first argument
        GetLua()->Call( 1, 1 );
        
        if ( GetLua()->IsBool() )
        {
            retVal = GetLua()->ToBool();
        }
    }

    return retVal;
}

/**
    If the DestroyGL function is bound then call it.
**/
void LuaWindow::DestroyGL()
{
    Lua::LuaStackRestore restore( *GetLua() );
    
    GetLuaValue( "DestroyGL" );
    
    if ( GetLua()->IsFunction() )
    {
        GetLuaObject(); // pass self as first argument
        GetLua()->Call( 1, 0 );
    }   
}

/**
    If Resize function is bound then call it.
    
    @param w New window width.
    @param h New window height.
**/
void LuaWindow::Resize( unsigned int w, unsigned int h )
{
    Lua::LuaStackRestore restore( *GetLua() );
    
    GetLuaValue( "Resize" );
    
    if ( GetLua()->IsFunction() )
    {
        GetLuaObject(); // pass self as first argument
        GetLua()->PushNumber( w );
        GetLua()->PushNumber( h );
        GetLua()->Call( 3, 0 );
    }
}

/**
    If TimerExpired function is bound then call it.
    
    If this function returns true then the Render function is called immediately.
    Otherwise the window will not call Update or Render again until a window event
    wakes it. Can be used to avoid unecessary processing/rendering.
    
    @param t The time in milliseconds since the last call to Render.
    
    @todo  Hmmm... having false as default return value means the window can freeze updates
    if called before lua update function is defined in a script.
**/
bool LuaWindow::Update( unsigned int t )
{
    bool retVal = false; // dodgy?

    Lua::LuaStackRestore restore( *GetLua() );
    
    GetLuaValue( "Update" );
    
    if ( GetLua()->IsFunction() )
    {
        GetLuaObject(); // pass self as first argument
        GetLua()->PushNumber( t );
        GetLua()->Call( 2, 1 );
        
        if ( GetLua()->IsBool() )
        {
            retVal = GetLua()->ToBool();
        }
    }

    return retVal;
}

/**
    If Render function is bound then call it.
**/
void LuaWindow::Render()
{
    Lua::LuaStackRestore restore( *GetLua() );
    
    GetLuaValue( "Render" );
    
    if ( GetLua()->IsFunction() )
    {
        GetLuaObject(); // pass self as first argument
        GetLua()->Call( 1, 0 );
    }
}

/**
    If TimerExpired function is bound then call it.
    
    @param id The timer's identifier.
**/
void LuaWindow::TimerExpired( int id )
{
    Lua::LuaStackRestore restore( *GetLua() );
    
    GetLuaValue( "TimerExpired" );
    
    if ( GetLua()->IsFunction() )
    {
        GetLuaObject(); // pass self as first argument
        GetLua()->PushNumber( id );
        GetLua()->Call( 2, 0 );
    }
}

} // end namespace GLK

