#ifndef GLK_LIST_H
#define GLK_LIST_H

#include <assert.h>

namespace GLK
{

template <class T> class List;
template <class T> class ListIterator;
template <class T> class ConstListIterator;

/**
    Template class for internal representation of list nodes in linked lists.
**/
template <class T>
class FW_API ListNode
{
    friend class List<T>;
    friend class ListIterator<T>;
    friend class ConstListIterator<T>;

private:
    ListNode( const T& item ) : m_item  (item) {};
    ListNode( const T& item, ListNode<T>* next, ListNode<T>* prev );
    ~ListNode() {};

    ListNode<T>* m_next;
    ListNode<T>* m_prev;
    T m_item;
};

/**
    Construct a new list node and insert it into the list
    (Insertion is explicitly enforced by requiring pointers
    to previous and next nodes be passed to the constructor).
**/
template <class T>
ListNode<T>::ListNode( const T& item, ListNode<T>* next, ListNode<T>* prev )
:
    m_next  (next),
    m_prev  (prev),
    m_item  (item)
{

}

/**
    Template class for doubly linked lists.
**/
template <class T>
class FW_API List
{
    friend class ListIterator<T>;
    friend class ConstListIterator<T>;

public:
    List();
    ~List();

    unsigned int Length() const  { return m_length; };

    void AddFront( const T& item );
    void AddBack( const T& item );

    T RemoveFirst();
    T RemoveLast();  
    void RemoveAll();

    T& GetFirst() { assert(m_head); return m_head->m_item; };
    const T& GetFirst() const { assert(m_head); return m_head->m_item; };
    T& GetLast() { assert(m_tail); return m_tail->m_item; };
    const T& GetLast() const { assert(m_tail); return m_tail->m_item; };

private:
    ListNode<T>* m_head;
    ListNode<T>* m_tail;
    unsigned int m_length;    
};

/**
    Initialise an empty list.
**/
template <class T>
List<T>::List()
: m_head(0), m_tail(0), m_length(0)
{

}

/**
    The list is emptied on destruction.
**/
template <class T>
List<T>::~List()
{
    RemoveAll();
}

/**
    Add an item to the front of the list.

    Allocates a new node holding a copy of item
    and adds it at the head of the List.
**/
template <class T>
void List<T>::AddFront( const T& item )
{
    ListNode<T>* node = new ListNode<T>( item, m_head, 0 );

    if ( m_head )
    {
        m_head->m_prev = node;
    }
    else
    {
        // this is first item so it is the tail:
        m_tail = node;
    }

    m_head = node; // Item Added to head of list
    m_length++;
}

/**
    Add an item to the back of the list.

    Allocates a new node holding a copy of item
    and adds it at the tail of the List.    
**/
template <class T>
void List<T>::AddBack( const T& item )
{
    ListNode<T>* node = new ListNode<T>( item, 0, m_tail );

    if ( m_tail )
    {
        m_tail->m_next = node;
    }
    else
    {
        m_head = node; // this is first node so it is also the head
    }

    m_tail = node; // Node becomes new tail
    m_length++;
}

/**
    Return the item at the front of the list and remove it.
**/
template <class T>
T List<T>::RemoveFirst()
{
    assert( m_head );
    
    T rval = m_head->m_item;
    ListNode<T>* tmp = m_head;
    m_head = m_head->m_next;
    
    if ( m_head )
    {
        m_head->m_prev = 0;
    }
    else
    {
        // head was also tail
        assert( m_length == 1 );
        m_tail = m_head;
    }
    
    m_length--;

    delete tmp;

    return rval;
}

/**
    Return the item at the back of the list and remove it.
**/
template <class T>
T List<T>::RemoveLast()
{
    assert( m_tail );
    
    T rval = m_tail->m_item;
    ListNode<T>* tmp = m_tail;
    m_tail = m_tail->m_prev;

    if ( m_tail )
    {
        m_tail->m_next = 0;
    }
    else
    {
        // tail was also head
        assert( m_length == 1 );
        m_head = m_tail;
    }

    m_length--;

    delete tmp;

    return rval;
}

/**
    
**/
template <class T>
void List<T>::RemoveAll()
{   
    ListNode<T>* current = m_head;
    m_head = 0;

    while ( current )
    {
        m_tail = current; // use m_tail as a temporary
        current = current->m_next;
        delete m_tail;
    }

    m_tail = 0;
    m_length = 0;
}

} // end namespace GLK

#endif // GLk_LIST_H
