// NO HEADER GUARD (DELIBERATELY)

#include "List.h"

#undef ITR_CLASSNAME
#undef ITR_CONST

#ifdef GLK_CONST_ITERATOR

// Compiling the const version:
#define ITR_CLASSNAME ConstListIterator
#define ITR_CONST const

#else

// Compiling the normal version
#define ITR_CLASSNAME ListIterator
#define ITR_CONST

#endif

namespace GLK
{

/**
    ! THIS FILE IS A PREPROCESSOR TEMPLATE AND SHOULD NOT BE INCLUDED DIRECTLY !

    Class for traversing all items of a list.

    These list iterators are based on the hasNext-getNext paradigm (also used in Java/Qt).    

**/
template <class T>
class FW_API ITR_CLASSNAME
{
public:
    explicit ITR_CLASSNAME( ITR_CONST List<T>& list ) : m_list(list), m_prev( 0 ), m_next( m_list.m_head ), m_fwd(true) { };

    //! Test if there is a next item in the list.
    bool HasNext() const { return m_next != 0; };
    inline ITR_CONST T& Next();

    //! Test if there is a previous item in the list.
    bool HasPrev() const { return m_prev != 0; };
    inline ITR_CONST T& Prev();

    //! Set the iterator to the beginning of the list
    void Begin() { m_prev = 0; m_next = m_list.m_head; m_fwd = true; };
    
    //! Set the iterator to the tail of the list
    void End() { m_prev = m_list.m_tail; m_next = 0; m_fwd = false; };

#ifndef GLK_CONST_ITERATOR
    
    //! Insert a new node between the current node and the next and move the iterator over the new node.
    void Insert( const T& );

    //! Remove the current node (the last one jumped over)
    void Remove();

    //! Set the value at the current node (the last one jumped over)
    void Set( const T& newValue );

#endif

private:
    SET_UNCOPYABLE( ITR_CLASSNAME );

    ITR_CONST List<T>& m_list;
    ListNode<T>* m_prev;
    ListNode<T>* m_next;
    bool m_fwd;
};

/**
    Return a reference to the next item in the list.

    There must exist such an item or the program will assert in
    debug mode (and crash otherwise).
    You can check there is a next item using ITR_CLASSNAME<T>::HasNext().
**/
template <class T>
ITR_CONST T& ITR_CLASSNAME<T>::Next()
{
    assert( m_next );
    m_prev = m_next;
    m_next = m_next->m_next;
    m_fwd = true;
    return m_prev->m_item;
}

/**
    Return a reference to the previous item in the list.
    
    There must exist such an item or the program will assert in
    debug mode (and go wrong otherwise).
    You can check there is a next item using ITR_CLASSNAME<T>::HasPrev().
**/
template <class T>
ITR_CONST T& ITR_CLASSNAME<T>::Prev()
{
    assert( m_prev );
    m_next = m_prev;
    m_prev = m_prev->m_prev;
    m_fwd = false;
    return m_next->m_item;
}

#ifndef GLK_CONST_ITERATOR

template <class T>
void ListIterator<T>::Insert( const T& )
{

}

template <class T>
void ListIterator<T>::Set( const T& newValue )
{
    if ( m_fwd )
    {
        // iterated forwards on operate on prev
        assert( m_prev );
        
        m_prev->m_item = newValue;
    }
    else
    {
        // iterated backwards on last move so operate on next
        assert(m_next);
        m_next->m_item = newValue;
    }
}

/**
    Remove the last node returned by Next() or Prev().
**/
template <class T>
void ListIterator<T>::Remove()
{
    ListNode<T>* tmp;

    if ( m_fwd )
    {
        // iterated forwards on last move so remove previous element
        assert( m_prev );
        
        // set the new previous node
        tmp = m_prev;
        m_prev = m_prev->m_prev;

        if ( m_prev )
        {
            // there is still a previous so update it's next
            m_prev->m_next = m_next;
        }
        else
        {
            // we removed head
            m_list.m_head = m_next;
        }

        if ( m_next )
        {
            // there is a next so update it's previous
            m_next->m_prev = m_prev;
        }
        else
        {
            // we removed tail
            m_list.m_tail = m_prev;
        }

    }
    else
    {
        // iterated backwards on last move so remove next element
        assert(m_next);
        
        // set the new previous node
        tmp = m_next;
        m_next = m_next->m_next;

        if ( m_next )
        {
            // there is still a next so update it's previous
            m_next->m_prev = m_prev;
        }
        else
        {
            // we removed tail
            m_list.m_tail = m_prev;
        }

        if ( m_prev )
        {
            // there is a previous so update it's next
            m_prev->m_next = m_next;
        }
        else
        {
            // we removed head
            m_list.m_head = m_next;
        }

    }

    // delete disconnected node:
    delete tmp;
    m_list.m_length--;
}

#endif

} // end namespace GLK
