#include "Glyph.h"

#include "Face.h"

#include <assert.h>

namespace Ft
{

/**
    Will take a copy of the glyph currently
    loaded by the specified Face object.
**/
Glyph::Glyph( const Face& face )
{
    m_err = FT_Get_Glyph( face.m_face->glyph, &m_glyph );
    assert( m_err == 0 );

    m_metrics = face.m_face->glyph->metrics;
}

/**
    Copy another glyph.
**/
Glyph::Glyph( const Glyph& g )
{
    m_err = FT_Glyph_Copy( g.m_glyph, &m_glyph );
    assert( m_err == 0 );
    m_metrics = g.m_metrics;
}

/**
    Release the glyph.
**/
Glyph::~Glyph()
{
    FT_Done_Glyph( m_glyph );
}

/**
    Copy operator.
**/
const Glyph& Glyph::operator = ( const Glyph& g )
{
    FT_Done_Glyph( m_glyph );
    m_err = FT_Glyph_Copy( g.m_glyph, &m_glyph );
    assert( m_err == 0 );
    
    m_metrics = g.m_metrics;

    return *this;
}

/**
    Return the dimensions of the control box.
**/
FT_BBox Glyph::GetControlBox( FT_UInt mode ) const
{
    FT_BBox bbox;
    FT_Glyph_Get_CBox( m_glyph, mode, &bbox );
    return bbox;
}

/**
    Apply an affine transform to the glyph.

    @param A Transformation matrix.
    @param t Translation vector (coords are 1/64th of a pixel).
**/
void Glyph::Transform( FT_Matrix* A, FT_Vector* t )
{
    FT_Glyph_Transform( m_glyph, A, t );
}

/**
    Convert the font to a bitmap. The bitmap can be retrieved using ExtractBitmap().

    @param mode The render mode.
    @param origin Pointer to offset vector for rendering. If null (default) then (0,0) will be used.
**/
FT_Error Glyph::Render( FT_Render_Mode mode, FT_Vector* origin )
{   
    m_err = FT_Glyph_To_Bitmap( &m_glyph, mode, origin, 1 );
    return m_err;
}

/**
    Copy the glyph's bitmap into the bitmap with the specified width, height, and stride.
    
    @param pDst The buffer into which the bitmap should be extracted.
    @param width Width of the destination bitmap.
    @param height of the destination bitmap.
    @param stride Number of bytes between rows of the destination bitmap.
    
    // @todo here we use height and width instead of h and w because h and w could be larger than bitmap size: need to fix allocation at caller.
**/
void Glyph::ExtractBitmap( unsigned char* pDst, int width, int height, int stride )
{
    FT_BitmapGlyph glyph = reinterpret_cast< FT_BitmapGlyph >( m_glyph );
    assert( glyph->bitmap.pixel_mode == FT_PIXEL_MODE_GRAY );

    unsigned char* pSrc = glyph->bitmap.buffer;

    int w = glyph->bitmap.width;
    int h = glyph->bitmap.rows;
    int s = glyph->bitmap.pitch - glyph->bitmap.width + (glyph->bitmap.width-width);
    int skip = stride-width;

    unsigned char* pDstEnd;

    // Here we use height and width instead of h and w because h and w could be larger than bitmap size
    // Need to fix allocation at caller.
    while ( height-- != 0 )
    {
        pDstEnd = pDst + width;
        while ( pDst != pDstEnd )
        {
            *pDst++ = *pSrc++;
        }

        pDst += skip;
        pSrc += s;
    }
}

int Glyph::GetRows() const
{
    FT_BitmapGlyph glyph = reinterpret_cast< FT_BitmapGlyph >( m_glyph );
    return glyph->bitmap.rows;
}

int Glyph::GetWidth() const
{
    FT_BitmapGlyph glyph = reinterpret_cast< FT_BitmapGlyph >( m_glyph );
    return glyph->bitmap.width;
}

int Glyph::GetPitch() const
{
    FT_BitmapGlyph glyph = reinterpret_cast< FT_BitmapGlyph >( m_glyph );
    return glyph->bitmap.pitch;
}

int Glyph::GetLeft() const
{
    FT_BitmapGlyph glyph = reinterpret_cast< FT_BitmapGlyph >( m_glyph );
    return glyph->left;        
}

int Glyph::GetTop() const
{
    FT_BitmapGlyph glyph = reinterpret_cast< FT_BitmapGlyph >( m_glyph );
    return glyph->top;
}

} //end namespace Ft

