#include "File.h"

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

#include <assert.h>

#ifdef _MSC_VER

#include <tchar.h>

// Dodgy hack because Microsoft don't implement vscanf.
// Well, actually they do but call it _input_l() and don't expose the function!
// It is likely that this hack will only work on VS2008.
// It also requires static linking to CRT as the function will not be in the CRT dlls.
//#ifdef __cplusplus
//extern "C"
//#endif
//int __cdecl _input_l (FILE *infile, const char *format, _locale_t plocinfo, va_list arglist );
//#define vfscanf(fp,f,a) _input_l((fp),(f),0,(a))

#endif

namespace GLK
{

File::File( const char* name, const char* mode )
{
    fp = fopen( name, mode );
}

File::~File()
{
    if ( fp )
    {
        fclose(fp);
    }
}

/**
    Writes a specified number of bytes to the file.

    @param buffer Array of char containing the bytes to be written.
    @param numBytes The number of bytes to write.
    @return The number of bytes successfully written.
**/
size_t File::Write( const char* buffer, size_t numBytes )
{
    assert( fp );

    return fwrite( buffer, 1, numBytes, fp );
}

/**
    Formatted print.

    The format specifiers are the same as for stdio's printf.
    
    @return The number of characters written.
**/
int File::Printf( const char* formatString, ... )
{
    assert( fp );

    va_list args;
    va_start( args, formatString );

    int n = vfprintf( fp, formatString, args );

    va_end( args );

    return n;
}

/**
    Read a specified number of bytes from the file.

    @param buffer Destination buffer for the bytes we read.
    @param numBytes The number of bytes to read.
    @return The number of bytes read.
**/
size_t File::Read( char* buffer, size_t numBytes )
{
    assert( fp );

    return fread( buffer, 1, numBytes, fp );
}

/**
    Performs a stdio-like scanf on the file at the current cursor position.

    Format specifiers are the same as for scanf.

    @param formatString The scanf format string.
    @return The number of fields successfully converted.
    
    @todo  Fix Scanf for MS VS2008!	
**/
int File::Scanf( const char* formatString, ... )
{
    assert( fp );
    
    va_list args;
    va_start( args, formatString );

#ifdef  _MSC_VER
    int n = 0;
#else
    int n = vfscanf( fp, formatString, args );
#endif

    va_end( args );

    return n;
}

/**
    Keep reading characters into buffer until CR or LF is
    encountered or until there is no more data available.

    CR+LF will appear as 2 lines - The second empty.

    The new-line character is not written to buffer.

    A terminating character is written to the buffer.

    An empty line will return a count of 1 and a terminating character will be written at buffer[0].

    If there is an end of file then ReadLine will return 0;

    @param buffer Buffer to receive the line. Must be at least maxBytes in size.
    @return The number of characters put in the buffer including the terminating character.
**/
size_t File::ReadLine( char* buffer, size_t maxBytes )
{
    char c = '\0';
    size_t count = 0;
    size_t numRead = 1;

    while (
            count != maxBytes
            && numRead
            && !( c == '\r' || c == '\n' )
        )
    {
        numRead = fread( buffer, 1, 1, fp );
        c = *buffer;
        count += numRead;
        buffer += numRead;
    }

    if (count)
    {
        *(buffer - 1) = '\0';
    }

    return count;
}

/**
    Convert the whole file into a string representation.
    The file is rewound, then the entire file is converted to a string.

    After calling ToString() the file cursor will be at the end of file
    so for further operations on the file it will need to be rewound (see Rewind());

    @param s The destination string which will hold the file's contents.
    @return The number of bytes in the file.
**/
size_t File::ToString( String& s )
{
    assert( fp );

    fseek( fp, 0, SEEK_END );
    long int end = ftell( fp );
    rewind( fp );
    long int start = ftell( fp );
    
    char* buf = new char[ end-start ];
    int n = fread( buf, 1, end-start, fp );   
    assert( n <= end-start );
    buf[n] = '\0';

    s = String( buf, n );

    return n;
}

/**
    Static utility function which will open a specified file,
    convert it's contents to a string and close the file.

    @param fileName Name of file to read from.
    @param fileContents Reference to string in which we will store the file's contents.
    @return true if the file was opened successfully, false otherwise.
**/
bool File::FileToString( const String& fileName, String& fileContents )
{
    bool ok;

    File f( fileName.cStr(), "r" );
    
    ok = f.IsOpen();
    
    if ( ok )
    {
        f.ToString( fileContents );
    }

    return ok;
}

} // end namespace GLK
