#include "LuaObject.h"
#include "LuaState.h"

extern "C" {
#include <lauxlib.h>
}

#include <assert.h>
#include <string.h>

namespace Lua
{

/**
 **/
int LuaObject::GlobalDispatch( lua_State* L )
{
    LuaState lua( L );

    assert( lua.IsLightUserData( lua_upvalueindex( 1 ) ) ); // the first upvalue is the object's this pointer

    LuaObject* obj = reinterpret_cast<LuaObject*>( lua.GetUserData( lua_upvalueindex( 1 ) ) );
    return obj->Dispatch( L, lua.GetUpvalue(2) );
}

LuaObject::LuaObject()
:
    m_luaName   (0),
    m_lua       (0)
{

}

/**
    Unregister object in the destructor - otherwise calls could be made
    from Lua which access the invalid this pointer.
**/
LuaObject::~LuaObject()
{
    UnregisterObject();
    delete [] m_luaName;
}

/**
    Allows us to register and access objects in Lua in an object oriented fashion.
    Create a named table in Lua (global variable) that will represent this object.

    If a global with the same name already exists it will be overwritten with the
    object's table (and then garbage collected at some later point).

    @param lua The Lua state into which the object is registered.
    @param name The name of the table.
**/
void LuaObject::RegisterObject( LuaState* lua, const char* name )
{
    assert(lua);

    // create a new table
    lua->NewTable();
    lua->SetGlobal( name );

    // Keep a record of the lua state with which the object is registered
    m_lua = lua;

    // Keep a record of the named-variable which holds this object's Lua representation.
    size_t n = strlen(name);
    delete [] m_luaName;
    m_luaName = new char[n+1];
    strncpy( m_luaName, name, n );
    m_luaName[n] = '\0';
}

/**
    Allows us to register and access objects in Lua in an object oriented fashion.

    Similar to RegisterObject but does not overwrite the named table - appends to it instead.
    This allows us to initialise variables in the table which can then be read by the object
    after it has registered itself. An example use would be object specific configuration values.

    If the named variable exists (is not nil) but is not a table then the object will not be registered
    with Lua.

    @param lua The Lua state into which the object is registered.
    @param name The name of the table.

    @return true if the object was successfully registered, false if the named variable was not nil or a table.
**/
bool LuaObject::JoinObject( LuaState* lua, const char* name )
{
    bool ok;

    // get the variable with the given name
    lua->GetGlobal( name );

    int type = lua->GetType();
    if ( type == LUA_TNIL || type == LUA_TTABLE )
    {
        if ( type == LUA_TNIL )
        {
            // The variable is nil so we must create a table
            lua->NewTable();
            lua->SetGlobal( name );
        }

        // Keep a record of the lua state with which the object is registered
        m_lua = lua;

        // Keep a record of the named-variable which holds this object's Lua representation.
        size_t n = strlen(name);
        delete [] m_luaName;
        m_luaName = new char[n+1];
        strncpy( m_luaName, name, n );
        m_luaName[n] = '\0';

        ok = true;
    }
    else
    {
        ok = false;
    }

    return ok;
}

/**
    Remove object from Lua by assigning to nil.
**/
void LuaObject::UnregisterObject()
{
    if ( m_lua )
    {
        m_lua->PushNil();
        m_lua->SetGlobal( m_luaName );
        m_lua->Pop( 1 );
     }
}

/**
    Register a method name with the object's Lua representation.

    Inserts a named function into the object's Lua table.
    When called, the 'id' the function was registered with will be passed
    to the correct object's Dispatch() method. You (the programmer) are
    responsible for calling the appropriate method based on this ID -
    you define the meaning of this ID when you implement Dispatch().

    @param name Function name that will appear in the object's table.
    @param id Unique id you choose for this function.
**/
void LuaObject::RegisterMethod( const char* name, int id )
{
    assert( m_lua && m_luaName ); // must have already registered the object to register a method.

    LuaStackRestore restore( *m_lua ); // set stack restore point

    m_lua->GetGlobal( m_luaName ); // get the table which represents this object

    assert( m_lua->GetType() == LUA_TTABLE );

    m_lua->PushString( name );  // will be used as table key

    m_lua->PushLightUserData( (void*)this ); // Push the this pointer
    m_lua->PushNumber( id );                 // Push the member id for dispatch
    m_lua->PushCClosure( LuaObject::GlobalDispatch, 2 );
    m_lua->SetTable( -3 ); // Stack is { closure, name, table } executes: table[name] = closure
}

/**
    Get the Lua table that represents this object.
    Leaves it on  top of stack.
**/
void LuaObject::GetLuaObject()
{
    m_lua->GetGlobal( m_luaName ); // get the object table
    
    assert( m_lua->GetType() == LUA_TTABLE );
}

/**
    Copy the item ontop of the stack into the Lua table
    representation of this object. (The item at the top
    of the stack is popped.)

    @param name The name of the variable to set in the object table.
**/
void LuaObject::SetLuaValue( const char* name )
{
    assert( m_lua->GetStackSize() > 0 );

    { // begin stack restore scope
        LuaStackRestore restore( *m_lua );

        GetLuaObject();

        m_lua->PushString( name );  // will be used as table key

        m_lua->PushValue( -3 ); // copy value to top of stack
        m_lua->SetTable( -3 ); // enter the value into the table using name as key

    } // end stack retore scope

    m_lua->Pop( 1 );
}

/**
    Get a value from the Lua table representation of this object,
    leaving the value ontop of the stack when done.

    @param name The name (the key in Lua terms) of the value to retrieve.
    
    @return true if the object was retrieved, false if it did not exist
**/
void LuaObject::GetLuaValue( const char* name )
{
    GetLuaObject();

    m_lua->PushString( name );  // will be used as table key

    m_lua->GetTable(-2); // uses key to index table at -2 in the stack (which should be the object's table)

    m_lua->Replace(-2); // overwrite the object table with the retreived value    
}

} // end namespace Lua
