#ifndef LUA_OBJECT_H
#define LUA_OBJECT_H

#include "../fwconfig.h"

extern "C" {
#include <lua.h>
}

namespace Lua
{

// fwd decl
class LuaState;

/**
    Interface for objects that want to register an interface with
    Lua so that they can used in scripts/interactively.

    The method chosen for exposing object-orientation in Lua is one of the
    simplest and slowest, but also the most flexible. When you register an
    object with Lua you create a named table to represent the interface to
    the object. The interface consists of Lua functions which call methods
    in the object and other lua values which can be read/written by the
    object.

    The Lua interface of the object is not the same as its C++ interface,
    and does not even need to be the same for objects of the same type.
    The programmer has complete control over the interface he/she exposes
    on a per object basis.
    
    Because a C++ object in Lua is just a table, if type information is
    required in scripts you will need to add a value (or method) to the
    table to distinguish type.

    Method Call Dispatching:
    
    The dispatch mechanism uses C closures (see Lua reference manual for
    more info). All method calls from Lua go through the static member
    function GlobalDispatch() (Lua can only call static members). This
    function retrieves the correct 'this' pointer (stored as the first
    upvalue). The 'this' pointer is then used to call the virtual
    Dispatch() with an ID indicating the actual member to call. (The ID
    is the second upvalue.)

    The dispatch mechanism has a number of important properties:
      
    1. By implementing Dispatch(), each object is responsible for its own
    method dispatch. It will usually just be a switch statement, but could
    be, for example, context sensitive. This is very flexible.

    2. Methods are registered per object - not per class. Different objects
    of the same type can register different methods, or register the same
    methods under different names.

    3. Dispatch goes through a Lua call and (at least) three native calls
    one of which is a virtual function call so it is relatively slow.

**/
class FW_API LuaObject
{

public:
    LuaObject();
    virtual ~LuaObject();

    void RegisterObject( LuaState* lua, const char* name );
    bool JoinObject( LuaState* lua, const char* name );   
    void UnregisterObject();
    
    void RegisterMethod( const char* name, int id );

    void GetLuaObject();
    void SetLuaValue( const char* name );
    void GetLuaValue( const char* name );

    virtual int Dispatch( lua_State* L, lua_Number id ) = 0;

protected:
    LuaState* GetLua() { return m_lua; };
    
private:
    static int GlobalDispatch( lua_State* L );
    char* m_luaName;
    LuaState* m_lua;
};

} // end namespace Lua

#endif // LUA_OBJECT_H
