#include "LuaState.h"

#include "../thread/Thread.h"

extern "C" {
#include <lauxlib.h>
#include <lualib.h>
}

#include <assert.h>

#ifdef __linux

#include <stdio.h>
#include <string.h>

#else

#define WIN32_MEAN_AND_LEAN
#include <windows.h>

#endif

namespace Lua
{

/**
    Add a sleep function accessible from Lua.

    Expects a single number as argument on lua stack: the time to sleep in milliseconds.

    Returns 0 results.
**/
int LuaState::sleep( lua_State* L )
{
    LuaState lua( L );

    if ( lua.GetStackSize() == 1 && lua.IsNumber() )
    {
        int millisecs = lua.ToInteger();
        GLK::Thread::Sleep( millisecs );
    }
    else
    {
        lua.GetGlobal( "print" );
        lua.PushString( "Error: sleep( millisecs ) requires a single number as argument.\n" );
        lua.Call( 1 );
    }

    return 0;
}

/**
    Function to overide Lua's print function (which will not work without proper stdio).
**/
int LuaState::print( lua_State* L )
{
    const LuaState lua( L );

    char output[128];

    int numParams = lua.GetStackSize();

    for ( int i=1; i<=numParams; ++i )
    {
        const char* write = 0;
        size_t n = 0;

        switch( lua.GetType(i) )
        {
        case LUA_TNIL:
            sprintf( output, "nil" );
            write = output;
            n = strlen( write );
            break;
        case LUA_TNUMBER:
            sprintf( output, "%f", lua.ToDouble(i) );
            write = output;
            n = strlen( write );
            break;

        case LUA_TBOOLEAN:
            if ( lua.ToBool(i) )
            {
                sprintf( output, "true" );
            }
            else
            {
                sprintf( output, "false" );
            }
            write = output;
            n = strlen( write );
            break;

        case LUA_TSTRING:
            write = lua.ToString( i, &n );
            break;

        case LUA_TTABLE:
            sprintf( output, "table: " );
            write = output;
            n = strlen( write );
            break;

        case LUA_TFUNCTION:
            sprintf( output, "function: " );
            write = output;
            n = strlen( write );
            break;
        case LUA_TUSERDATA:
            sprintf( output, "user data: " );
            write = output;
            n = strlen( write );
            break;
        case LUA_TTHREAD:
            sprintf( output, "thread: " );
            write = output;
            n = strlen( write );
            break;
        case LUA_TLIGHTUSERDATA:
            sprintf( output, "light user data: " );
            write = output;
            n = strlen( write );
            break;


        default:
            break;
        }

        if ( write )
        {

#ifdef __linux
            fprintf( stderr, "%s", write );

            if ( i == numParams )
            {
                fprintf(stderr, "\n" );
            }
            else
            {
                fprintf(stderr, "     " );
            }
#else
            WriteConsole( GetStdHandle(STD_ERROR_HANDLE), write, n, 0, 0 );

            if ( i == numParams )
            {
                WriteConsole( GetStdHandle(STD_ERROR_HANDLE), "\n",  1, 0, 0 );
            }
            else
            {
                WriteConsole( GetStdHandle(STD_ERROR_HANDLE), "     ",  4, 0, 0 );
            }
#endif

        }

    }

    return 0;
}

/**
    Creates a new lua state object.

    lua_open() is called in the constructor:
    user should check lua_open() was successful
    by calling LuaState::Ok().

    @param name Give a name to the interpreter: useful for debugging
    purposes in programs which use multiple Lua states.
**/
LuaState::LuaState()
:
    m_lua           (0),
    m_isReference   (false)
{
    m_lua = lua_open();
    if ( m_lua == 0 )
    {
        fprintf( stderr, "\nError in LuaState::LuaState(): not enough memory to create lua state\n" );
    }
}

/**

**/
LuaState::LuaState( lua_State* lua_reference )
:
    m_lua           (lua_reference),
    m_isReference   (true)
{

}

/**
    Destroy lua state object.
    lua_close() is called here provided that the lua
    state object was not created as a reference
    (A reference is created by passing in a lua_State* to the contstructor).
**/
LuaState::~LuaState()
{
    if ( !m_isReference )
    {
        lua_close( m_lua );
    }
}

/**
    Returns true if lua state was created successfully.
**/
bool LuaState::Ok() const
{
    return m_lua != 0;
}

/**
    Return true if the object was created as a
    reference to an already existing lua_State structure,
    false otherwise.
**/
bool LuaState::IsReference() const
{
    return m_isReference;
}

/**
    Open the Lua base libraries.
**/
void LuaState::OpenLibraries()
{
    luaL_openlibs( m_lua );
}

/**
    Execute the script with the given file name.
**/
int LuaState::DoFile( const char* fileName )
{
	return luaL_dofile( m_lua, fileName );
}

/**
    Loads and runs the specified string.
    (s should contain a valid Lua command and be zero terminated).
**/
int LuaState::DoString( const char* s )
{
    assert( s != 0 );

    int rval = luaL_loadstring( m_lua, s );

    if ( rval == 0 )
    {
        rval = lua_pcall( m_lua, 0, LUA_MULTRET, 0 );
    }

    return rval;
}

/**
    Access/query/control Lua's garbage collector (gc).

    @param what Indicates the value/funcitonality we are interested in. what can be any of the folowing:
    - LUA_GCSTOP : Stop gc
    - LUA_GCRESTART : Start gc.
    - LUA_GCCOLLECT : Perform gc (full cycle)
    - LUA_GCCOUNT : amount of memory in use in Kilobytes
    - LUA_GCCOUNTB : remainder of memory in use in bytes/1024
    - LUA_GCSTEP : Incremental steps of garbage collection, data controls num steps. returns 1 if a full cycle was completed
    - LUA_GCSETPAUSE : Set the pause value to data (see luaconf.h).
    - LUA_GCSETSTEPMUL : Set the multiplier value to data 9 see luaconf.h).
    @param data Meaning depends on value of what.
**/
int LuaState::Gc( int what, int data )
{
    return lua_gc( m_lua, what, data );
}

/**
    Gets named global variable from Lua and push it onto stack.
    @param variableName The name of the variable whos value will be retrieved.
**/
void LuaState::GetGlobal( const char* variableName )
{
    lua_getglobal( m_lua, variableName );
}

/**
    Set the named global variable's value to equal whatever is on top of the stack.
    @param variableName The name of the variable whos value will be set.
**/
void LuaState::SetGlobal( const char* variableName )
{
    lua_setglobal( m_lua, variableName );
}

/**
    Return the number of items on the stack.
**/
int LuaState::GetStackSize() const
{
    return lua_gettop( m_lua );
}

/**
    Ensures that there are at least extra free slots on the stack.
    Note, the stack will not increase in size if it is already large enough.
**/
int LuaState::GrowStack( int extra )
{
    return lua_checkstack( m_lua, extra );
}

/**
    Set the top of the stack to specified index.
    If the new top is higher than the current top
    then the extra elements are filled with 'nil'.

    Does not allocate more stack space, use LuaState::GrowStack()
    for this.

    @param index The new top of the stack. If 0 then the stack is cleared.
**/
void LuaState::SetTop( int index )
{
    lua_settop( m_lua, index );
}

/**
    Remove all items from the stack.
**/
void LuaState::ClearStack()
{
    //Pop( GetStackSize() );
    SetTop( 0 );
}

/**
    Push a number onto stack.
**/
void LuaState::PushNumber( lua_Number number )
{
    lua_pushnumber( m_lua, number );
}

/**
    Push a (zero-terminated) string onto the stack.
    Lua creates (or reuses) its own internal copy of
    the string so the data pointed to by s can be
    safely discarded or modified.

    @param s Pointer to zero-terminated string.
**/
void LuaState::PushString( const char* s )
{
    lua_pushstring( m_lua, s );
}

/**
    Push the specified bool onto the stack.
    In lua nil is false while all real-numbers
    are true. (I.e. 0 is true).

    @param b Boolean to push.
**/
void LuaState::PushBoolean( bool b )
{
    lua_pushboolean( m_lua, b );
}

/**
    Push a copy of the item at index onto the top of the stack.
    @param index position of item to be copied to top of stack.
**/
void LuaState::PushValue( int index )
{
    lua_pushvalue( m_lua, index );
}

/**
    Remove the item from the top of the stack.
**/
void LuaState::Pop( int n )
{
    lua_pop( m_lua, n );
}

/**
    Move the item ontop of the stack into a position within
    the stack (shifting up all higher items to create a gap).
    @param index position where the item should be inserted.
**/
void LuaState::Insert( int index )
{
    lua_insert( m_lua, index );
}

/**
    Remove item from stack and shift higher elements to fill gap.

    @param index position of item to be removed. Cannot be a psuedo index.
**/
void LuaState::Remove( int index )
{
    lua_remove( m_lua, index );
}

/**
    Move the item ontop of teh stack to the specified index.
    No items in the stack are shifted so that the item at
    'index' is overwritten.

    @param index position of item to be replaced.
**/
void LuaState::Replace( int index )
{
    lua_replace( m_lua, index );
}

/**
    Determine if the items on stack are equal (can call metamethod for equals).

    @return true if item at index1 equals item at index2, false if they are not
    equal or if either of the indices are invalid.
**/
bool LuaState::Equal( int index1, int index2 )
{
    return ( 1 == lua_equal( m_lua, index1, index2 ) );
}

/**
    Calls the last function that was pushed onto the stack.
    The function and its arguments are automatically popped from the stack.

    @param nargs The number of arguments (which must have been pushed onto the stack AFTER the function)
    @param nresults The number of results wanted or LUA_MULTRET (default) if all return values are wanted.
**/
void LuaState::Call( int nargs, int nresults )
{
    lua_call( m_lua, nargs, nresults );
}

/**
    Calls the last function that was pushed onto the stack in protected mode.
    Same as Call except if there is an error the error message is pushed onto
    the stack and a non-zero value is returned.

    @param nargs The number of arguments (which must have been pushed onto the stack AFTER the function)
    @param nresults The number of results wanted or LUA_MULTRET (default) if all return values are wanted.

    @return Error code: 0 for success or one of the following for failure:
    LUA_ERRRUN	
    LUA_ERRSYNTAX	
    LUA_ERRMEM	
    LUA_ERRERR	
**/
int LuaState::CallProtected( int nargs, int nresults )
{
    return lua_pcall( m_lua, nargs, nresults, 0 );
}

/**
    Register the C function so that it can be called from within lua scripts.
    The function will be a global varaible - functions are variables in Lua.
**/
void LuaState::RegisterCFunction( const char* funcName, lua_CFunction func )
{
    lua_register( m_lua, funcName, func );
}

/**
    Register a C function but specify the upvalues at the same time (creating what is called a closure).
    The function will be a global variable - functions are variables in Lua.
    The upvalues to be used must be at the top of the stack (and will be popped from the stack).

    @param funcName Name of global variable that will identify function in Lua.
    @param func The C function to register.
    @param numUpvalues The number of upvalues to be popped from the stack.
**/
void LuaState::RegisterCClosure( const char* funcName, lua_CFunction func, int numUpvalues )
{
    lua_pushcclosure( m_lua, func, numUpvalues );
    lua_setglobal( m_lua, funcName );
}

/**
    Pushes a C function onto the Lua stack.
    @param func The C function to push.
**/
void LuaState::PushCFunction( lua_CFunction func )
{
    lua_pushcfunction( m_lua, func );
}

/**
    Pushes a C-closure onto the stack.
    
    A C-closure consists of a C-function and a number of 'upvalues'.
    Upvalues are values uniquely available to the C-function when it
    is called from Lua.

    Before calling PushCClosure the stack should be as follows:

    { upvalue_n, ..., upvalue_2, upvalue_1, 'anything', ... }
    
    Calling PushCClosure( func, n ) then creates the C-closure and
    pops all the upvalues leaving:
    
    { 'newly created c-closure', 'anything', ... }

    @param func The C-function to be pushed.
    @param numUpvalues The number of upvalues to be pushed (these will be popped from the stack).
**/
void LuaState::PushCClosure( lua_CFunction func, int numUpvalues )
{
    lua_pushcclosure( m_lua, func, numUpvalues );
}

/**
    Return the upvalue at the specified stack index
    (Do not use to retrieve user data (C-pointers), use GetUserData instead).
**/
lua_Number LuaState::GetUpvalue( int n ) const
{
    return lua_tonumber( m_lua, lua_upvalueindex( n ) );
}

/**
    Retrieve user data from stack index.
**/
void* LuaState::GetUserData( int n ) const
{
    return lua_touserdata( m_lua, n );
}

/**
    Register a static function which receives this pointer as solitary upvalue.
    
    @param funcName Name of global variable that will identify function in Lua.
    @param func The C function to register.
**/
void LuaState::RegisterFunction( const char* funcName, lua_CFunction func, void* object )
{
    PushLightUserData( object ); // Push the this pointer for some object
    RegisterCClosure( funcName, func, 1 );
}

/**
    Create a new (empty) table and push it ontop of the stack.
**/
void LuaState::NewTable()
{
    lua_newtable( m_lua );
}

/**
    Pushes the value t[k] onto the stack.
    t is the value at 'index' and k is the value at the top of the stack.

    This function replaces 'k' on the stack with the result of th table lookup.
    May trigger a metamethod (see Lua reference manual).

    @param index Position in stack to convert. Default (-1), checks the top of the stack.
**/
void LuaState::GetTable( int index )
{
    lua_gettable( m_lua, index );
}

/**
    Puts the value at the top of the stack into the table residing at index using the key
    which is second down on the stack.

    The key and value (k and v) are both popped from the stack.

    If the stack is { v, k, ... } and table at 'index' is t.
    then this call does t[k] = v, removing v and k from the stack.

    @param index Stack position where table resides.
**/
void LuaState::SetTable( int index )
{
    lua_settable( m_lua, index );
}

/**
    Used when iterating over a table.

    Pops a key from the top of lua's stack, then
    pushes a key/value pair or the next item in the table
    to the top of the stack.

    @index Stack index for the table.
    @return false if there are no more items, true otherwise.
**/
bool LuaState::Next( int index )
{
    return lua_next( m_lua, index ) != 0;
}

/**
    Converts the item on top of the stack to an integer and returns it.
    @param index Position in stack to convert. Default (-1), checks the top of the stack.
**/
int LuaState::ToInteger( int index ) const
{
   return static_cast<int>( lua_tointeger( m_lua, index ) );
}

/**
    Convert item on top of stack to a double and return it.
    @param index Position in stack to convert. Default (-1), checks the top of the stack.
**/
double LuaState::ToDouble( int index ) const
{
    return lua_tonumber( m_lua, index );
}

/**
    Convert item on top of stack to a bool and return it.
    @param index Position in stack to convert. Default (-1), checks the top of the stack.
**/
bool LuaState::ToBool( int index ) const
{
    return ( lua_toboolean( m_lua, index ) == 1 );
}

/**
    Convert the item on top of the stack to a string.

    The  pointer returned is only valid as long as
    the string exists in Lua (i.e. it remains in the
    stack or is stored in a table).

    If the string is popped from the stack it can be garbage
    collected at any time so the pointer becomes invalid.
    @param index Position in stack to return as string.
    Default (-1), checks the top of the stack.
    @param length Pointer to int where string length will be stored (if it is not null)
**/
const char* LuaState::ToString( int index, size_t* length ) const
{
    const char* rval = lua_tolstring( m_lua, index, length );
    return rval;
}

/**
    Returns true if the item at index of the stack is a number.
    @param index The position on the stack of the value we are interested in checking.
    Default (-1), checks the top of the stack.
**/
bool LuaState::IsNumber( int index ) const
{
    return lua_isnumber( m_lua, index ) != 0;
}

/**
    Return true if the item at index of the stack is a string.
    @param index The position on the stack of the value we are interested in checking.
    Default (-1), checks the top of the stack.
**/
bool LuaState::IsString( int index ) const
{
    return lua_isstring( m_lua, index ) != 0;
}

/**
    Return true if the item at index is nil.
    @param index The position on the stack of the value we are interested in checking.
    Default (-1), checks the top of the stack.
**/
bool LuaState::IsNil( int index ) const
{
    return lua_isnil( m_lua, index ) != 0;
}

bool LuaState::IsCFunction( int index ) const
{
    return (1 == lua_iscfunction( m_lua, index ) );
}

bool LuaState::IsFunction( int index ) const
{
    return (1 == lua_isfunction( m_lua, index ) );
}

bool LuaState::IsLightUserData( int index ) const
{
    return (1 == lua_islightuserdata( m_lua, index ) );
}

bool LuaState::IsTable( int index ) const
{
    return (1 == lua_istable( m_lua, index ) );    
}

bool LuaState::IsBool( int index ) const
{
    return (1== lua_isboolean( m_lua, index ) );
}

/**
    Return the type of the item at index of the stack.
    @param index The position on the stack of the value we are interested in checking.
    @return Integer with one of the following constant values defined in lua.h:
    LUA_TNIL, LUA_TNUMBER, LUA_TBOOLEAN, LUA_TSTRING, LUA_TTABLE, LUA_TFUNCTION,
    LUA_TUSERDATA, LUA_TTHREAD, and LUA_TLIGHTUSERDATA
**/
int LuaState::GetType( int index ) const
{
    return lua_type( m_lua, index );
}

/**
    Convenience function for indexing a table.
    By default the table ontop of the stack is indexed.
    After returning the contents of the t[key] will be
    ontop of the stack. (Where t is the table ontop of
    lua's stack)

    @key Key to index table
    @returns Reference to lua state to faclitate the following notation: lua[ "item" ].ToInteger()
**/
LuaState& LuaState::operator[] ( const char* key )
{
    lua_pushstring( m_lua, key );
    GetTable( -2 );
    return *this;
}

} // end namespace Lua

