#ifndef LUA_STATE_H
#define LUA_STATE_H

#include "../fwconfig.h"

#include "../utility/String.h"
#include "../utility/GlobalDefs.h"

extern "C" {
#include <lua.h>
}

// Macro can be used to register a function using the literal function name
#define LUA_REGISTER_FUNCTION(l,f) ( (l)->RegisterCFunction((#f),(f)) )

// Macro can be used to redefine a LuaState instance's print() function
#define LUA_REGISTER_PRINT(l) ( (l)->RegisterCFunction( "print", Lua::LuaState::print ) )

// Macro used to add sleep function to lua
#define LUA_REGISTER_SLEEP(l) ( (l)->RegisterCFunction( "sleep", Lua::LuaState::sleep ) )

namespace Lua
{

/**
    Class which encapsulates a lua_State struct. All Lua programming within the
    framework should utilise this class rather than using lua_State structures
    directly.

    Many of the methods are simple wrappers for functions from the Lua C API.
**/
class FW_API LuaState
{
    friend class LuaObject;

public:
    static int sleep( lua_State* L );
    static int print( lua_State* L );

    LuaState();
    explicit LuaState( lua_State* lua_reference );
    virtual ~LuaState();

    template <class T> void PrintError( int err, T& printer );

    // Query LuaState object
    bool Ok() const;
    bool IsReference() const;

    // Executing commands
    void OpenLibraries();
    int DoFile( const char* fileName );
    int DoString( const char* string );

    // Garbage Collection
    int Gc( int what, int data );

    // Lua Stack Manipulation
    void GetGlobal( const char* variableName );
    void SetGlobal( const char* variableName );
    int GetStackSize() const;
    int GrowStack( int extra );
    void SetTop( int index );
    void ClearStack();
    
    void PushNil() { lua_pushnil( m_lua ); };
    void PushNumber( lua_Number number );
    void PushString( const char* s );
    void PushBoolean( bool b );
    void PushValue( int index );
    void PushLightUserData( void* p ) { lua_pushlightuserdata( m_lua, p ); };

    void Pop( int n );
    void Insert( int index );
    void Remove( int index );
    void Replace( int index );
    bool Equal( int index1, int index2 );

    // Functions and closures
    void Call( int nargs, int nresults = LUA_MULTRET );
    int CallProtected( int nargs, int nresults = LUA_MULTRET );   
    void RegisterCFunction( const char* funcName, lua_CFunction func );
    void RegisterCClosure( const char* funcName, lua_CFunction func, int numUpvalues );
    void PushCFunction( lua_CFunction func );
    void PushCClosure( lua_CFunction func, int n );
    lua_Number GetUpvalue( int n ) const;
    void* GetUserData( int n ) const;

    void RegisterFunction( const char* funcName, lua_CFunction func, void* oject );

    // Tables
    void NewTable();
    void GetTable( int index = -1 );
    void SetTable( int index );
    bool Next( int index = -1 );

    // Manipulating Lua types
    int         ToInteger   ( int index = -1 ) const;
    double      ToDouble    ( int index = -1 ) const;
    bool        ToBool      ( int index = -1 ) const;
    const char* ToString    ( int index = -1, size_t* length = 0 ) const;
    bool IsNumber( int index = -1 ) const;
    bool IsString( int index = -1 ) const;
    bool IsNil( int index = -1 ) const;
    bool IsCFunction( int index = -1 ) const;
    bool IsFunction( int index = -1 ) const;
    bool IsLightUserData( int index = -1 ) const;
    bool IsTable( int index = -1 ) const;
    bool IsBool( int index = -1 ) const;
    int  GetType ( int index = -1 ) const;

    // experimental
    LuaState& operator [] ( const char* key );

protected:

private:
    SET_UNCOPYABLE( LuaState )

    lua_State* m_lua;
    bool m_isReference;
};

/**
    A simple class which allows us to restore the lua stack at the end
    of scope, without keeping track of what we did to it in between.
    (Same design principle as a Mutex locking class).

    Only checks stack size, so can restore pops/pushes but not
    insertions/deletions or moer complex stack manipulation etc.
**/
class LuaStackRestore
{
public:
    explicit LuaStackRestore( LuaState& lua ) : m_lua(lua), s( lua.GetStackSize() ) {};
    ~LuaStackRestore() { m_lua.Pop( m_lua.GetStackSize() - s ); };

private:
    SET_UNCOPYABLE( LuaStackRestore )

    LuaState& m_lua;
    int s;
};

/**
    Template for error printing. If err is 0 then nothing is printed.

    @param err Error code returned from Lua.
    @param w Template class which must support 'Printf' and 'WriteText' interfaces.
**/
template <class T>
void LuaState::PrintError( int err, T& w )
{
    switch (err)
    {
        case LUA_YIELD:
            w.Printf( "Lua Yield\n" );
            break;
        case LUA_ERRSYNTAX:
            w.Printf( "Syntax Error\n" );
            break;
        
        case LUA_ERRMEM:
            w.Printf( "Memory Allocation Error\n" );    
            break;
        case LUA_ERRRUN:
            {
                const char* s = 0;
                size_t l = 0;
                if ( GetStackSize() > 0 && IsString() )
                {
                    s = ToString( -1, &l );
                    w.Write( s, l );
                    w.Write( "\n", 1 );
                    Pop(1);
                }
                else
                {
                    w.Printf( "Unknown Runtime Error\n" );
                }
            }
            break;

        case LUA_ERRERR:
            w.Printf( "Error Handler Failed\n" );    
            break;

        default:
            break;
    }
}

} // end namespace Lua

#endif // LUA_STATE_H

