#ifdef USE_MEM_ANALYSER

#include "MemoryAnalyser.h"

#include <malloc.h>
#include <new>
#include <assert.h>

#include "../thread/MutexLock.h"

namespace GLK
{

#ifdef _MSC_VER

inline void* MY_ALLOC( size_t size, size_t alignment )
{
    return _aligned_malloc( size, alignment );
}

inline void MY_FREE( void* p )
{
    _aligned_free( p );
}

#else

inline void* MY_ALLOC( size_t size, size_t alignment )
{
    return malloc( size );
}

inline void MY_FREE( void* p )
{
    free( p );
}

#endif

MemoryAnalyser* MemoryAnalyser::m_singleton = 0;

MemoryAnalyser* MemoryAnalyser::GetSingleton()
{
    if ( !m_singleton )
    {
        m_singleton = new MemoryAnalyser();
    }

    return m_singleton;
}

/**
    Assign the single instance if this is the first MemoryAnalyser.
    It is an error to create more than one instance of MemoryAnalyser.
**/
MemoryAnalyser::MemoryAnalyser()
:
    m_allocs    (0),
    m_bytes     (0),
    m_maxBytes  (0)
{
    MutexLock locker( m_lock );

    if ( !m_singleton )
    {
        m_singleton = this;
    }
    else
    {
        assert(0); // trying to create more than one MemoryAnalyser.
    }

    for (unsigned int i=0;i<m_maxRecords;++i)
    {
        m_record[i].p = 0;
    }
}

MemoryAnalyser::~MemoryAnalyser()
{

}

void* MemoryAnalyser::operator new ( size_t size )
{
    return MY_ALLOC( size, 4 );
}

void MemoryAnalyser::operator delete ( void* p )
{
    return MY_FREE( p );
}

void* MemoryAnalyser::Allocate( size_t size )
{
    MutexLock locker( m_lock );

    void* p = MY_ALLOC( size, m_alignment );

    while ( p == 0 )
    {
        std::new_handler nh = std::set_new_handler(0);
        if (nh)
        {
            std::set_new_handler( nh );
            (*nh)();
        }
        else
        {
            throw std::bad_alloc();
        }
    }

    m_record[m_allocs].p = p;
    m_record[m_allocs].size = size;
    m_allocs++;
    m_bytes += size;

    if ( m_bytes > m_maxBytes )
    {
        m_maxBytes = m_bytes;
    }

    return p;
}

void MemoryAnalyser::Free( void* p )
{
    if ( p != 0 )
    {
        MutexLock locker( m_lock );

        // Search for pointer in recorded allocations.
        // Search begins at most recent for efficiency.
        int i = m_allocs;
        if ( i > m_maxRecords)
        {
            i = m_maxRecords;
        }
        i -= 1;
        while ( i >= 0 && m_record[i].p != p )
        {
            --i;
        }

        if ( i >= 0 )
        {
            unsigned int iu = i; // fixes signed/unsigned mismatch warning
            m_record[iu].p = 0;
            m_bytes -= m_record[i].size;

            // Found a record so shift down all later allocations to fill hole (could be slow!)
            while ( iu < m_allocs && iu+1 < m_maxRecords )
            {
                m_record[iu] = m_record[iu+1];
                iu++;
            }
        }

        m_allocs--;
        MY_FREE( p );
    }
}

} // end namespace Fw

///////////////////////////////////////
/// Overides for global new/delete: ///
///////////////////////////////////////

void* operator new ( size_t size ) throw( std::bad_alloc )
{
    Fw::MemoryAnalyser* mem = Fw::MemoryAnalyser::GetSingleton();
    return mem->Allocate( size );
}

void operator delete ( void* p ) throw()
{
    Fw::MemoryAnalyser* mem = Fw::MemoryAnalyser::GetSingleton();
    mem->Free( p );
}

void* operator new [] ( size_t size ) throw( std::bad_alloc )
{
    Fw::MemoryAnalyser* mem = Fw::MemoryAnalyser::GetSingleton();
    return mem->Allocate( size );
}

void operator delete [] ( void* p ) throw()
{
    Fw::MemoryAnalyser* mem = Fw::MemoryAnalyser::GetSingleton();
    mem->Free( p );
}

#endif

