#ifndef MEMORY_ANALYSER_H
#define MEMORY_ANALYSER_H

#include "../thread/Mutex.h"

namespace GLK
{

/**
    Singleton class which can be used in
    dynamic memory allocation analysis and debugging.

    Overides of global new/delete are provided that
    use this singleton to log dynamic memory usage.
**/
class MemoryAnalyser
{
public:
    static MemoryAnalyser* GetSingleton();

    void* Allocate( size_t size );
    void Free( void* p );

    // Memory analyser needs its own new/delete so that it doesn't try to log its own allocation:
    void* operator new ( size_t );
    void operator delete ( void* );

protected:
    
private:
    // private constructors and destructors because (the) single
    // MemoryAnalyser will be implicitly created by the first new.
    MemoryAnalyser();
    ~MemoryAnalyser();

    static MemoryAnalyser* m_singleton;

    //! Byte alignment for all allocations.
    const static size_t m_alignment = 4;
    //! Maximum number of allocations that can be fully recorded.
    const static size_t m_maxRecords = 4096;

    Mutex m_lock;

    //! Total number of allocations.
    size_t m_allocs;

    //! Total number of bytes allocated.
    size_t m_bytes;

    //! Maximum number of bytes ever allocated
    size_t m_maxBytes;

    struct Record
    {
        void* p;
        size_t size;
    };

    Record m_record[m_maxRecords];
};

} // end namespace GLK

#endif // MEMORY_ANALYSER_H
