#include "Barrier.h"

#include "MutexLock.h"

namespace GLK
{

/**
	Set up a new barrier which will synchronise the specified number of threads.
	
	@param count The number of threads which must reach the barrier before any is allowed to continue.
**/
Barrier::Barrier( unsigned int count )
:
	m_count		( count ),
	m_stored	( count ),
	m_cycle		( 0 )
{
	assert (m_count>0);
}

/**
	Destroys a barrier. It is an error to destroy a
	barrier that has threads waiting on it.
**/
Barrier::~Barrier()
{
	assert( m_count == m_stored ); // Debug - check there are no threads wating at the barrier.
}

/**
	Makes the calling thread wait at the barrier.
	Once all threads have reached the barrier the
	barrier resets itself (starts a new cycle).
	
	@return A boolean set to true if the caller was last to arrive at the barrier, false otherwise.
**/
bool Barrier::Wait()
{
	bool retval;

	MutexLock locker( m_lock );

	if ( --m_count == 0 )
	{
		m_count = m_stored;		// reset the count.
		m_cycle++;				// move onto next cycle.
		m_cond.WakeAll();		// wake all other threads at the barrier.
		retval = true;			// caller was last to arrive at the barrier.
	}
	else
	{
		unsigned int cycle = m_cycle; // remember the cycle when caller began waiting at the barrier
		while ( cycle == m_cycle )
		{
			m_cond.Wait( m_lock ); // waits until barrier moves onto next cycle
		}
		retval = false; // caller was not last to arrive at the barrier
	}

	return retval;
}

} // namespace GLK
