#ifndef GLK_BARRIER_H
#define GLK_BARRIER_H

#include <assert.h>

#include "Mutex.h"
#include "ConditionVariable.h"

#include "../utility/GlobalDefs.h"

namespace GLK
{

/**
	Implements the barrier synchronisation primitive.
	A barrier will block all threads that use it as a
	wait point until a set number of threads have 'reached'
	(that is - 'are waiting on') the same barrier.

	This allows us, for example, to stop any member of a
	group of threads from continuing beyond some point
	until all threads have reached that point.

	Once all threads have reached the barrier it is
	reset so that it can be used repeatedly. Each reset
	of the barrier completes one 'cycle'. The barrier
	keeps count of the number of cycles it completes.
**/
class FW_API Barrier
{
public:

    static void Test();

	Barrier( unsigned int count );
	~Barrier();

	//! Return the number of cycles this barrier has completed. No mutex is locked so the actual cycle number might be larger.
	unsigned int NumCycles() const { return m_cycle; };

	//! Return the number of threads waiting at the barrier. No mutex is locked so the actual number of waiters might be smaller.
	unsigned int NumWaiting() const { return m_stored-m_count; };

	bool Wait();

private:
	SET_UNCOPYABLE(Barrier);

	Mutex m_lock;
	ConditionVariable m_cond;
	unsigned int m_count;			//<! Stores the number of threads we are currently waiting for.
	const unsigned int m_stored;	//<! Record the original count so we can reset the barrier.
	unsigned int m_cycle;			//<! The number of cycles executed by the barrier.
};

} // end of namespace GLK

#endif // end BARRIER_H
