#ifndef FW_MESSAGE_QUEUE_H
#define FW_MESSAGE_QUEUE_H

#include "../utility/GlobalDefs.h"

#include "Mutex.h"
#include "MutexLock.h"
#include "Semaphore.h"

namespace GLK
{

/**
	Template for first-in-first-out (FIFO) message queues.
	These queues can be used for inter-thread communication.

	The message type is a template parameter, so message
	queues for specific purposes should specialise this class.

	The copy constructor for the class T is invoked for every
	message which is written or read.
**/
template <class T>
class FW_API MessageQueue
{
public:
    
    static void Test( void );

	explicit MessageQueue( unsigned int size );
	virtual ~MessageQueue();

	//! Allows us to query the size of message without knowing the template type.
	static size_t SizeOfMessage() { return sizeof(T); };

	void Read( T* pMsgs, unsigned int n=1 );
	void Write( const T* pMsgs, unsigned int n=1 );

    bool TryWrite( const T* pMsgs, unsigned int n=1 );
    bool TryRead( T* pMsgs, unsigned int n=1 );
    void WaitMessage();    

	//! Return the number of messages in the buffer. (Not atomic).
	unsigned int NumFilled() { return m_count; };

	//! Return the number of free slots in the buffer. (Not atomic).
	unsigned int NumFree() { return m_maxMsgs-m_count; };

protected:
	bool IsFull() const;
	bool IsEmpty() const;

private:
	SET_UNCOPYABLE( MessageQueue )

    void WritePostAcquire( const T* pMsgs, unsigned int n );
    void ReadPostAcquire( T* pMsgs, unsigned int n );

	Mutex	  m_lock;
	Semaphore m_slotsFree;
	Semaphore m_slotsFilled;

	T* m_pMsg;
	const unsigned int m_maxMsgs;
	unsigned int m_count;	        // Can be accessed (red-only) in non-thread safe way but is deliberately non-volatile
	unsigned int m_first;			// first message written to the queue
	unsigned int m_last;			// last message written to the queue
};

/**
	Construct a message queue which can hold (at most) size messages.
	@param size Number of messages which can be buffered in this queue.
**/
template <class T>
MessageQueue<T>::MessageQueue( unsigned int size )
:
	m_slotsFree		(size),
	m_slotsFilled	(size),
    m_pMsg			(0),
	m_maxMsgs		(size),
	m_count			(0),
	m_first			(0),
	m_last			(0)
{
	MutexLock lock( m_lock );
	m_slotsFilled.Acquire(size);
	m_pMsg = new T[ m_maxMsgs ];
}

/**
	Atomically releases memory allocated for the message buffer.
**/
template <class T>
MessageQueue<T>::~MessageQueue()
{
	MutexLock lock( m_lock );
	delete [] m_pMsg;
}

/**
	Read the next n messages from the queue. Blocks if there are
	not n messages in the queue. Modifications to the message
	buffer are performed atomically.

	The pointer pMsgs must point to space for at least n messages.

	@param pMsgs Pointer to storage for copies of messages we read.
	@param n The number of messages to read.
**/
template <class T>
void MessageQueue<T>::Read( T* pMsgs, unsigned int n )
{
    assert( n <= m_maxMsgs );

	m_slotsFilled.Acquire(n);
    ReadPostAcquire( pMsgs, n );
}

/**
	Write n messages to the queue. Blocks if there is not space
	for n messages in the queue. Modifications to the message
	buffer are performed atomically.

	@param pMsgs Pointer to messages: must contain at least n messages.
	@param n The number of messages to write.
**/
template <class T>
void MessageQueue<T>::Write( const T* pMsgs, unsigned int n )
{
    assert( n <= m_maxMsgs );

	m_slotsFree.Acquire(n);
	WritePostAcquire( pMsgs, n );
}

/**
	Try to read the next n messages from the queue but return
	immediately if there are not n messages in the queue.
	Modifications to the message buffer are performed atomically.

	The pointer pMsgs must point to space for at least n messages.

	@param pMsgs Pointer to storage for copies of messages we read.
	@param n The number of messages to read.

	@return True if n messages were read, false if no messages were read.
**/
template <class T>
bool MessageQueue<T>::TryRead( T* pMsgs, unsigned int n )
{
    assert( n <= m_maxMsgs );
    bool rval;

	if ( m_slotsFilled.TryAcquire(n) )
	{
        ReadPostAcquire( pMsgs, n );
        rval = true;
    }
    else
    {
        rval = false;
    }

    return rval;
}

/**
	Try and write n messages to the queue but return immediately if there
	is not space for n messages in the queue. Modifications to the message
	buffer are performed atomically.

	@param pMsgs Pointer to messages: must contain at least n messages.
	@param n The number of messages to write.

	@return True if n messages were written, false if no messages were written.
**/
template <class T>
bool MessageQueue<T>::TryWrite( const T* pMsgs, unsigned int n )
{
    assert( n <= m_maxMsgs );
    bool rval;

	if ( m_slotsFree.TryAcquire(n) )
	{
        WritePostAcquire( pMsgs, n );
        rval = true;
    }
    else
    {
        rval = false;
    }

    return rval;
}

/**
    Wait for the queue to become not empty.

    The caller sleeps until the queue's not-empty condition is signalled.
**/
template <class T>
void MessageQueue<T>::WaitMessage()
{
    m_slotsFilled.Acquire();
    m_slotsFilled.Release();
}

/**
    Atomically writes to the message queue. Assumes that the speciied number of resources have been
    successfully acquired. This avoids duplicating code between Write and TryWrite.

    The specified number of resources are released before exiting so the caller
    does not need to do so.
**/
template <class T>
void MessageQueue<T>::WritePostAcquire( const T* pMsgs, unsigned int n )
{
    MutexLock lock(m_lock);

    m_count += n;
    for ( unsigned int i=0; i<n; ++i )
    {
        m_pMsg[m_last] = *pMsgs;
        m_last++;
        pMsgs++;
        m_last %= m_maxMsgs;
    }

    m_slotsFilled.Release(n);
}

/**
    Atomically reads from the message queue. Assumes that the speciied number of resources have been
    successfully acquired. This avoids duplicating code between Read and TryRead.

    The specified number of resources are released before exiting so the caller
    does not need to do so.
**/
template <class T>
void MessageQueue<T>::ReadPostAcquire( T* pMsgs, unsigned int n )
{
	MutexLock lock( m_lock );

	m_count -= n;
	for ( unsigned int i=0; i<n; ++i )
	{
		*pMsgs = m_pMsg[m_first];
		m_first++;
		pMsgs++;
		m_first %= m_maxMsgs;
	}

	m_slotsFree.Release(n);
}

/**
	Check if the queue is full. This test is not atomic.
	@return true if the queue is full.
**/
template <class T>
bool MessageQueue<T>::IsFull() const
{
	// if the first message is 1 index ahead of the last message then the queue is full
	int diff	= m_last - m_first;
	bool rval	= diff == -1;
	rval	   |= diff == -m_maxMsgs; // need to check for wrap around case as well
	return rval;
}

/**
	Check if the queue is empty. This test is not atomic.
**/
template <class T>
bool MessageQueue<T>::IsEmpty() const
{
	return m_first == m_last;
}

} // END namespace GLK

#endif // end MESSAGE_QUEUE_H
