#include "Semaphore.h"

#include "MutexLock.h"

namespace GLK
{

/**
  Constructs a semaphore with a specified number of resources.
  @param count The number of resources available from this semaphore.
 **/
Semaphore::Semaphore( int count )
:
	m_count     ( count ),
    m_required  ( 0 )
{
	assert( m_count >= 0 );
}

/**
    No action to take in destructor.
    In debug mode we do a check and assert if there were
    any threads waiting on the semaphore when it was destroyed.
**/
Semaphore::~Semaphore()
{
    assert( m_count >= 0 ); // check there are no waiters
}

/**
    Take specified number of resources.
    @param n Number of resources to acquire.
**/
void Semaphore::Acquire( int n )
{
	MutexLock lock( m_lock );

    m_required += n; // indicate the number of resources required

	while ( m_count < n )
	{
		m_cond.Wait( m_lock ); // wait for more resources
	}

	m_count -= n;    // take the resources
    m_required -= n; // resources are no longer required
}

/**
    Try to take the specified number of resources but return immediately if they are unavailable.

    @param n Number of resources to acquire.
    @return true if the resources were available, false otherwise.
**/
bool Semaphore::TryAcquire( int n )
{
    MutexLock lock( m_lock );
    bool rval;

    if ( m_count >= n )
    {
        m_count -= n;
        rval = true;
    }
    else
    {
        rval = false;
    }

    return rval;
}

/**
  Give back specified number of resources.
  @param n Number of resources to release.
**/
void Semaphore::Release( int n )
{
	MutexLock lock( m_lock );

	m_count += n;

	if ( m_required != 0 )
	{
        // there were threads waiting so wake one up
		m_cond.WakeOne();
	}
}

} // namespace GLK

