#ifndef GLK_POSIX_CONDITION_VARIABLE_H
#define GLK_POSIX_CONDITION_VARIABLE_H

#include <pthread.h>
#include <time.h>

#ifdef WIN32
#include <sys/timeb.h>
#else
#include <sys/errno.h>
#include <sys/time.h>
#endif

#include "../../utility/GlobalDefs.h"

namespace GLK
{

/**
	Implements the condition variable synchronisation primitive.
	
    Currently just a wrapper for pthread_cond_t.

	Features:
		- Uncopyable.	
**/
class FW_API ConditionVariable
{
public:
	//! Creates and initialises a condition variable.
	ConditionVariable() { pthread_cond_init( &m_cond, 0 ); };
	
	//! Destroys the condition variable. It is an error to destroy a condition variable on which a thread is waiting.
	~ConditionVariable() { pthread_cond_destroy( &m_cond ); };

	inline bool Wait( Mutex& m );
	inline bool TimedWait( Mutex& m, unsigned long int millisecs );
	inline void WakeOne();
	inline void WakeAll();

private:
	SET_UNCOPYABLE( ConditionVariable );

	pthread_cond_t m_cond;
};

/**
	Have the calling thread block until the condition variable is
	signalled.

	First, the specified mutex is unlocked, then calling thread sleeps until
	the condition variable is signalled.  If the mutex is not already
	locked then the function immediately returns false.

	Otherwise the thread blocks until the condition is signalled, at
	which point the mutex is re-locked and the function returns true.

	@param mutex A (locked) mutex that will be used to protect the condition.	
	
	@return False in case of an error (if the mutex is not owned by the calling thread or is not locked), true otherwise.
**/
bool ConditionVariable::Wait( Mutex& mutex )
{
	int val = pthread_cond_wait( &m_cond, &mutex.m_mutex );
	assert( val != EINVAL );
	return val == 0;
}

/**
	Same as Wait( Mutex& mutex ) but only waits up to a specified amount of time.

    In debug mode any error causes an assertion to fail.

	@param mutex A (locked) mutex that will be used to protect the condition.
	@param millisecs Maximum time to wait in microseconds.

	@return false in case of timeout, true otherwise.
**/
bool ConditionVariable::TimedWait( Mutex& mutex, unsigned long int millisecs )
{
#ifdef WIN32
    _timeb time;
	timespec ts;
	_ftime_s( &time );
	unsigned long int sec  = millisecs / 1000000;
	unsigned long int usec = millisecs % 1000000;
	ts.tv_sec  = static_cast<long>( time.time + sec );
	ts.tv_nsec = (time.millitm/1000) + (usec/1000);
#elif defined __linux
	timeval tv;
	timespec ts;
	gettimeofday( &tv, 0 );
	unsigned long int sec  = millisecs / 1000000;
	unsigned long int usec = millisecs % 1000000;
	ts.tv_sec  = tv.tv_sec + sec;
	ts.tv_nsec = (tv.tv_usec/1000) + (usec/1000);
#endif

    int err = pthread_cond_timedwait( &m_cond, &mutex.m_mutex, &ts );
    assert( err == 0 || err == ETIMEDOUT );
    return err != ETIMEDOUT;
}

/**
	Wake at least one thread that is waiting on this condition.
**/
void ConditionVariable::WakeOne()
{
	pthread_cond_signal( &m_cond );
}

/**
	Wake all threads that are waiting on this condition.
**/
void ConditionVariable::WakeAll()
{
	pthread_cond_broadcast( &m_cond );
}

} // END namespace GLK

#endif // GLK_POSIX_CONDITION_VARIABLE_H
