#ifndef GLK_POSIX_MUTEX_H
#define GLK_POSIX_MUTEX_H

#include "../../fwconfig.h"

#include <pthread.h>
#include <assert.h>

#include "../../utility/GlobalDefs.h"
#include "Thread.h"

#ifndef WIN32
#include <sys/errno.h>
#endif

namespace GLK
{

class MutexLock;

/**
   Implements the Mutex primitive for creating critical sections/mutual exclusion between threads.
   Note this type of mutex can not be used for inter-process synchronisation. It only manages
   inter-thread locking within a process.

   Currently just a wrapper for pthread_mutex_t.
**/
class FW_API Mutex
{
    friend class ConditionVariable;

public:

    enum Mode
    {
        Recursive,
        NonRecursive
    };

    static void Test();

    inline Mutex( Mode mode = NonRecursive );
    inline ~Mutex();

    inline void Lock();
    inline void Unlock();
    inline bool TryLock();

protected:
    pthread_mutex_t m_mutex;

private:
    SET_UNCOPYABLE(Mutex)
};

/**
    The constructor initialises the mutex.
    Mutex is initially unlocked.

    A mutex can be Recursive or NonRecursive.
    A Recursive mutex can be locked multiple times (but only by the same thead).
    It mus be unlocked the same numer of times before it can be locked by a
    different thread. If a thread attempts to relock a non-recursive mutex then
    the behaviour is undefined but it will possibly block forever.

    @param mode Indicates whether the mutex should be recursive or not.
**/
Mutex::Mutex( Mode mode )
{
    switch (mode)
    {
        case Recursive:
        {
            pthread_mutexattr_t attr;
            pthread_mutexattr_init( &attr );
            pthread_mutexattr_settype( &attr, PTHREAD_MUTEX_RECURSIVE );
            pthread_mutex_init( &m_mutex, &attr );
            pthread_mutexattr_destroy( &attr );
            break;
        }
        case NonRecursive:
        {
            pthread_mutex_init( &m_mutex, 0 );
            break;
        }
    }
}

/**
   Destroy the mutex.
   If the mutex is locked unlock it.
**/
Mutex::~Mutex()
{
    int val = pthread_mutex_destroy( &m_mutex );
    assert( val != EBUSY ); // mutex was locked or referenced by a condition variable
                            // (other errors should not be possible - would have already asserted)
}

/**
   Locks the mutex.
   If the mutex is already locked, call will block until it is available for locking.
**/
void Mutex::Lock()
{
    int err = pthread_mutex_lock( &m_mutex );
    assert( err == 0 );
}

/**
   Unlocks the mutex.
**/
void Mutex::Unlock()
{
    int err = pthread_mutex_unlock( &m_mutex );
    assert( err == 0 );
}

/**
   Try and lock a mutex but do not block if it fails.
   @return True if locking suceeds, false otherwise.
**/
bool Mutex::TryLock()
{
    bool rval = false;

    if ( pthread_mutex_trylock( &m_mutex ) == 0 )
    {
        rval = true;
    }

    return rval;
}

} // end namespace GLK

#endif // GLK_POSIX_MUTEX_H

