#include "Thread.h"

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#else
#include <unistd.h>
#include <sys/time.h>
#endif

#include <assert.h>
namespace GLK
{

/**
	Utility function used with pthread to launch a Thread object's Run() method.
**/
void* Thread::ThreadStarter( void* arg )
{
	pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, 0 );

	Thread* thread    = static_cast<Thread*>(arg);

	thread->m_running = true;

	thread->Run();

	thread->m_running = false;
	thread->m_termReq = false;
	return 0;
}

/**
	Create a new thread.
	Threads do not execute until Start() is called.
**/
Thread::Thread()
:
	m_running (false),
	m_termReq (false)
{
}

/**
	Destructor implicitly stops thread.
**/
Thread::~Thread()
{
	Stop();
}

/**
	Begins concurrent execution of Run() method.

	@return true if execution began successfully, false otherwise.
**/
void Thread::Start()
{
    int err = pthread_create( &m_id, 0, ThreadStarter, (void*)this);
    assert( err == 0 );
}

/**
	Calling thread requests the thread to terminate and then blocks until it does so.

	If the thread ignores your request then you will potentially be waiting forever.

	The request can only be cleared by a thread returning from its Run() method, so a request cannot be 'missed', just ignored (accidentally or on purpose).
**/
bool Thread::Stop()
{
	if ( IsRunning() )
	{
		RequestTermination();
		pthread_join( m_id, 0 );
		return true;
	}
	return false;
}

/**
	Caller blocks until the thread's Run() method terminates.

    Calling Join from multiple different threads results in undefined behaviour
    (this is due to the specification for pthread_join).
**/
bool Thread::Join() const
{
 	if ( 0 == pthread_join( m_id, 0 ) )
	{
		return true;
	}
	else
	{
		return false;
	}
}

/**
    Forcibly cancel the current thread. This is done using asynchronous cancellation so
    by calling this the thread could exit at any point (potentially leaking resources). 
**/
void Thread::Kill()
{
    pthread_cancel( m_id );
}

/**
	Ask the thread to terminate its run method. This is more polite than Kill() since it gives
    the thread the opportunity to clean up. The implementor of the thread must explicitly handle
    these requests e.g:

\code
    MyThread::Run()
    {
        while ( !TerminationRequested() )
        {
            // do processing
        }

        // free resources here
    }
\endcode

	If the Run() method does not check for termination requests by calling TermintationRequested() then the request is effectively ignored. The Run() method may also check and then ignore the request.

	If the thread is not running then this call has no effect.
**/
void Thread::RequestTermination()
{
    if ( IsRunning() )
    {
    m_termReq = true;
    }
}
/**
    Test if the thread is running.

    @return True if the thread's Run() method is active, false otherwise.
**/
bool Thread::IsRunning() const
{
    return m_running;
}

/**
    Test if two thread objects refer to the same running thread.
    
    If either of the threads are not running then this returns false.

    @return True If t is running and equal to this thread.
**/
bool Thread::operator ==  ( const Thread& t ) const
{
    bool equal;

    if ( !IsRunning() || !t.IsRunning() )
    {
        equal = false;
    }
    else
    {
        equal = 0 != pthread_equal( m_id, t.m_id );
    }

    return equal;
} 

/**
	Make (the calling thread) sleep for a specified number of milliseconds.
	@param msecs Time to sleep in milliseconds
**/
void Thread::Sleep( unsigned int msecs )
{
#ifdef WIN32
	::Sleep( msecs );
#else
	usleep( msecs*1000 );
#endif
}

} // namespace GLK

