#ifndef GLK_WIN_CONDITION_VARIABLE_H
#define GLK_WIN_CONDITION_VARIABLE_H

#include <time.h>
#include <sys/timeb.h>

#include "../utility/GlobalDefs.h"

namespace GLK
{

/**
	Implements the condition variable synchronisation primitive.
	
    This is a wrapper around a CONDITION_VARIABLE object.

	Features:
		- Uncopyable.	
**/
class FW_API ConditionVariable
{
public:
	//! Creates and initialises a condition variable.
	ConditionVariable() { InitializeConditionVariable( &m_cond ); };
	
	//! Destroys the condition variable. It is an error to destroy a condition variable on which a thread is waiting.
	~ConditionVariable() { };

	inline bool Wait( Mutex& m );
	inline bool TimedWait( Mutex& m, unsigned long int millisecs );
	inline void WakeOne();
	inline void WakeAll();

private:
	SET_UNCOPYABLE( ConditionVariable );

	CONDITION_VARIABLE m_cond;
};

/**
	Have the calling thread block until the condition variable is
	signalled.

	First, the specified mutex is unlocked, then calling thread sleeps until
	the condition variable is signalled.  If the mutex is not already
	locked then the function immediately returns false.

	Otherwise the thread blocks until the condition is signalled, at
	which point the mutex is re-locked and the function returns true.

	@param mutex A (locked) mutex that will be used to protect the condition.	
	
	@return False in case of an error (if the mutex is not owned by the calling thread or is not locked), true otherwise.
**/
bool ConditionVariable::Wait( Mutex& mutex )
{
	bool val = SleepConditionVariableCS( &m_cond, &mutex.m_critical, INFINITE );
	assert( val != 0 );
	return val;
}

/**
	Same as Wait( Mutex& mutex ) but only waits up to a specified amount of time.

    In debug mode any error causes an assertion to fail.

	@param mutex A (locked) mutex that will be used to protect the condition.
	@param millisecs Maximum time to wait in microseconds.

	@return false in case of timeout, true otherwise.
**/
bool ConditionVariable::TimedWait( Mutex& mutex, unsigned long int millisecs )
{
	bool val = SleepConditionVariableCS( &m_cond, &mutex.m_critical, millisecs );
	assert( val != 0 );
	return val;
}

/**
	Wake at least one thread that is waiting on this condition.
**/
void ConditionVariable::WakeOne()
{
	WakeConditionVariable( &m_cond );
}

/**
	Wake all threads that are waiting on this condition.
**/
void ConditionVariable::WakeAll()
{
	WakeAllConditionVariable( &m_cond );
}

} // END namespace GLK

#endif // GLK_WIN_CONDITION_VARIABLE_H
