#ifndef GLK_WIN_MUTEX_H
#define GLK_WIN_MUTEX_H

#include "../fwconfig.h"

#include <assert.h>

#include "../utility/GlobalDefs.h"
#include "Thread.h"

namespace GLK
{

class MutexLock;

/**
   Implements the Mutex primitive for creating critical sections/mutual exclusion between threads.

   Note this type of mutex can not be used for inter-process synchronisation. It only manages
   inter-thread locking within a process.

   This is a wrapper around a CRITICAL_SECTION object.
**/
class FW_API Mutex
{
    friend class ConditionVariable;

public:

    enum Mode
    {
        Recursive,
        NonRecursive
    };

    static void Test();

    inline Mutex( Mode mode = NonRecursive );
    inline ~Mutex();

    inline void Lock();
    inline void Unlock();
    inline bool TryLock();

protected:
    CRITICAL_SECTION m_critical;

private:
    SET_UNCOPYABLE(Mutex)
};

/**
    The constructor initialises the mutex.
    Mutex is initially unlocked.

    A mutex can be locked multiple times (but only by the same thead).
    It must be unlocked the same numer of times before it can be locked by a
    different thread. If a thread attempts to relock a non-recursive mutex then
    the behaviour is undefined but it will possibly block forever.

	NOTE: This is a wrapper around a CRITICAL_SECTION which is always recursive.

    @param mode Indicates whether the mutex should be recursive or not.
**/
Mutex::Mutex( Mode mode )
{
    switch ( mode )
    {
        case Recursive:
        {
            break;
        }
        case NonRecursive:
        {
            break;
        }
    }

    InitializeCriticalSection( &m_critical );
}

/**
   Destroy the mutex.
   If the mutex is locked unlock it.
**/
Mutex::~Mutex()
{
    DeleteCriticalSection( &m_critical );
}

/**
   Locks the mutex.
   If the mutex is already locked, call will block until it is available for locking.
**/
void Mutex::Lock()
{
    EnterCriticalSection( &m_critical );
}

/**
   Unlocks the mutex.
**/
void Mutex::Unlock()
{
    LeaveCriticalSection( &m_critical );
}

/**
   Try and lock a mutex but do not block if it fails.
   @return True if locking suceeds, false otherwise.
**/
bool Mutex::TryLock()
{
    return TryEnterCriticalSection( &m_critical );
}

} // end namespace GLK

#endif // GLK_WIN_MUTEX_H
