#ifndef LINUX_TIMER_H
#define LINUX_TIMER_H

#include <sys/time.h>
#include <stdint.h>

namespace GLK
{

class Timer
{
 public:
  inline Timer();

  inline void Reset();
  inline float GetSeconds();
  inline uint64_t GetMicroSeconds();
  inline uint32_t GetMilliSeconds();

 private:
  timeval m_init;
};

/**
   Construct a timer.
   Calls Reset(), so subsequent calls to
   GetSeconds() return time since object
   construction.
**/
Timer::Timer()
{
  Reset();
}

/**
   Reset the timer.
**/
void Timer::Reset()
{
  gettimeofday( &m_init, 0 );
}

/**
    Returns the time since Reset() was last called.
    If no call to Reset() has been made then the
    return value is the time elapsed since the object
    was created.

    @return The elapsed time in seconds.
**/
float Timer::GetSeconds()
{
    timeval current;
    gettimeofday( &current, 0 );
  
    long int secs = current.tv_sec - m_init.tv_sec;
    long int usecs = current.tv_usec - m_init.tv_usec;

    return secs + (usecs*0.000001f);
}

/**
    Returns the time since Reset() was last called.
    If no call to Reset() has been made then the
    return value is the time elapsed since the object
    was created.

    @return The elapsed time in milliseconds.
**/
uint32_t Timer::GetMilliSeconds()
{
    timeval current;
    gettimeofday( &current, 0 );
  
    long int secs  = current.tv_sec  - m_init.tv_sec;
    long int usecs = current.tv_usec - m_init.tv_usec;
 
    return (secs*1000) + (usecs/1000);
}

/**
    Returns the time since Reset() was last called.
    If no call to Reset() has been made then the
    return value is the time elapsed since the object
    was created.

    @return The elapsed time in microseconds.
**/
uint64_t Timer::GetMicroSeconds()
{
    timeval current;
    gettimeofday( &current, 0 );
  
    long int secs  = current.tv_sec  - m_init.tv_sec;
    long int usecs = current.tv_usec - m_init.tv_usec;

    return (secs*1000000) + usecs;
}

} //  end namespace GLK

#endif // end LINUX_TIMER_H
