#include "String.h"

#include <string.h>

#include <assert.h>

#include <ctype.h>

namespace GLK
{

/**
	Construct an empty string.
	Optionally allocate some initial capacity.

	@param l Initial capacity.
**/
String::String( unsigned int l )
:
	m_pData		(0),
	m_capacity	(l),
	m_size		(0)
{
	if ( m_capacity	> 0 )
	{
		m_pData = new char[m_capacity];
	}
}

/**
	Construct a string from an array of type 'char'.
	If capacity is not specified then allocate space only for strlen(data)+1,
	otherwise allocate the string with extra capcity (if capacity is smaller
	than the length of data then enough space to fit the data is allocated).
**/
String::String( const char* data, unsigned int capacity )
:
	m_pData (0),
	m_size	( strlen( data ) )
{
	if ( m_size + 1 < capacity )
	{
		m_capacity = capacity;
	}
	else
	{
		m_capacity = m_size + 1;
	}

	assert( m_size < m_capacity );

	m_pData = new char[m_capacity];

	strncpy( m_pData, data, m_size );
    m_pData[ m_size ] = '\0';
}

String::String( const String& s )
:
	m_pData		(0),
	m_capacity	(s.m_capacity),
	m_size		(s.m_size)
{
	if ( s.m_capacity != 0 )
	{
		m_pData = new char[s.m_capacity];
		strncpy( m_pData, s.m_pData, s.m_size );
        m_pData[ m_size ] = '\0';
	}
}

/**
	Delete the string's data.
**/
String::~String()
{
	delete [] m_pData;
}

/**
	Print the string to a file.
**/
void String::Print( FILE* fp )
{
	if ( IsEmpty())
	{
		fprintf( fp, "(empty)" );
	}
	else
	{
		fprintf( fp, "%s", m_pData );
	}
}

/**
	Make a copy of the string data.
	The copy does not necessarily have the same
	capacity as 's' or as it did before the assignment.
**/
String& String::operator = ( const String& s )
{
	m_size = s.m_size;

	if ( m_capacity < m_size+1 )
	{
		m_capacity = m_size+1;
		delete [] m_pData;
		m_pData = new char[m_capacity];
	}

	strncpy( m_pData, s.m_pData, m_size );
	m_pData[m_size] = '\0';

	return *this;
}

/**
	Append 's' to the end of this string;
**/
String& String::operator += ( const String& s )
{
	if ( IsEmpty() )
	{
		*this = s;
	}
	else
	{
		size_t new_size = m_size + s.m_size;

		if ( m_capacity < new_size+1 )
		{
		    m_capacity = new_size+1;

			char* oldBuff = m_pData;
			m_pData = new char[m_capacity];

			strncpy( m_pData, oldBuff, m_size );
			strncpy( m_pData+m_size, s.m_pData, s.m_size );

            m_pData[new_size] = '\0';

			delete [] oldBuff;
		}
		else
		{
			strncpy( m_pData+m_size, s.m_pData, s.m_size );
		}

		m_size = new_size;
	}

	return *this;
}

/**
	Append character 'c' to the end of this string;
**/
String& String::operator += ( char c )
{
    if ( IsEmpty() )
	{
		*this = String( "", 1 );
	}

	size_t new_size = m_size + 1;

	if ( m_capacity < new_size+1 )
	{
			m_capacity = new_size+1;

			char* oldBuff = m_pData;
			m_pData = new char[m_capacity];

			strncpy( m_pData, oldBuff, m_size );
			*(m_pData+m_size) = c;

			delete [] oldBuff;
	}
	else
	{
        *(m_pData+m_size) = c;
	}

	m_size = new_size;
    m_pData[ m_size ] = '\0';

	return *this;
}

/**
    Compare two strings for equality.
    
    Currently a wrapper around strncmp.

    @return true if strings are equal, false otherwise.
**/
bool String::operator == ( const String& s ) const
{
    return strncmp( cStr(), s.cStr(), m_size ) == 0;
}

/**
    Compare two strings for inequality.
    
    Currently a wrapper around strncmp.

    @return true if strings are different, false otherwise.
**/
bool String::operator != ( const String& s ) const
{
    return !(*this == s);
}

/**
	Store the substring from index 'start' to index 'end' plus a string terminator, into 's'.
**/
void String::SubString( String& s, size_t start, size_t end )
{
    assert( start < m_size );

    if ( end > m_size-1 )
    {
        end = m_size-1;
    }

    s.m_size = end - start + 1;

    if ( s.m_capacity < s.m_size + 1 )
    {
        s.m_capacity = s.m_size + 1;
        delete [] s.m_pData;
        s.m_pData = new char[s.m_capacity];
    }

    strncpy( s.m_pData, m_pData+start, s.m_size );
    s.m_pData[s.m_size] = '\0';
}

/**
    Insert the string after the specified index.
**/
void String::Insert( const String& s, unsigned int i )
{
    if ( s.m_size != 0 )
    {
        assert( i < m_size+1 );

        m_capacity += s.m_size;
        char* pTmp = new char[ m_capacity ];

        memcpy( pTmp, m_pData, i );
        memcpy( pTmp+i, s.m_pData, s.m_size );
        memcpy( pTmp+i+s.m_size, m_pData+i, m_size-i );

        m_size += s.m_size;

        delete [] m_pData;
        m_pData = pTmp;
        m_pData[m_size] = '\0';
    }
}

/**
    Insert the given character at the specified index.
**/
void String::Insert( char c, unsigned int i )
{
    assert( i <= m_size );

    (*this) += c; // just to make space if it isn't there

    unsigned int j;
    for ( j=m_size-1; j>i; --j )
    {
        m_pData[j] = m_pData[j-1];
    }
    m_pData[j] = c;
}

/**
    Remove the character at the given index
    (Moving all later chars down to fill gap).

    E.g:

    string.Insert( 'x', i );
    string.Delete( i );

    would leave string unchanged (except capacity may have increased).

**/
void String::Delete( unsigned int i )
{
    assert( i < m_size );

    for ( unsigned int j=i; j<m_size; ++j )
    {
        m_pData[j] = m_pData[j+1];
    }

    m_size -= 1;
}

/**
    Remove whitespace from beginning and end of string, storing the result in a new string.
**/
void String::Trim( String& trimmedString )
{
    // first find new beginning and end position
    int start = 0;
    while ( isspace( m_pData[start] ) )
    {
        ++start;
    }
    
    int end = Size()-1;
    while ( isspace( m_pData[end] ) )
    {
        --end;
    }

    SubString( trimmedString, start, end );
}

} // END namespace GLK
