#include "glk.h"
#include "libluaglk.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <assert.h>

extern "C"
{
#include <lauxlib.h>
}

#define LUA_IMPORT_NAME "glk"

#ifdef NDEBUG
// Load debug versions of glk libs:
#define LIBGLKCORE_NAME "libglkcore_d.so"
#define LIBGLK_NAME "libglk_d.so"
#else
// Load release versions of glk libs:
#define LIBGLKCORE_NAME "libglkcore.so"
#define LIBGLK_NAME "libglk.so"
#endif

#define MAX_LIBS 16

/*
    These are all the libraries needed to load GLK on Linux,
    (ending with glk itself):
*/
static const char* libs[MAX_LIBS]=
{
    "librt.so",
    "libX11.so",
    "libXrender.so",
    "libGL.so",
    "libfreetype.so",
    LIBGLKCORE_NAME, 
    LIBGLK_NAME,
    0             // <= end of list indicated with null string pointer! 
};

static void* handles[MAX_LIBS];

/**
    Return the named library's handle or null if it is not found/loaded.
**/
void* HandleFromLibName( const char* libname )
{
    unsigned int i = 0;
    while ( libs[i] != 0 )
    {
        if ( 0 == strncmp( libname, libs[i], strlen( libname ) ) )
        {
            return handles[i];
        }
        ++i;
    }
    
    return 0;
}

/**
    Register the GLK api with the specified Lua state.
**/
int RegisterLuaGlk( lua_State* L )
{
    void* glkHandle = HandleFromLibName( "libglk_d.so" );

    // Dynamically load the API ( defined in src/glk/lua/StandAloneLua.cpp):
    typedef int(*LuaFuncType)(lua_State*);
    LuaFuncType NewWindow = (LuaFuncType)dlsym( glkHandle, "NewWindow" );
    LuaFuncType DeleteWindow = (LuaFuncType)dlsym( glkHandle, "DeleteWindow" );

    // Register the API with the glk library object:
    static const luaL_reg libglk[] = {
        { "NewWindow", NewWindow },
        { "DeleteWindow", DeleteWindow },
        { NULL, NULL}
    };

    luaL_register( L, LUA_IMPORT_NAME, libglk );
    
    // Set some non-function fields e.g. version numbers etc:
    lua_pushnumber( L, GLK_MAJOR_VERSION );
    lua_setfield( L, -2, "major_version" );
    lua_pushnumber( L, GLK_MINOR_VERSION );
    lua_setfield( L, -2, "minor_version" );
    
    return 1; // The glk library object is now on the top of the stack.
}

/**
    Initialises the luaglk library in Lua.

    It loads the glk libs and their dependencies, and exposes
    (some minimal) functionality.
**/
LUALIB_API int luaopen_libluaglk( lua_State *L )
{
    // Import all the libs specified in the libs array
    // storing the handles in the handles array and
    // reporting errors:
    unsigned int i = 0;
    while ( libs[i] != 0 )
    {
        handles[i] = dlopen( libs[i], RTLD_GLOBAL | RTLD_NOW );
        ++i;
    
        char* err = dlerror();
        if ( err != 0 )
        {
            fprintf( stderr, "\t%s\n", err );
        }
    }
    
    // Now we can register the API:
    return RegisterLuaGlk( L );
}

