#include <glkcore.h>

#include <string.h>
#include <stdio.h>

#include <gtest/gtest.h>

#include "TestThread.h"

namespace GLK
{

void String::Test()
{
    // test creating empty string
    String s("");
    ASSERT_TRUE( s.IsEmpty() );
    ASSERT_TRUE( s.Size() == 0 );

    // test assignment
    s = String( "123" );
    ASSERT_TRUE( s.Size() == 3 );
    ASSERT_TRUE( s.IsEmpty() == false );

    // Test copy constructor:
    String copy( s );
    ASSERT_STREQ( "123", copy.cStr() );

    // test cStr()
    ASSERT_STREQ( "123", s.cStr() );

    // test +=
    s += String("ab");
    s += 'c';
    ASSERT_STREQ( "123abc", s.cStr() );

    // test insert
    s.Insert( String("HELLO"), 3 );
    ASSERT_STREQ( "123HELLOabc", s.cStr() );

    s.Insert( String("WORLD"), 11 );
    ASSERT_STREQ( "123HELLOabcWORLD", s.cStr() );

    s.Insert( '!', 16 );
    ASSERT_STREQ( "123HELLOabcWORLD!", s.cStr() );

    // Test sub-string
    s = String( "123HELLOabc" );
    ASSERT_STREQ( "123HELLOabc", s.cStr() );

    String sub;
    // check substring from 0..Size() gives whole string:
    s.SubString( sub, 0, s.Size() );
    ASSERT_STREQ( s.cStr(), sub.cStr() );
    
    s.SubString( sub, 0, 2 );
    ASSERT_STREQ( "123", sub.cStr() );

    s.SubString( sub, 3, 7 );
    ASSERT_STREQ( "HELLO", sub.cStr() );

    assert( s[10] == 'c');

    s.SubString( sub, 10, 11 );
    ASSERT_STREQ( "c", sub.cStr() );

    s.SubString( sub, 10, 1000 );
    ASSERT_STREQ( "c", sub.cStr() );

    // Test delete
    s.Delete( 3 );
    ASSERT_STREQ( "123ELLOabc", s.cStr() );

    s.Delete( 0 );
    ASSERT_STREQ( "23ELLOabc", s.cStr() );

    s.Delete( s.Size()-1 );
    ASSERT_STREQ( "23ELLOab", s.cStr() );
    
    String trimMe( "   +_Wjsdh i \t \r\n " );
    trimMe.Trim( sub );
    ASSERT_STREQ( "+_Wjsdh i", sub.cStr() );
}

/**
    Static test function.

    Opens a new file for writing, writes know contents, closes it.
    Then open same file for reading, read back and check contents using various member functions.
**/
void File::Test()
{
    // create some test data
    char data[1024] = "";
    int length = 0;
    for ( int i=0; i<10; ++i )
    {
        length += sprintf( data + length, "tuple: %d, %d\n", i, i+1 );
    }

    // open for writing test
    File* f = new File( "test-file.txt", "w" );
    ASSERT_TRUE( f->IsOpen() );
    f->Write( data, length );
    delete f;
    
    // open for reading test
    f = new File( "test-file.txt", "r" );
    ASSERT_TRUE( f->IsOpen() );

    // test Scanf
    int a = -1;
    int b = -1;
    for ( int i=0; i<10; ++i )
    {
        f->Scanf( "tuple: %d,%d\n", &a, &b );
#ifndef WIN32
        // Scanf doesnt work on win32 because of Microsoft
        ASSERT_TRUE( a==i && b==i+1 );
#endif
    }    

    // test Rewind and ReadLine
    f->Rewind();

    char tmp[128];
    for ( int i=0; i<10; ++i )
    {
        f->ReadLine( tmp, 128 );
        sscanf( tmp, "tuple: %d,%d\n", &a, &b );
        ASSERT_TRUE( a==i && b==i+1 );
    }  

    delete f;
    
    // test ToString
    String s;
    bool ok = File::FileToString( String("test-file.txt"), s );
    ASSERT_TRUE( ok );
    ASSERT_STREQ( data, s.cStr() );
}

/**
    Simple test which checks that a thread's run method gets called.
**/
void Thread::Test()
{
    TestThread t;
    t.Start(); 
    t.Join();

    ASSERT_TRUE( t.WasRun() );
}

void Mutex::Test()
{
    {   // Test non-recursive mutex
        Mutex m( NonRecursive );
        EXPECT_TRUE( m.TryLock() );
        EXPECT_FALSE( m.TryLock() );
        m.Unlock();
    }    
    
    {   // Test a recursive mutex
        Mutex m( Recursive );
        EXPECT_TRUE( m.TryLock() );     
        EXPECT_TRUE( m.TryLock() );
        m.Unlock();
        m.Unlock();
    }
}

void Semaphore::Test()
{
    Semaphore s( 0 );

    ASSERT_FALSE( s.TryAcquire() );

    s.Release( 1 );

    ASSERT_FALSE( s.TryAcquire(2) );
    ASSERT_TRUE( s.TryAcquire(1) );
}

void Barrier::Test()
{
    TestBarrier* t[5];
    Barrier b( 6 );

    EXPECT_TRUE( b.NumWaiting() == 0 );

    for ( int i = 0; i<5; ++i )
    {
        t[i] = new TestBarrier( b );
        t[i]->Start();
    }

    Thread::Sleep( 5 );

    for ( int i = 0; i<5; ++i )
    {
        EXPECT_FALSE( t[i]->PastBarrier() );
    }

    EXPECT_TRUE( b.NumWaiting() == 5 );
    b.Wait();

    Thread::Sleep( 5 );

    for ( int i = 0; i<5; ++i )
    {
        EXPECT_TRUE( t[i]->PastBarrier() );
    }    
    
    EXPECT_TRUE( b.NumWaiting() == 0 );
    EXPECT_TRUE( b.NumCycles() == 1 );
}

template <>
FW_API void MessageQueue<int>::Test()
{
    MessageQueue<int> m( 10 );

    ASSERT_TRUE( m.IsEmpty() );
    ASSERT_TRUE( m.NumFilled() == 0 );
    ASSERT_TRUE( m.NumFree() == 10 );

    // Test nothing to read from empty queue
    assert( m.TryRead( 0 ) == false );

    for ( int i=0; i<10; ++i )
    {
        m.Write( &i );
    }

    // Test can't write to empty queue
    ASSERT_TRUE( m.TryWrite( 0 ) == false );

    // Test read (should be FIFO)
    int j;
    for ( int i=0; i<10; ++i )
    {
        m.Read( &j );
        ASSERT_TRUE( j == i );
    }
}

} // end of namespace GLK

